/* Store data for a polygon mesh: vertexes in space, polygons, attributes
from polygons.

Vertexes: are locations in space (have 3 components: X, Y, Z)
Polygons: define index vertexes that compose one polygon
Attributes: define if polygon is a portal and your texture ID,...

WARNING:
- after change mesh data call CMesh::mBeginAddMesh and imediatly before
  call CMesh::mEndAddMesh.
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "node.h"

#include "nodelii.h"

#include "list.h"

#include "map.h"

#include "polygon.h"

#include "nodepol.h"

#include "nodeflo.h"

#include "nodexyz.h"

#include "math.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"




/* constructor - init empty sector
**********************************************************/
CMesh::CMesh ()
{
}



/* destructor - desalocate data, if alocated
**********************************************************/
CMesh::~CMesh ()
{
   mDestroy ();
}



/* Prepare mesh to reseve data: disable direct access feature.
return TRUE - ok / FALSE - error
Call after change mesh data, before call CMesh::mEndAddMesh
**********************************************************/
BOOL CMesh::mBeginAddMesh ()
{
   m_oPolygons.mDisableDirectAccess ();
   m_oVertexes.mDisableDirectAccess ();
   m_oVertLight.mDisableDirectAccess ();

   return TRUE;
}



/* Update mesh data acording polygons and vertexes definition:
recompute polygons normals and enable direct access to data lists.
return TRUE - ok / FALSE - error
Call after define all data and before use this data.
**********************************************************/
BOOL CMesh::mEndAddMesh ()
{
   // try enable direct access feature (speed up random access)
   if ( ! m_oPolygons.mEnableDirectAccess () ||
        ! m_oVertexes.mEnableDirectAccess () ||
        ! m_oVertLight.mEnableDirectAccess () )
      return FALSE;   // if error ocurr return fail

   // compute polygons plan equation (used in collision detection,...)
   CMath
      oM;    // the math function is inside this object
   m_oPolygons.mSelectStart ();
   while ( m_oPolygons.mRetExistSelection () ) {

      // the current polygon
      CNodePolygon
         *poPol = (CNodePolygon*) m_oPolygons.mRetSelected ();
      
      oM.mDefinePolyPlanEquation ( m_oVertexes, poPol );

      // next polygon
      m_oPolygons.mSelectNext ();
   }

   return TRUE;
}



/* Desalocate mesh data, if alocated
**********************************************************/
void CMesh::mDestroy ()
{
   // remove data
   m_oPolygons.mRemoveAll ();
   m_oVertexes.mRemoveAll ();
   m_oVertLight.mRemoveAll ();
}



/* Append the mesh with more data.
Remember call before: mBeginAddMesh and after mEndAddMesh.
**********************************************************/
void CMesh::mAppendObj ( CMesh& oSrc )
{
   // copy the polygons from the source to new
   oSrc.m_oPolygons.mSelectStart ();
   while ( oSrc.m_oPolygons.mRetExistSelection () ) {

      mAddPolygon ( (CNodePolygon*) oSrc.m_oPolygons.mRetSelected () );
      oSrc.m_oPolygons.mSelectNext ();
   }

   // copy the vertexes from the source to new
   oSrc.m_oVertexes.mSelectStart ();
   while ( oSrc.m_oVertexes.mRetExistSelection () ) {

      mAddVertex (
         ( (CNodeXYZ*) oSrc.m_oVertexes.mRetSelected () ) -> m_stXYZ.fX,
         ( (CNodeXYZ*) oSrc.m_oVertexes.mRetSelected () ) -> m_stXYZ.fY,
         ( (CNodeXYZ*) oSrc.m_oVertexes.mRetSelected () ) -> m_stXYZ.fZ,
         1.0, DEFAULT_LIGHT_LEVEL );   // no scale (model was allready scaled)
     oSrc.m_oVertexes.mSelectNext ();
   }

   // copy vertexes light levels from the source to new
   oSrc.m_oVertLight.mSelectStart ();
   m_oVertLight.mSelectStart ();          // nodes created in mAddVertex
   while ( oSrc.m_oVertLight.mRetExistSelection () ) {

      // attribution
      ( (CNodeLightInfo*) m_oVertLight.mRetSelected () ) -> m_stValue = 
         ( (CNodeLightInfo*) oSrc.m_oVertLight.mRetSelected () ) -> m_stValue;
      
      // next
      oSrc.m_oVertLight.mSelectNext ();
      m_oVertLight.mSelectNext ();
   }
}



/* 1 - Add polygons
oIndexes is the list with vertex indexes that compose the polygon,
the indexes are acording vertex list in mesh.
fLight - the default poly surface light level
dwLightID - the default poly surface light table
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddPolygon ( CList& oIndexes, DWORD dwLightID, float fLight )
{
   // try create new polygon
   CNodePolygon
      *poPol = new CNodePolygon;
   if ( poPol == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // create a new polygon
   poPol->m_oPolygon.mCreate ( oIndexes, dwLightID, fLight );

   // update mesh adding polygon and attribute
   m_oPolygons.mSetEnd ( (CNode*) poPol );

   return TRUE;
}



/* 2- Add polygons
Make a copy from the source polygon poSrc.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddPolygon ( CNodePolygon* poSrc )
{
   // try create new polygon
   CNodePolygon
      *poPol = new CNodePolygon;
   if ( poPol == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // copy
   poPol->m_oPolygon.mSetObj ( poSrc->m_oPolygon );

   // update mesh adding polygon
   m_oPolygons.mSetEnd ( (CNode*) poPol );

   return TRUE;
}



/* Add vertexes coordenates.
fScale - the scale factor for the vector (default is 1)
fNormLightLevel - the initial vertex light level (default DEFAULT_LIGHT_LEVEL)
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddVertex ( float fX, float fY, float fZ, float fScale, float fNormLightLevel )
{
   // try create new vertex
   CNodeXYZ
      *poVert = new CNodeXYZ;
   if ( poVert == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define vertex data (scale vertex)
   poVert->m_stXYZ.fX = fX * fScale;
   poVert->m_stXYZ.fY = fY * fScale;
   poVert->m_stXYZ.fZ = fZ * fScale;

   // update data
   m_oVertexes.mSetEnd ( (CNode*) poVert );

   // try create vertex light information
   CNodeLightInfo
      *poLight = new CNodeLightInfo;
   if ( poLight == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define initial light info
   poLight->m_stValue.fAmbient = fNormLightLevel;
   poLight->m_stValue.fSource = 0;

   // update data
   m_oVertLight.mSetEnd ( (CNode*) poLight );

   return TRUE;
}



/* Set polygon's data.
nTypeID - type of data
pafData - array with the data
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their data changed.
Convension: TRUE = 1, FALSE = 0
return TRUE - ok / FALSE - indexes out of range
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mSetDataValues ( INT nTypeID, float* pafData, DWORD dwNdxPol1, DWORD dwNdxPol2 )
{
   // sort indexes
   DWORD
      dwMin = dwNdxPol1,
      dwMax = dwNdxPol2;
   if ( dwNdxPol1 > dwNdxPol2 ) {
      dwMin = dwNdxPol2,
      dwMax = dwNdxPol1;
   }

   // verify if polygons indexes in mesh are out of range, if yes return
   if ( dwMax >= m_oPolygons.mRetNumNodes () ) {
      printf ( MSG_ERROR_INDEX_OUT_OF_RANGE );
      return FALSE;
   }

   // set values
   for ( DWORD dwcI = dwMin; dwcI <= dwMax; dwcI ++ ) {

      // the polygon
      CNodePolygon
         *poPol = (CNodePolygon*) m_oPolygons [ dwcI ];

      // poly data to update
      switch ( nTypeID ) {

         case ID_PORTALS:
            poPol->m_oPolygon.m_stPortal.yCode = (BYTE) pafData [ 0 ];
            poPol->m_oPolygon.m_stPortal.dwSectorID = (DWORD) pafData [ 1 ];
            poPol->m_oPolygon.m_stPortal.poSector = NULL;
            break;

         case ID_SOLIDSTATE:
            poPol->m_oPolygon.m_yCrossable = (BYTE) pafData [ 0 ];
            break;

         case ID_POLYGONSLIGHT:
            poPol->m_oPolygon.m_dwLightID = (DWORD) pafData [ 0 ];
            poPol->m_oPolygon.m_fLight = pafData [ 1 ];
            break;

         case ID_POLYGONSSURFACE:
            poPol->m_oPolygon.m_dwTextureID = (DWORD) pafData [ 0 ];
            break;
      }
   }

   return TRUE;
}



/* Make collision detection between a sphere and the mesh.
stPoint - the point to test (sphere center)
fRadius - sphere radius
return:
        - if portal crossing, point the portal poly;
        - if colliding with solid poly (no portal crossing)
          point the first detected poly;
        - if no collision: NULL.
return yCode:
       bit 0: 0 - no collision
              1 - collision with a solid polygon
       bit 1: 0 - no portal crossing
              1 - portal crossing
to access the bit values use: value & 2^bit_order
**********************************************************/
CNodePolygon* CMesh::mDetectCollision ( tagfXYZ& stPosNew,
                                        float fRadius,
                                        BYTE& yCode )
{
   // TRUE - if colliding with one solid poly
   BOOL
      bCollide = FALSE;

   yCode = 0;

   // default: no poly collision
   CNodePolygon
      *poPolCollide = NULL;

   // scan all polygons verifying the distance
   // for each polygon in mesh
   m_oPolygons.mSelectStart ();
   while ( m_oPolygons.mRetExistSelection () ) {

      // the current polygon to test
      CNodePolygon
         *poPol = (CNodePolygon*) m_oPolygons.mRetSelected ();

      // find the distance between you and the poly
      float
         fDistance = m_oMath.mDistancePointPlan ( stPosNew,
                                                  poPol->m_oPolygon.m_stNormal,
                                                  poPol->m_oPolygon.m_fD );

      // if your sphere intersect the poly plan
      if ( fDistance <= fRadius ) {   // distance test

         // verify if need test if intersection point is inside poly
         // edges or not
         // flag if point is inside (TRUE) or outside (FALSE)
         BOOL
            bPointInside = FALSE;
         // verify collision detection testing if point is inside poly
         if ( poPol->m_oPolygon.m_yCrossable >= 2 ) {      // TXE code 2 or 3

            // the intersection point between plan and sphere (projected center
            // of sphere to poly plan)
            tagfXYZ
               stPoint;
            stPoint.fX = stPosNew.fX - poPol->m_oPolygon.m_stNormal.fX * fDistance;
            stPoint.fY = stPosNew.fY - poPol->m_oPolygon.m_stNormal.fY * fDistance;
            stPoint.fZ = stPosNew.fZ - poPol->m_oPolygon.m_stNormal.fZ * fDistance;

            // test    REVIEW - buggy?
            bPointInside = m_oMath.mPointInsidePoly ( stPoint, poPol, m_oVertexes );
         }

         // verify collision detection:
         // object not has collided in other poly
         // if only need test plan distance (poPol->m_oPolygon.m_stPol.bCrossable==1)
         // if need test if intersection point is inside poly (bPointInside TRUE)
         if ( ! bCollide                                       && // not collided
              ( poPol->m_oPolygon.m_yCrossable == 1 ||                       // plan test
                ( poPol->m_oPolygon.m_yCrossable == 2 && bPointInside ) ) && // point test
              fDistance >= - fRadius ) {                          // sphere completely behind poly

            yCode = yCode | 1;
            bCollide = TRUE;
            // if no portal crossing detected in previus polys
            if ( ! ( yCode & 2 ) )
               poPolCollide = poPol;
         }
         else
            // verify portal crossing:
            // if the polygon is crossable (with plan or point collision),
            // and is a portal and you are behind
            if ( poPol->m_oPolygon.m_stPortal.yCode != (BYTE) 0 && // portal
                 ( poPol->m_oPolygon.m_yCrossable == 0 ||          // crossable (plan)
                   ( poPol->m_oPolygon.m_yCrossable == 3 && bPointInside ) ) && // crossable (point)
                 fDistance < 0 ) {                   // negative distance

               yCode = yCode | 2;
               poPolCollide = poPol;
            }
      }

      // next polygon to test
      m_oPolygons.mSelectNext ();
   }

   return poPolCollide;
}



/* Update the mesh portal's pointer. Set pointer to the right
destin sector (an adjacent sector or internal sector).
poSectors - the set of adjacent sectors
**********************************************************/
void CMesh::mSetPortalPointers ( CMap* poSectors )
{
   // for each poly
   m_oPolygons.mSelectStart ();
   while ( m_oPolygons.mRetExistSelection () ) {

      // the current polygon
      CNodePolygon
         *poPol = (CNodePolygon*) m_oPolygons.mRetSelected ();
      
      // verify if poly is a portal
      if ( poPol->m_oPolygon.m_stPortal.yCode != 0 ) // if yes

         // Verify if poly is a portal to interior or between 2 sectors
         // inside the object.
         if ( poPol->m_oPolygon.m_stPortal.yCode == 1 )
            // update pointer
            poPol->m_oPolygon.m_stPortal.poSector = (CNodeSector*) poSectors->mRetValue ( poPol->m_oPolygon.m_stPortal.dwSectorID );

      // next polygon
      m_oPolygons.mSelectNext ();
   }
}
