/* Implement a backpropagation neural network. Use Sigma for adaptation.

 neuron:
                     X0 = +1
         W0
 X0--------\
            \                                training vectors:
         W1  \
 X1-----------+---+  net +---+   Y            X0  X1 ...    D (desired)
              | S |------| f |---      k=0: [ 1   1     ] [ 1 ... ]
    ....      +---+      +---+         k=1: [ 1  -1     ] [ 1     ]
             /                         k=.: [ .   .     ] [ .     ]
 Xn---------/                          k=.: [ .   .     ] [ .     ]
         Wn                            k=.: [ .   .     ] [ .     ]

                    net = X0*W0 + X1*W1 + ... + XnWn
                    Y = tanh ( net )

TIPS:
- the minimum number of layers are 2: the hidden layer and the output layer
- the first hiden layer use an input vector as training vector
- input must be between -1 and 1 (bipolar)
- Layers are added like a heap: the output layer is always the last layer
  inserted and the first hidden layer is the first inserted.

CAUTION:
- when creating the network first define input vertors only then the layers
*/



#include <string.h>

#include "defs.h"

#ifdef TRG_DJGPP
   #include <bios.h>
#endif

#include "mem.h"

#include "string.h"

#include "file.h"

#include "node.h"

#include "nodeflo.h"

#include "list.h"

#include "nodelist.h"

#include "neuron.h"

#include "nodenet.h"

#include "bpn.h"



/* constructor - init random number gererator, set default
 adaptation rate
**********************************************************/
CBPN::CBPN ()
{
   m_dwMaxNeurons = 0;
   m_fRate = 0.1;

   // init the random number generator (Linux init when system boot)
   #ifdef TRG_DJGPP
      srandom ( (int) biostime ( 0, 0 ) );
   #endif
}



/* destructor
**********************************************************/
CBPN::~CBPN ()
{
   mDestroy ();
}



/* Destroy network.
**********************************************************/
void CBPN::mDestroy ()
{
   // remove all training vectors
   // for each training vector
   m_oInputs.mSelectStart ();
   // while exist training vectors (selection not NULL)
   while ( m_oInputs.mRetExistSelection () ) {

      // remove all elements in training vector
      CNodeList
         *poVector = (CNodeList*) m_oInputs.mRetSelected ();
      poVector->m_oList.mRemoveAll ();

      // select next vector
      m_oInputs.mSelectNext ();
   }

   // remove all vectors pointers
   m_oInputs.mRemoveAll ();

   // remove all network layers
   // for each layer
   m_oLayers.mSelectStart ();
   // while exist layer (selection not NULL)
   while ( m_oLayers.mRetExistSelection () ) {

      // remove all layer neurons
      CNodeList
         *poLayer = (CNodeList*) m_oLayers.mRetSelected ();
      poLayer->m_oList.mRemoveAll ();

      // select next layer
      m_oLayers.mSelectNext ();
   }

   // remove all layers pointers
   m_oLayers.mRemoveAll ();

   // remove all desired values
   // for each desired vector for train vector
   m_oD.mSelectStart ();
   // while exist desired vectors (selection not NULL)
   while ( m_oD.mRetExistSelection () ) {

      // remove all layer values
      CNodeList
         *poD = (CNodeList*) m_oD.mRetSelected ();
      poD->m_oList.mRemoveAll ();

      // select next layer
      m_oD.mSelectNext ();
   }

   // remove desired values for previous layers
   m_oDPrevLayer.mRemoveAll ();

   // remove sum values for previous layers
   m_oSum.mRemoveAll ();

   m_dwMaxNeurons = 0;
}




/* Save networks data to a file, if file exist overwrite it.
If you don't want to overwrite the file, just append, use
CBPN::mAppendSave.
Save the adaptation rate and neuron weights and net.
oszFile - string with the file and path to save
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CBPN::mSave ( CString oszFile )
{
   // create a binary file for write
   CFile
      oFile;
   if ( ! oFile.mOpen ( oszFile, "wb" ) )
      return FALSE;

   // try append data to this file
   if ( ! mAppendSave ( oFile ) )
      return FALSE;

   // close file
   oFile.mClose ();

   return TRUE;
}



/* Load networks data from a file, start reading from the
beginning of the file.
Load the adaptation rate and neuron weights and net.
This function does not create the network, only change the
current values acording file data, the layers and neurons
must be previously created. No input vectors load.
Before call this function create and fill the input vectors data.
If you don't want to start reading from the
beginning use CBPN::mAppendLoad.
oszFile - string with the file and path to load
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CBPN::mLoad ( CString oszFile )
{
   // open a binary file for read
   CFile
      oFile;
   if ( ! oFile.mOpen ( oszFile, "rb" ) )
      return FALSE;

   // try read data from this file
   if ( ! mAppendLoad ( oFile ) )
      return FALSE;

   // close file
   oFile.mClose ();

   return TRUE;
}
   


/* Append data to a file. Start writing from the current
file position (usualy the end).
Save the adaptation rate and neuron weights and net.
oFile - the file object to save data
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CBPN::mAppendSave ( CFile& oFile )
{
   // save the adaptation rate
   if ( ! oFile.mSetMem ( sizeof ( float ), (void*) &m_fRate ) )
       return FALSE;

   // save all network neurons
   
   // for each layer
   m_oLayers.mSelectStart ();
   // while exist layer (selection not NULL)
   while ( m_oLayers.mRetExistSelection () ) {

      // for each layer neuron
      CNodeList
         *poLayer = (CNodeList*) m_oLayers.mRetSelected ();
      poLayer->m_oList.mSelectStart ();
      // while exist neuron (selection not NULL)
      while ( poLayer->m_oList.mRetExistSelection () ) {

         // point neuron
         CNodeNeuron
            *poNeuron = (CNodeNeuron*) poLayer->m_oList.mRetSelected ();
         // try save neuron
         if ( ! poNeuron->m_oNeuron.mSave ( oFile ) )
            return FALSE;

         // next neuron
         poLayer->m_oList.mSelectNext ();
      }

      // select next layer
      m_oLayers.mSelectNext ();
   }

   return TRUE;
}



/* Load networks data from a file, start reading from the
current file position.
Load the adaptation rate and neuron weights and net.
This function does not create the network, only change the
current values acording file data, the layers and neurons
must be previously created. No input vectors load.
Before call this function create and fill the input vectors data.
oszFile - string with the file and path to load
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CBPN::mAppendLoad ( CFile& oFile )
{
   // load the adaptation rate
   if ( ! oFile.mRetMem ( sizeof ( float ), (void*) &m_fRate ) )
       return FALSE;

   // load all network neurons
   
   // for each layer
   m_oLayers.mSelectStart ();
   // while exist layer (selection not NULL)
   while ( m_oLayers.mRetExistSelection () ) {

      // for each layer neuron
      CNodeList
         *poLayer = (CNodeList*) m_oLayers.mRetSelected ();
      poLayer->m_oList.mSelectStart ();
      // while exist neuron (selection not NULL)
      while ( poLayer->m_oList.mRetExistSelection () ) {

         // point neuron
         CNodeNeuron
            *poNeuron = (CNodeNeuron*) poLayer->m_oList.mRetSelected ();
         // try load neuron
         if ( ! poNeuron->m_oNeuron.mLoad ( oFile ) )
            return FALSE;

         // next neuron
         poLayer->m_oList.mSelectNext ();
      }

      // select next layer
      m_oLayers.mSelectNext ();
   }

   return TRUE;
}



/* Define weights change rate
**********************************************************/
void CBPN::mSetRate ( float fRate )
{
   m_fRate = fRate;
}



/* Return weights change rate
**********************************************************/
float CBPN::mRetRate ()
{
   return m_fRate;
}



/* Add a layer to end of neural network
dwNumNeurons - number of neurons in layer
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CBPN::mAddLayer ( DWORD dwNumNeurons )
{
   // try create a new layer
   CNodeList
      *poLayer = new CNodeList;
   if ( poLayer == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // create neurons layer
   for ( DWORD dwcN = 0; dwcN < dwNumNeurons; dwcN ++ ) {

      // try create a new neuron
      CNodeNeuron
         *poNeuron = new CNodeNeuron;
      if ( poNeuron == NULL ) {
         printf ( MSG_ERROR_MEMORY_LOW );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }

      // verify if is not the first layer, if is not attach neuron
      // to previous neurons layer
      if ( m_oLayers.mRetNumNodes () != 0 ) {
         // try create neuron
         INT
            nCreateCode = poNeuron->m_oNeuron.mCreate ( * ( (CNodeList*) m_oLayers.mRetEnd () ) );
         if ( nCreateCode != 1 )
            return FALSE;
      }
      else {
         // verify if training vectors are not defined
         if ( m_oInputs.mRetNumNodes () == 0 ) {
            printf ( MSG_ERROR_TRAINING_VECTORS_UNDEFINIED );
            printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
            return FALSE;
         }

         // define the number of nodes in input vectors
         CNodeList
            *poTrainVector = (CNodeList*) m_oInputs.mRetStart ();

         // try create neuron weights
         INT
            nCreateCode = poNeuron->m_oNeuron.mCreate ( *poTrainVector );
         if ( nCreateCode != 1 )
            return FALSE;
      }

      //add neuron to layer
      poLayer->m_oList.mSetEnd ( poNeuron );
   }

   // verify if is the biggest layer
   if ( m_dwMaxNeurons < dwNumNeurons )
      m_dwMaxNeurons = dwNumNeurons;

   // add layer to layer list
   m_oLayers.mSetEnd ( poLayer );

   // define last inserted layer as output vector
   m_oOutput.mSetObj ( poLayer->m_oList );

   return TRUE;
}



/* Make network propagation computing neurons net value.
dwVector define the index of the input vector from
m_oInputs to be used. The index start in 0.
**********************************************************/
void CBPN::mPropagate ( DWORD dwVector )
{
   // flag first layer
   BOOL
      bFirst = TRUE;

   // for each layer
   m_oLayers.mSelectStart ();
   // while exist layer (selection not NULL)
   while ( m_oLayers.mRetExistSelection () ) {

      // for each neuron in layer
      CNodeList
         *poLayer = (CNodeList*) m_oLayers.mRetSelected ();
      poLayer->m_oList.mSelectStart ();
      // while exist neuron (selection not NULL)
      while ( poLayer->m_oList.mRetExistSelection () ) {

         // point neuron
         CNodeNeuron
            *poNeuron = (CNodeNeuron*) poLayer->m_oList.mRetSelected ();
         // call neuron propagation function
         if ( bFirst ) {
            // if neuron in first layer (the input is the training or input vectors)
            CNodeList
               oInput;
            oInput.m_oList.mSetObj ( ( (CNodeList*) m_oInputs.mRetNode ( dwVector ) )->m_oList );
            poNeuron->m_oNeuron.mPropagate ( oInput.m_oList );
         } 
         else
            // if neuron in other layers
            poNeuron->m_oNeuron.mPropagate ();

         // select next neuron in layer
         poLayer->m_oList.mSelectNext ();
      }

      // select next layer
      m_oLayers.mSelectNext ();

      // no more first layer
      bFirst = FALSE;
   }
}



/* Make network adaptation, change neurons weights.
dwVector define the index of the input vector used in
propagation and their desired values to be used in adaptation.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CBPN::mAdapt ( DWORD dwVector )
{
   // define current output desired for all output layer neurons
   CNodeList
      *poDOut = (CNodeList*) m_oD.mRetNode ( dwVector );

   // for each layer
   m_oLayers.mSelectEnd ();   // adapt from back to front
   // define current layer
   CNodeList
      *poLayer = (CNodeList*) m_oLayers.mRetSelected ();

   // verify if need grow delta values list (insert new nodes)
   if ( m_oDPrevLayer.mRetNumNodes () < m_dwMaxNeurons ) {
      // number of nodes to insert
      DWORD
         dwNumNew = m_dwMaxNeurons - m_oDPrevLayer.mRetNumNodes ();
      for ( DWORD dwcD = 0; dwcD < dwNumNew; dwcD ++ ) {
         // try create a new node
         CNodeFloat
            *poNode = new CNodeFloat;
         if ( poNode == NULL ) {
            printf ( MSG_ERROR_MEMORY_LOW );
            printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
            return FALSE;
         }
         // add to delta values
         m_oDPrevLayer.mSetEnd ( poNode );
      }
   }

   // verify if need grow sum values list (insert new nodes)
   if ( m_oSum.mRetNumNodes () < m_dwMaxNeurons ) {
      // number of nodes to insert
      DWORD
         dwNumNew = m_dwMaxNeurons - m_oSum.mRetNumNodes ();
      for ( DWORD dwcD = 0; dwcD < dwNumNew; dwcD ++ ) {
         // try create a new node
         CNodeFloat
            *poNode = new CNodeFloat;
         if ( poNode == NULL ) {
            printf ( MSG_ERROR_MEMORY_LOW );
            printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
            return FALSE;
         }
         // add to sum values
         m_oSum.mSetEnd ( poNode );
      }
   }

   // compute delta for output layer
   // select first value, neuron and output (if exist)
   m_oDPrevLayer.mSelectStart ();
   poLayer->m_oList.mSelectStart ();
   poDOut->m_oList.mSelectStart ();
   // while exist values to define (selection not NULL)
   while ( poLayer->m_oList.mRetExistSelection () ) {

      // define current value
      CNodeFloat
         *poDPrev = (CNodeFloat*) m_oDPrevLayer.mRetSelected ();
      // define current neuron
      CNodeNeuron
         *poNeuron = (CNodeNeuron*) poLayer->m_oList.mRetSelected ();
      // define current desired output value
      CNodeFloat
         *poYOut = (CNodeFloat*) poDOut->m_oList.mRetSelected ();

      // compute: D = (Y - out) * (1-out*out)
      poDPrev->m_fValue = ( poYOut->m_fValue - poNeuron->m_oNeuron.mRetAdaptationValue () ) *
                          ( 1 - poNeuron->m_oNeuron.mRetAdaptationValue () *
                                poNeuron->m_oNeuron.mRetAdaptationValue () );

      // next nodes
      m_oDPrevLayer.mSelectNext ();
      poLayer->m_oList.mSelectNext ();
      poDOut->m_oList.mSelectNext ();
   }

   // reset sum values to 0
   // select first value
   m_oSum.mSelectStart ();
   // while exist values to define (selection not NULL)
   while ( m_oSum.mRetExistSelection () ) {

      ( (CNodeFloat*) m_oSum.mRetSelected () )->m_fValue = 0;

      // next nodes
      m_oSum.mSelectNext ();
   }

   // adapt neurons for each layer
   m_oLayers.mSelectEnd ();   // adapt from back to front
   // while exist layer (selection not NULL)
   while ( m_oLayers.mRetExistSelection () ) {

      // for each neuron in current layer
      poLayer->m_oList.mSelectStart ();
      m_oDPrevLayer.mSelectStart ();
      // while exist neurons (selection not NULL)
      while ( poLayer->m_oList.mRetExistSelection () ) {

         // define current neuron
         CNodeNeuron
            *poNeuron = (CNodeNeuron*) poLayer->m_oList.mRetSelected ();

         float
            fDelta = ( (CNodeFloat*) m_oDPrevLayer.mRetSelected () )->m_fValue;

         // make neuron adaptation
         // verify if this is the last layer to adapt
         if ( m_oLayers.mRetSelected () == m_oLayers.mRetStart () )
            // if the last layer (last use train vectors as input)
            poNeuron->m_oNeuron.mAdaptLast ( m_fRate, fDelta, ( (CNodeList*) m_oInputs.mRetNode ( dwVector ) )->m_oList );
         else
            poNeuron->m_oNeuron.mAdapt ( m_fRate, fDelta, m_oSum );

         // next neuron
         poLayer->m_oList.mSelectNext ();
         m_oDPrevLayer.mSelectNext ();
      }

      // goto previous layer
      m_oLayers.mSelectPrevious ();

      // verify if this is not the last layer to adapt
      if ( m_oLayers.mRetExistSelection () ) {
         // if is not the last layer (last use train vectors as input)
         
         // points layer
         poLayer = (CNodeList*) m_oLayers.mRetSelected ();

         // compute delta
         // select first value, neuron and output (if exist)
         poLayer->m_oList.mSelectStart ();
         m_oDPrevLayer.mSelectStart ();
         m_oSum.mSelectStart ();
         // while exist values to define (selection not NULL)
         while ( poLayer->m_oList.mRetExistSelection () ) {

            // define current neuron
            CNodeNeuron
               *poNeuron = (CNodeNeuron*) poLayer->m_oList.mRetSelected ();
            // define current value
            CNodeFloat
               *poDPrev = (CNodeFloat*) m_oDPrevLayer.mRetSelected ();
            // define current desired output value
            CNodeFloat
               *poSumValue = (CNodeFloat*) m_oSum.mRetSelected ();

            // compute: D = sum * (1-out*out)
            poDPrev->m_fValue = poSumValue->m_fValue *
                                ( 1 - poNeuron->m_oNeuron.mRetAdaptationValue () *
                                      poNeuron->m_oNeuron.mRetAdaptationValue () );

            // next nodes
            poLayer->m_oList.mSelectNext ();
            m_oDPrevLayer.mSelectNext ();
            m_oSum.mSelectNext ();
         }

         // reset sum values to 0
         // select first value
         m_oSum.mSelectStart ();
         // while exist values to define (selection not NULL)
         while ( m_oSum.mRetExistSelection () ) {

            ( (CNodeFloat*) m_oSum.mRetSelected () )->m_fValue = 0;

            // next nodes
            m_oSum.mSelectNext ();
         }
      }
   }

   return TRUE;
}



/* Make network training.
dwNum is the number of interactions
**********************************************************/
void CBPN::mTrain ( DWORD dwNum )
{
   // define number of input (training) vectors
   DWORD
      dwNumVectors = m_oInputs.mRetNumNodes ();

   // make interactions
   for ( DWORD dwcI = 0; dwcI < dwNum; dwcI ++ )
      // for each input vector
      for ( DWORD dwcV = 0; dwcV < dwNumVectors; dwcV ++ ) {
         mPropagate ( dwcV );
         mAdapt ( dwcV );
      }
}
