/* Low level video graphics functions for MCGA mode: 320x200 with 256 colors.

You have some video direct access functions to put frame buffers on screen,
programm palette, draw pixels, clear screen, ... Use CDraw to draw more
complex imagens (with polygons, ...).

TIPS:
- the reference system is the same as the video:   (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX

PROGRAMING (Assembler AT&T + DJGPP):
- if writing in variable passed as argument in asm code put memory
  in third ":"
- __volatile__ enable compiler to optimize code
*/



#include <go32.h>
#include <dos.h>

#include "..\defs.h"

#include "..\mem.h"

#include "gr.h"



/* constructor
**********************************************************/
CGr::CGr ()
{
}



/* constructor - init video mode
**********************************************************/
CGr::CGr ( DWORD dwMode )
{
   mSetVideoMode ( dwMode );
}



/* destructor
**********************************************************/
CGr::~CGr ()
{
}



/* Change current video mode
**********************************************************/
void CGr::mSetVideoMode ( DWORD dwMode )
{
   REGS
      stRegsE, stRegsS;      // registers

   // define video mode code
   if ( dwMode == MODE_TEXT )
      stRegsE.x.ax = 0x0003;
   if ( dwMode == MODE_MCGA )
      stRegsE.x.ax = 0x0013;

   int86 ( 0x10, &stRegsE, &stRegsS );
}



/* Change XxY rate in MCGA mode (320x200x256).
 Only work in some video cards.
**********************************************************/
void CGr::mChangeMCGARate ()
{
   outportb ( 0x3C2, 0xE3 );
}



/* Set the RGB of one palette entry.
**********************************************************/
void CGr::mSetPalColor ( BYTE yCol, BYTE yR, BYTE yG, BYTE yB )
{
   // recall change palette
   outportb ( 0x03C8, yCol );
   // write
   outportb ( 0x03C9, yR );
   outportb ( 0x03C9, yG );
   outportb ( 0x03C9, yB );
}



/* Return RGB of one palette entry.
**********************************************************/
void CGr::mRetPalColor ( BYTE yCol, BYTE& yR, BYTE& yG, BYTE& yB )
{
   // recall read palette
   outportb ( 0x03C7, yCol );
   // read
   yR = inportb ( 0x03C9 );
   yG = inportb ( 0x03C9 );
   yB = inportb ( 0x03C9 );
}



/* Save palette in BYTE array of size 768. Use CGr:mWritePal
 to restoure a saved palette.
**********************************************************/
void CGr::mReadPal ( CMem& oMem )
{
   // alocate needed memory (768 bytes for 256 VGA palette)
   oMem.mCreate ( 768 );

   // seek all palette reading RGB
   for ( DWORD dwcCol = 0; dwcCol < 768; dwcCol += 3 )
      mRetPalColor ( dwcCol / 3, oMem.m_payMem [ dwcCol ],
                                 oMem.m_payMem [ dwcCol + 1 ],
                                 oMem.m_payMem [ dwcCol + 2 ] );
}



/* Write a memory array of RGB data in palette.
**********************************************************/
void CGr::mWritePal ( CMem& oMem )
{
   // seek all palette writing RGB
   for ( DWORD dwcCol = 0; dwcCol < 768; dwcCol += 3 )
      mSetPalColor ( dwcCol / 3, oMem.m_payMem [ dwcCol ],
                                 oMem.m_payMem [ dwcCol + 1 ],
                                 oMem.m_payMem [ dwcCol + 2 ] );
}



/* Wait video retrace.
**********************************************************/
void CGr::mWaitRetrace ()
{
   __asm__ __volatile__ ("

          movw  $0x03DA, %%dx
      L1: inb   %%dx,    %%al
          andb  $0x8,    %%al
          jnz   L1

      L2: inb   %%dx,    %%al
          andb  $0x8,    %%al
          jz    L2
      "
      :
      :
      : "%dx", "%al"
   );
}



/* Change video color. Do not perform wait retrace test.
**********************************************************/
void CGr::mSetVideoColor ( BYTE yCol )
{
   // define a 32 bits array with same color
   BYTE
      ayColors [ 4 ];
   ayColors [ 0 ] = ayColors [ 1 ] = ayColors [ 2 ] = ayColors [ 3 ] = yCol;

   for ( DWORD dwcTrans = 0; dwcTrans < VIDEOSIZE; dwcTrans += 4 )
      _dosmemputl ( (void*) ayColors, 1, 0xA0000 + dwcTrans );
}



/* Put one pixel on screen
**********************************************************/
void CGr::mSetPixel ( INT nX, INT nY, BYTE yCol )
{
   dosmemput ( (void*) &yCol, 1, 0xA0000 + nX + ( nY * RESX ) );
}



/* 1 - Show a frame buffer on screen in MCGA mode.
**********************************************************/
void CGr::mShow ( CMem& oMem )
{
   _movedatal ( _my_ds (), (PTRVAL) oMem.m_payMem, _dos_ds, 0xA0000, VIDEOSIZE4 );
}



/* 2 - Show a bitmap on screen. (dwOX, dwOY) are the
screen position for the upper left bitmap point (these values
start in upper left screen corner).
dwWidth and dwHeidht are the dimensions of the bitmap.
No test if bitmap is bigger that the screen in any axis (GPF).
**********************************************************/
void CGr::mShow ( INT nOX, INT nOY, INT nWidth, INT nHeight, CMem& oMem )
{
   INT
      nSizeLine4 = nWidth / 4;
   // copy each bitmap line
   for ( INT ncY = 0; ncY < nHeight; ncY ++ )
      _movedatal ( _my_ds (),
                   (PTRVAL) oMem.m_payMem + ncY * nWidth,
                   _dos_ds,
                   0xA0000 + ( ncY * RESX ) + nOX + ( nOY * RESX ),
                   nSizeLine4 );
}
