/* Implement keyboard functions to manage multiple keys presses without
repet delays. Keyboard handler is a function that is called each time an
interruption 9d is genereted by the press of one or more keys. To access
keys status you can't use getch, scanf,... you need the functions
below, see tips to know how use getch, scanf,... with keyboard handler.

FEATURES:
- substitute BIOS keyboard function
- don't use keyboard buffer

TIPS:
- to use getch (), scanf () and other functions desinstall the keyboard
  handler with CKeyboard::mDesinstallKeyboardHandler (). After reinstall it
  using CKeyboard::mInstallKeyboardHandler ()

CAUTION:
- all keys must be released before desinstall the handler, system will crach

BUGS:
- Print Screen and Control+Break are ignored while keyboard handler is on

TECHINFO:
- keyboard interruption: 9d
- port to read keys: 0x60
- port to signal interrupt captured: 0x20, value 0x20
- byte's values:
   If special key:
      give 224d from port 0x60 and
      next byte gived from port 0x60 is:
      If key pressed:
         key value        (array position: value + 128)
      If key released:
         key value + 128  (array position: value + 128)
   If not special key:
      If key pressed:
         key value        (array position: value)
      If key released:
         key value + 128  (array position: value)
*/



#include <dpmi.h>
#include <go32.h>
#include <string.h>
#include <dos.h>

#include "..\defs.h"

// Data below must be global to enable handler and local functions access.
BOOL ayKEYS [ 256 ];    // key table: TRUE: given key is pressed else FALSE
BOOL bSPECIAL = FALSE;  // special key pressed
BOOL bINSTALL = FALSE;  // Signals handler is installed, if TRUE it can be
                        // desintalled else no.
_go32_dpmi_seginfo stOLD_KEYBOARD_HANDLER;   // original keyboard handler
_go32_dpmi_seginfo stNEW_KEYBOARD_HANDLER;   // this class keyboard handler

#include "keyboard.h"



/* constructor - if is not installed keyboard handler install it
**********************************************************/
CKeyboard::CKeyboard ()
{
   // verify if need installation
   if ( ! bINSTALL )
      mInstallKeyboardHandler ();
}



/* destructor - desinstall the keyboard handler if:
 it is installed and only if this instance has installed it
**********************************************************/
CKeyboard::~CKeyboard ()
{
   // verify if must try desinstall handler
   if ( m_bInstallHandler )
      mDesinstallKeyboardHandler ();
}



/* Insplements the keyboard handler. Each time is generated an
interruption request this function is called.
SEE CKeyboard::mNewKeyboardHandler___ !!!!!
**********************************************************/
void CKeyboard::mNewKeyboardHandler ()
{
   // read key value
   BYTE
      yKey = inportb ( 0x60 );

   // verify if last call is not by an special key press
   if ( ! bSPECIAL )
      // detect special key press
      if ( yKey == 224 )
         bSPECIAL = TRUE;
      else
         // If not special key press, update table of key presses.
         // verify if key is beeing pressed (bits 7 == 0)
         if ( ( yKey & 128 ) == 0 )
            // if press set TRUE
            * ( ayKEYS + ( yKey & 127 ) ) = TRUE;
         else
            // if release (bit 7 == 1) set FALSE
            * ( ayKEYS + ( yKey & 127 ) ) = FALSE;
   else {
      // if special key had be pressed
      // verify if key is beeing pressed (bits 7 == 0)
      if ( ( yKey & 128 ) == 0 )
         * ( ayKEYS + 128 + ( yKey & 127 ) ) = TRUE;
      else
         // if release (bit 7 == 1) set FALSE
         * ( ayKEYS + 128 + ( yKey & 127 ) ) = FALSE;

      bSPECIAL = FALSE;
   }

   // signal insterruption end
   outportb ( 0x20, 0x20 );
}



/* Used only to know the size off code CKeyboard::mNewKeyboardHandler
ALWAYS MUST CAME AFTER CKeyboard::mNewKeyboardHandler.
This information is used in installation process.
**********************************************************/
void CKeyboard::mNewKeyboardHandler___ ()
{
}



/* Install keyboard handler if it's not installed
**********************************************************/
void CKeyboard::mInstallKeyboardHandler ()
{
   // verify if is not installed
   if ( ! bINSTALL ) {
      // save old keyboard handler
      _go32_dpmi_get_protected_mode_interrupt_vector ( 9, &stOLD_KEYBOARD_HANDLER );

      // save memory where new keyboard handler is from virtual memory swaps
      _go32_dpmi_lock_code ( (void*) mNewKeyboardHandler, (PTRVAL) mNewKeyboardHandler___ - (PTRVAL) mNewKeyboardHandler );

      // save memory where keys table is from virtual memory swaps
      _go32_dpmi_lock_data ( ayKEYS, sizeof ( ayKEYS ) );

      // substitue current keyboard handler to class keyboard handler
      stNEW_KEYBOARD_HANDLER.pm_offset   = (PTRVAL) mNewKeyboardHandler;
      stNEW_KEYBOARD_HANDLER.pm_selector = _go32_my_cs();
      _go32_dpmi_allocate_iret_wrapper ( &stNEW_KEYBOARD_HANDLER );
      _go32_dpmi_set_protected_mode_interrupt_vector ( 9, &stNEW_KEYBOARD_HANDLER );

      // reset key table (all FALSE)
      memset ( (void*) ayKEYS, FALSE, sizeof ( ayKEYS ) );

      // flag handler intalalled
      bINSTALL = m_bInstallHandler = TRUE;
   }
}



/* Desinstall keyboard handler if it's installed. Use
CKeyboard::mInstallKeyboardHandler to reinstall.
**********************************************************/
void CKeyboard::mDesinstallKeyboardHandler ()
{
   // verify if it's installed
   if ( bINSTALL ) {

      // wait all keys release (else you can have GPF)
      mWaitAllRelease ();

      // swap current handler with old
      _go32_dpmi_set_protected_mode_interrupt_vector ( 9, &stOLD_KEYBOARD_HANDLER );

      // free wrapper
      _go32_dpmi_free_iret_wrapper ( &stNEW_KEYBOARD_HANDLER );

      // flag desinstall
      bINSTALL = m_bInstallHandler = FALSE;
   }
}



/* Verify if a specific key is pressed now.
 return TRUE - yes; FALSE - no
**********************************************************/
BOOL CKeyboard::mTestKeyPress ( BYTE yKey )
{
   return * ( ayKEYS + yKey );
}



/* Return the code of the current key pressed.
If no key pressed return 0.
**********************************************************/
BYTE CKeyboard::mTestAllPress ()
{
   // see all keys's table
   for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
      // if find TRUE return it's code
      if ( mTestKeyPress ( ucT ) == TRUE )
         return ucT;

   return 0;
}



/* Wait the specific key be pressed. If bTestPress == TRUE then
only wait after the specific key be released.
Default bTestPress = TRUE
**********************************************************/
void CKeyboard::mWaitKeyPress ( BYTE yKey, BOOL bTestPress )
{
   // verify if must wait key realease
   if ( bTestPress )
      // while key pressed
      while ( mTestKeyPress ( yKey ) == TRUE );

   // while key released (wait press)
   while ( mTestKeyPress ( yKey ) == FALSE );
}



/* Wait some key be pressed. If bTestPress == TRUE then
only wait after all keys be released.
Default bTestPress = TRUE. return key code.
**********************************************************/
BYTE CKeyboard::mWaitAllPress ( BOOL bTestPress )
{
   // verify if must wait all keys be realeased
   if ( bTestPress ) {
      BOOL
         bRelease;
      // while one key be pressed
      do {
         bRelease = TRUE;
         // see all key's table
         for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
            // if find one pressed
            if ( mTestKeyPress ( ucT ) == TRUE )
               bRelease = FALSE;
      } while ( bRelease == FALSE );
   }

   // while any key not pressed
   while ( 1 )
      // see all key's table
      for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
         // if find one pressed return his code
         if ( mTestKeyPress ( ucT ) == TRUE )
            return ucT;
}



/* Wait all key be released
**********************************************************/
void CKeyboard::mWaitAllRelease ()
{
   BOOL
      bRelease;
   // while any key be pressed
   do {
      bRelease = TRUE;
      // see key's table
      for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
         // if find any key pressed
         if ( mTestKeyPress ( ucT ) == TRUE )
            bRelease = FALSE;
   } while ( bRelease == FALSE );
}
