/* Drawing functions for memory frame buffers. A frame buffer is an instance
of a CMem class and can be shown on screen by a CScreen instance.
Use CDraw to draw in frame buffer: polygons, pixels, ... or clear it.

FEATURES:
- draw polygons with one color (flat polygons) or with texture. The polygons
  can have any number of vertexes.
- is possible draw some types of polygons making clipping through a edge
  buffer. This edge buffer is used to define the left and right limits of X
  that can be drawn per scanline, providing the clipping. The buffer can
  be defined as the edges of a clipping polygon using the metod
  CDraw::mSetClipEdgeBuffer and can be reseted by CDraw::mResetClipEdgeBuffer
  The clip edge buffer must be an array of type tagInterval with Y resolution
  size. This is the way how portal clipping is implemented.
- draw shaded polygons.

TIPS:
- you can do "enviroment mapping" using a texture as the frame buffer.
- you can save an image in RAW file format saving the frame buffer to a file.
- texture width and height must not be equal.
- textures must be with size 256x256.
- the reference system is the same as the video:    (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX
- for more information about lighting see lighttab.cc

WARNING:
- polygons to be drawn must be convex
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "texture.h"

#include "lighttab.h"

#include "asm/draw.inc"

#include "draw.h"



// compute increase rate for x
#define INCREASEX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* constructor - init empty object
**********************************************************/
CDraw::CDraw ()
{
   m_nMidX = m_nMidY = m_nMaxX = m_nMaxY = m_nResX = m_nResY = m_nSize = 0;
   m_poFrBuff = NULL;
   m_pastForwardEdges = m_pastBackwardEdges = NULL;
   m_poLightTable = NULL;
}



/* constructor - init with an specific resolution.
**********************************************************/
CDraw::CDraw ( CLightTable& oLgT, INT nResX, INT nResY, CMem& oFrBuff )
{
   m_nMidX = m_nMidY = m_nMaxX = m_nMaxY = m_nResX = m_nResY = m_nSize = 0;
   m_poFrBuff = NULL;
   m_pastForwardEdges = m_pastBackwardEdges = NULL;
   m_poLightTable = NULL;

   mCreate ( oLgT, nResX, nResY, oFrBuff );
}



/* destructor - empty object
**********************************************************/
CDraw::~CDraw ()
{
   mDestroy ();
}



/* Set object member vriables. If frame buffer is
not bigger enought to this resolution realocate it.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CDraw::mCreate ( CLightTable& oLgT, INT nResX, INT nResY, CMem& oFrBuff )
{
   mDestroy ();

   m_nResX = nResX;
   m_nResY = nResY;
   m_poFrBuff = &oFrBuff;
   m_nSize = m_nResX * m_nResY;
   m_nMaxX = m_nResX - 1;
   m_nMaxY = m_nResY - 1;
   m_nMidX = m_nResX >> 1;  //    >> 1 is equal / 2
   m_nMidY = m_nResY >> 1;
   m_poLightTable = &oLgT;

   // try alocate memory for edge buffer for forward scan conversion
   if ( ( m_pastForwardEdges = new tagPolyEdge [ m_nResY ] ) == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // try alocate memory for edge buffer for backward scan conversion
   if ( ( m_pastBackwardEdges = new tagPolyEdge [ m_nResY ] ) == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   return TRUE;
}



/* Empty member variables (memory desalocation)
**********************************************************/
void CDraw::mDestroy ()
{
   // verify if object have some value in variables
   if ( m_poFrBuff != NULL ) {
      m_nMidX = m_nMidY = m_nResX = m_nResY = m_nSize = 0;
      m_poFrBuff = NULL;
      m_poLightTable = NULL;
   }

   // verify if object have some value in variable
   if ( m_poLightTable != NULL )
      m_poLightTable = NULL;

   // verify if edge buffer for forward scan is alocated, if yes desalocate
   if ( m_pastForwardEdges != NULL ) {
      delete m_pastForwardEdges;
      m_pastForwardEdges = NULL;
   }

   // verify if edge buffer for backward scan is alocated, if yes desalocate
   if ( m_pastBackwardEdges != NULL ) {
      delete m_pastBackwardEdges;
      m_pastBackwardEdges = NULL;
   }
}



///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// SIMPLE DRAW PRIMITIVES: LINES, PIXELS, FRAME BUFFER,...
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////



/* Set all frame buffer with a specific color.
**********************************************************/
void CDraw::mClear ( BYTE yCol )
{
   memset ( (void*) *m_poFrBuff, yCol, m_nSize ); // (void*) *m_poFrBuff is
                                         // one way to access memory buffer.
}



/* Put one pixel on frame buffer.
**********************************************************/
void CDraw::mSetPixel ( INT nX, INT nY, BYTE yCol )
{
   // do clip
   if ( nX >= 0 && nX < m_nResX && nY >= 0 && nY < m_nResY )
      memset ( (void*) m_poFrBuff->m_payMem + nX + ( nY * m_nResX ), yCol, 1 );
}



/* Copy an other frame buffer over the current frame buffer.
oFrBuffSource is the source frame buffer and must have
the same size as the current frame buffer.
**********************************************************/
void CDraw::mCopy ( CMem& oFrBuffSource )
{
   memcpy ( (void*) *m_poFrBuff, (void*) oFrBuffSource, m_poFrBuff->m_dwSize );
}



/* 2 - Copy an other frame buffer (or bitmap) over the
current frame buffer. (nOX, nOY) are the frame buffer
position for the upper left bitmap point (these values
start in upper left frame buffer corner).
nWidth and nHeidht are the dimensions of the bitmap.
No test if bitmap is bigger that the frame buffer in any axis (GPF)!
**********************************************************/
void CDraw::mCopy ( INT nOX, INT nOY, INT nWidth, INT nHeight, CMem& oFrBuffSource )
{
   // copy each bitmap line, no clip!
   for ( INT ncY = 0; ncY < nHeight; ncY ++ )

      memcpy ( (void*) m_poFrBuff->m_payMem + ( ncY * m_nResX ) + ( nOY * m_nResX ) + nOX,
               (void*) oFrBuffSource.m_payMem + ncY * nWidth,
               nWidth );
}



/* Draw a line with the specific color. Low optimized code.
Give the points (nXI, nYI) -> (nXF,nYF). Do line clip.
**********************************************************/
void CDraw::mLine ( INT nXI, INT nYI, INT nXF, INT nYF, BYTE yCol )
{
   // define start and end Y (the less and greater Y)
   INT
      nYStart = nYI,
      nYEnd = nYF,
      nSignalY = 1;
   if ( nYF < nYStart ) {
      nYStart = nYF;
      nYEnd = nYI;
      nSignalY = -1;
   }

   // define start and end X (the less and greater X)
   INT
      nXStart = nXI,
      nXEnd = nXF,
      nSignalX = 1;
   if ( nXF < nXStart ) {
      nXStart = nXF;
      nXEnd = nXI;
      nSignalX = -1;
   }

   // compute delta between the given points
   INT
      nDeltaX = nXF - nXI,
      nDeltaY = nYF - nYI;

   // if delta X is the greater you draw the line using Y as function of X
   if ( abs ( nDeltaX ) > abs ( nDeltaY ) ) {

      float
         fYRate = (float) nDeltaY / nDeltaX, // define Y rate
         fY;                                            // start Y
      // define start line
      if ( nSignalY == -1 )
         fY = (float) nYEnd;
      else
         fY = (float) nYStart;

      // draw line
      // test if X is growing left (1) or right (-1)
      if ( nSignalX == 1 )
         for ( INT nX = nXStart; nX < nXEnd + 1; nX ++ ) {

            // do clip
            if ( nX >= 0 && nX <= m_nMaxX && (INT) fY >= 0 && (INT) fY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + nX + ( (INT) fY * m_nResX ), yCol, 1 );

            fY += fYRate;
         }
      else {

         fYRate = - fYRate;

         for ( INT nX = nXEnd; nX > nXStart; nX -- ) {

            // do clip
            if ( nX >= 0 && nX <= m_nMaxX && (INT) fY >= 0 && (INT) fY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + nX + ( (INT) fY * m_nResX ), yCol, 1 );

            fY += fYRate;
         }
      }
   }
   else {   // draw the line using X as function of Y

      float
         fXRate = (float) nDeltaX / nDeltaY, // define X rate
         fX;                                            // start X
      // define start colum
      if ( nSignalX == -1 )
         fX = (float) nXEnd;
      else
         fX = (float) nXStart;

      // draw line
      // test if Y is growing down (1) or up (-1)
      if ( nSignalY == 1 )
         for ( INT nY = nYStart; nY < nYEnd + 1; nY ++ ) {

            // do clip
            if ( (INT) fX >= 0 && (INT) fX <= m_nMaxX && nY >= 0 && nY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + (INT) fX + ( nY * m_nResX ), yCol, 1 );

            fX += fXRate;
         }
      else {

         fXRate = - fXRate;

         for ( INT nY = nYEnd; nY > nYStart; nY -- ) {

            // do clip
            if ( (INT) fX >= 0 && (INT) fX <= m_nMaxX && nY >= 0 && nY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + (INT) fX + ( nY * m_nResX ), yCol, 1 );

            fX += fXRate;
         }
      }
   }
}



///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// PORTAL MAKE PRIMITIVES
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////



/* Reset the given edge buffer (pastEdgeBuffer): set the
start limit to the minimum x and the end limit to maximum
x acording the frame buffer. Do not perform clipping.
**********************************************************/
void CDraw::mResetClipEdgeBuffer ( tagInterval* pastEdgeBuffer )
{
   tagInterval
      *pstInt = pastEdgeBuffer;

   // scan all edge buffer setting their values
   for ( INT ncYScanLine = 0; ncYScanLine < m_nResY; ncYScanLine ++ ) {
      pstInt->nStart = 0;
      pstInt->nEnd = m_nMaxX;
      pstInt ++;
   }
}



/* Set the given clip buffer using the edges defined by the
given polygon.
The given polygon is clipped through the given edge buffer.
pastXYZes have the array of polygon vertexes in clockwise
or anticlockwise order used to define the edge buffer.
nNumVertexes is the number of polygon vertexes.
pastEdgeBuffer is the clip edge buffer, must have the same size
stClipBox is the bounding box around the clip polygon that can be
used in fast poly clip, enter the current clip box and return the new.
as the vertical resolution (Y).
return TRUE - poly drawn (not totaly clipped) / FALSE - poly
totaly invisible (completely clipped)
**********************************************************/
BOOL CDraw::mSetClipEdgeBuffer ( tagXY* pastXYZes, INT nNumVertexes,
                                 tagInterval* pastEdgeBuffer,
                                 tagRectangle& stClipBox )
{
   // flag if poly has some part (or completely) visible
   BOOL
      bVisible = FALSE;        // originaly not visible
   
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastXYZes [ 0 ].nY,   // y of the top vertex
      nYMax = pastXYZes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastXYZes [ ncVert ].nY > nYMax ) {
         nYMax = pastXYZes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastXYZes [ ncVert ].nY < nYMin ) {
         nYMin = pastXYZes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort edge fill
   if ( nYMax == nYMin ) {
      stClipBox.nStartY = nYMax;
      stClipBox.nEndY = nYMax;
      return bVisible;
   }

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_pastForwardEdges, pastXYZes [ nPos1 ], pastXYZes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_pastBackwardEdges, pastXYZes [ nPos1 ], pastXYZes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // Update edge buffer using 3 loops: the first set the area
   // between clip buffer start and clipping polygon begin. The second
   // set the area over the clipping polygon. The third set the area
   // between end clipping polygon and end buffer.
   // define start buffer position (Y) changed by clipping polygon
   INT
      nY = nYMin;
   if ( nY < 0 )
      nY = 0;     // start buffer line changed by polygon edges
   // define bottom buffer position (Y) changed by clipping polygon
   INT
      nYBottom = nYMax;
   if ( nYBottom > m_nMaxY )
      nYBottom = m_nMaxY;
   // start edge buffer position
   tagInterval
      *pstInt = pastEdgeBuffer,
      *pstEnd = &pastEdgeBuffer [ m_nMaxY ];

   // verify if can change the Y up clip box limit
   if ( nY > stClipBox.nStartY )
      stClipBox.nStartY = nY;
   // verify if can change the Y bottom clip box limit
   if ( nYBottom < stClipBox.nEndY )
      stClipBox.nEndY = nYBottom;

   // store the less and greatter X for this clip edge buffer
   tagInterval
      stXLim;
   // REVIEW
   //stXLim.nStart = m_nMaxX;
   //stXLim.nEnd = 0;

   // scan edge buffer begin to start polygon edges
   if ( nY < m_nResY )
      for ( INT ncY = 0; ncY < nY; ncY ++ ) {
         // set edge buffer: no space available for drawing
         if ( pstInt <= pstEnd ) {
            pstInt->nStart = m_nMaxX;
            pstInt->nEnd = 0;
            if ( pstInt < pstEnd ) 
               pstInt ++;
         }
      }

   // scan edge buffer from top polygon to bottom polygon
   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )
      while ( nY <= nYBottom ) {

         if ( nY < m_nResY ) {

            // make edge buffer X clip and define new buffer X limits
            // clip start X
            if ( m_pastBackwardEdges [ nY ].nX > pstInt->nStart ) {
               if ( pstInt <= pstEnd )
                  if ( m_pastBackwardEdges [ nY ].nX < 0 )
                     pstInt->nStart = 0;
                  else
                     pstInt->nStart = m_pastBackwardEdges [ nY ].nX;

               // search the minimum X in clip edge buffer
               if ( pstInt->nStart < stXLim.nStart )
                  stXLim.nStart = pstInt->nStart;
            }

            // clip end X
            if ( m_pastForwardEdges [ nY ].nX < pstInt->nEnd ) {
               if ( pstInt <= pstEnd ) 
                  if ( m_pastForwardEdges [ nY ].nX > m_nMaxX )
                     pstInt->nEnd = m_nMaxX;
                  else
                     pstInt->nEnd = m_pastForwardEdges [ nY ].nX;
   
               // search the maximum X in clip edge buffer
               if ( pstInt->nEnd > stXLim.nEnd )
                  stXLim.nEnd = pstInt->nEnd;
            }

            // if line visible
            if ( pstInt->nStart >= stClipBox.nStartX && pstInt->nEnd <= stClipBox.nEndX )
               bVisible = TRUE;   // poly is visible (partialy or completely)

            if ( pstInt < pstEnd ) 
               pstInt ++;
         }
         
         // goto next scanline
         nY ++;
      }
   else 
      while ( nY <= nYBottom ) {

         if ( nY < m_nResY ) {

            // make edge buffer X clip and define new buffer X limits
            // clip start X
            if ( m_pastForwardEdges [ nY ].nX > pstInt->nStart ) {
               if ( pstInt <= pstEnd ) 
                  if ( m_pastForwardEdges [ nY ].nX < 0 )
                     pstInt->nStart = 0;
                  else
                     pstInt->nStart = m_pastForwardEdges [ nY ].nX;

               // search the minimum X in clip edge buffer
               if ( pstInt->nStart < stXLim.nStart )
                  stXLim.nStart = pstInt->nStart;
            }

            // clip end X
            if ( m_pastBackwardEdges [ nY ].nX < pstInt->nEnd ) {
               if ( pstInt <= pstEnd )
                  if ( m_pastBackwardEdges [ nY ].nX > m_nMaxX )
                     pstInt->nEnd = m_nMaxX;
                  else
                     pstInt->nEnd = m_pastBackwardEdges [ nY ].nX;

               // search the maximum X in clip edge buffer
               if ( pstInt->nEnd > stXLim.nEnd )
                  stXLim.nEnd = pstInt->nEnd;
            }

            // if line visible
            if ( pstInt->nStart >= stClipBox.nStartX && pstInt->nEnd <= stClipBox.nEndX )
               bVisible = TRUE;   // poly is visible (partialy or completely)

            if ( pstInt < pstEnd ) 
               pstInt ++;
         }
         
         // goto next scanline
         nY ++;
      }

   // scan end clip polygon to buffer end
   for ( INT ncY = nYBottom + 1; ncY < m_nResY; ncY ++ ) {
      if ( nY < m_nResY ) {
         // set edge buffer: no space available for drawing
         if ( pstInt <= pstEnd ) {
            pstInt->nStart = m_nMaxX;
            pstInt->nEnd = 0;
         }
         if ( pstInt < pstEnd ) 
            pstInt ++;
      }
   }

   // REVIEW
   // verify if have updated X start and end, if no use original values
   //if ( bVisible ) {  // poly visible, must have good values for X limits
      //if ( stXLim.nStart == stClipBox.nEndX )
         //stXLim.nStart = stClipBox.nStartX;
      //if ( stXLim.nEnd == stClipBox.nStartX )
         //stXLim.nEnd = stClipBox.nEndX;
   //}

   // verify if can update clip box X limits
   /*if ( stClipBox.nStartX < stXLim.nStart )
      stClipBox.nStartX = stXLim.nStart;
   if ( stClipBox.nEndX > stXLim.nEnd )
      stClipBox.nEndX = stXLim.nEnd;*/

/*for (int n=0;n<320;n++){
   if(pastEdgeBuffer[n].nStart<0)
      ABORT ( "baixo", __FILE__, __LINE__ )
   if(pastEdgeBuffer[n].nEnd>320)
      ABORT ( "alto", __FILE__, __LINE__ )
}  */

   return bVisible;
}



///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// POLYGON DRAW PRIMITIVES: FLAT/GOURAUD SHADED
// POLYS, TEXTURED FLAT/GOURAUD SHADED POLYS,...
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////



/* Draw a flat polygon with only 4 vertex in frame buffer.
Medium optimized. The vertexes must be in clockwise or
anticlockwise. yCol is the polygon fill color.
This function use in implementation fix point 16.16.
Perform clip.

                        (x4, y4)
                        *                -
                       / \                 loop
                      /   \                    |
            (x1, y1) *     \                   |
                      \     * (x3, y3)         | compute x positions
                       \   /                   | between edges
                        \ /                    V
                         *               -
                         (x2, y2)

Line equation:
Y = aX + b, where
            a is a increase rate
            b is the offset of the line in y axis
Equation used in intersections (how x change for each scanline + x vertex)
X = ( deltaY * increaseX ) + Xvertex, where
            increaseX is 1 / a, how x change per scanline
            deltaY is the distance between a Y vertex and Y loop
            Xvertex is the X in vertex used in computation of deltaY
**********************************************************/
void CDraw::mFlatPolygon4 ( INT nX1, INT nY1, INT nX2, INT nY2,
                            INT nX3, INT nY3, INT nX4, INT nY4, BYTE yCol )
{
   // clip y edges
   // define less (top) and greater (bottom) y values
   INT
      nYSup = nY1,
      nYInf = nY1;
   // test y2
   if ( nYSup > nY2 )
      nYSup = nY2;
   else
      if ( nYInf < nY2 )
         nYInf = nY2;
   // test y3
   if ( nYSup > nY3 )
      nYSup = nY3;
   else
      if ( nYInf < nY3 )
         nYInf = nY3;
   // test y4
   if ( nYSup > nY4 )
      nYSup = nY4;
   else
      if ( nYInf < nY4 )
         nYInf = nY4;
   // test if points are outside frame buffer limits, if yes change values
   if ( nYSup < 0 )
      nYSup = 0;
   if ( nYInf > m_nMaxY )
      nYInf = m_nMaxY;

   // define if be possible compute x increase rates, if yes do
   INT
      nIncX12 = 0,
      nIncX14 = 0,
      nIncX23 = 0,
      nIncX34 = 0;
   BOOL
      bIncValid12 = TRUE,
      bIncValid14 = TRUE,
      bIncValid23 = TRUE,
      bIncValid34 = TRUE;
   // be computed valores only for points that are not in same lines
   if ( nY1 != nY2 )
      nIncX12 = (INT) ( INCREASEX ( nX1, nY1, nX2, nY2 ) * 65536 );
   else
      bIncValid12 = FALSE;
   if ( nY1 != nY4 )
      nIncX14 = (INT) ( INCREASEX ( nX1, nY1, nX4, nY4 ) * 65536 );
   else
      bIncValid14 = FALSE;
   if ( nY2 != nY3 )
      nIncX23 = (INT) ( INCREASEX ( nX2, nY2, nX3, nY3 ) * 65536 );
   else
      bIncValid23 = FALSE;
   if ( nY3 != nY4 )
      nIncX34 = (INT) ( INCREASEX ( nX3, nY3, nX4, nY4 ) * 65536 );
   else
      bIncValid34 = FALSE;

   // define address of the beging for the first line in frame buffer
   UINT
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + m_nResX * nYSup;

   // define color fill
   UINT
      uCol = yCol | ( yCol << 8 ) | ( yCol << 16 ) | ( yCol << 24 );

   // goto polygon start to end in Y axis
   INT
      nXIni,    // start x in scanline
      nXEnd,    // end x in scanline
      nX;       // x value of intersection by 2 lines
   for ( INT ncY = nYSup; ncY <= nYInf; ncY ++ ) {

      // initial x values in scanline
      nXIni = 32767;
      nXEnd = - 32768;

      // compute x between vertex 1 and 4
      if ( bIncValid14 &&
           ( ncY >= nY1 && ncY <= nY4 ) ||
           ( ncY <= nY1 && ncY >= nY4 ) ) {

         nX = (INT) ( ( ncY - nY4 ) * nIncX14 >> 16 ) + nX4;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 1 and 2
      if ( bIncValid12 &&
           ( ncY >= nY1 && ncY <= nY2 ) ||
           ( ncY <= nY1 && ncY >= nY2 ) ) {

         nX = (INT) ( ( ncY - nY2 ) * nIncX12 >> 16 ) + nX2;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 2 and 3
      if ( bIncValid23 &&
           ( ncY >= nY2 && ncY <= nY3 ) ||
           ( ncY <= nY2 && ncY >= nY3 ) ){

         nX = (INT) ( ( ncY - nY3 ) * nIncX23 >> 16 ) + nX3;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 3 and 4
      if ( bIncValid34 &&
           ( ncY >= nY3 && ncY <= nY4 ) ||
           ( ncY <= nY3 && ncY >= nY4 ) ) {

         nX = (INT) ( ( ncY - nY4 ) * nIncX34 >> 16 ) + nX4;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // clip x values
      // verify start
      if ( nXIni < 0 )
         nXIni = 0;
      else
         // start can be in right (line outside frame buffer)
         if ( nXIni > m_nMaxX ) {
            // compute next scanline address
            pLineAddress += m_nResX;
            // do not draw line
            continue;
         }
      // verify end
      if ( nXEnd > m_nMaxX )
         nXEnd = m_nMaxX;
      else
         // end can be in left (line outside frame buffer)
         if ( nXEnd < 0 ) {
            // compute next scanline address
            pLineAddress += m_nResX;
            // do not draw line
            continue;
         }

      // draw horizontal line
      // if compile with asm code
      #ifdef COMP_ASM
      ASM_HLINE_FLAT ( pLineAddress + nXIni, nXEnd - nXIni + 1, uCol );
      #endif
      // compile C only (no asm code)
      #ifndef COMP_ASM
      memset ( (void*) ( pLineAddress + nXIni ), uCol, nXEnd - nXIni + 1 );
      #endif

      // compute next scanline address
      pLineAddress += m_nResX;
   }
}



/* Draw a flat polygon with any number of vertexes (points).
pastXYZes have the array of polygon vertexes in
clockwise or anticlockwise order. nNumVertexes is the
number of vertexes in array. yCol is the polygon fill color.
Perform frame buffer clip. Medium optimized.
**********************************************************/
void CDraw::mFlatPolygon ( tagXY* pastXYZes,
                           INT nNumVertexes, BYTE yCol )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastXYZes [ 0 ].nY,   // y of the top vertex
      nYMax = pastXYZes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastXYZes [ ncVert ].nY > nYMax ) {
         nYMax = pastXYZes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastXYZes [ ncVert ].nY < nYMin ) {
         nYMin = pastXYZes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // define color fill
   UINT
      uCol = yCol | ( yCol << 8 ) | ( yCol << 16 ) | ( yCol << 24 );

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_pastForwardEdges, pastXYZes [ nPos1 ], pastXYZes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_pastBackwardEdges, pastXYZes [ nPos1 ], pastXYZes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = nYMin < 0 ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )

      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalLine ( m_pastBackwardEdges [ nY ],
                           m_pastForwardEdges [ nY ],
                           uCol, pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalLine ( m_pastForwardEdges [ nY ],
                           m_pastBackwardEdges [ nY ],
                           uCol, pLineAddress );
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
}



/* Draw a inpolated light shaded polygon (gouraud like) with any
number of vertexes (points).
pastXYZes have the array of polygon vertexes with it's light
  intensitys (levels), vertexes in clockwise or anticlockwise order.
payColors have the array of vertexes colors (palette indexes).
Vertexes in clockwise or anticlockwise order.
nNumVertexes is the number of vertexes in array.
Perform frame buffer clip.
**********************************************************/
void CDraw::mIntLgtPolygon ( tagXY* pastXYZes,
                              BYTE* payColors, INT nNumVertexes )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastXYZes [ 0 ].nY,   // y of the top vertex
      nYMax = pastXYZes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastXYZes [ ncVert ].nY > nYMax ) {
         nYMax = pastXYZes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastXYZes [ ncVert ].nY < nYMin ) {
         nYMin = pastXYZes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanIntLgtEdge ( m_pastForwardEdges,
                         pastXYZes [ nPos1 ], pastXYZes [ nPos2 ],
                         payColors [ nPos1 ], payColors [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanIntLgtEdge ( m_pastBackwardEdges,
                         pastXYZes [ nPos1 ], pastXYZes [ nPos2 ],
                         payColors [ nPos1 ], payColors [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = nYMin < 0 ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )

      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalIntLgtLine ( m_pastBackwardEdges [ nY ],
                                  m_pastForwardEdges [ nY ],
                                  pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
   
         mHorizontalIntLgtLine ( m_pastForwardEdges [ nY ],
                                  m_pastBackwardEdges [ nY ],
                                  pLineAddress );
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
}



/* Draw a affine texture mapped polygon on frame buffer.
The poly is flat shaded.
Make clip through the given edge buffer (portal).
pastPolVertexes is the array of polygon vertexes;
pastTexVertexes is the array of texture vertexes;
One vertex in polygon has one vertex in texture.
All vertexes must be in clockwise or anticlockwise order.
nNumVertex is the number of polygon vertexes.
oTx is the texture to use in polygon.
fNormLightLevel has the normalized polygon light level (between 0 and 1)
pastEdgeBuffer is the clip buffer to be used (in portal clip),
 if set to NULL only frame buffer clip is made (default NULL).
**********************************************************/
void CDraw::mTexturedPolygon ( tagXY* pastPolVertexes,
                               tagXY* pastTexVertexes,
                               INT nNumVertexes, CTexture& oTx,
                               tagInterval* pastEdgeBuffer,
                               float fNormLightLevel )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastPolVertexes [ 0 ].nY,   // y of the top vertex
      nYMax = pastPolVertexes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastPolVertexes [ ncVert ].nY > nYMax ) {
         nYMax = pastPolVertexes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastPolVertexes [ ncVert ].nY < nYMin ) {
         nYMin = pastPolVertexes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon and texture edges, the X for each scanline defined
      // by the line vertex 1 ----> vertex 2
      mScanTexturedEdge ( m_pastForwardEdges,
                          pastPolVertexes [ nPos1 ],
                          pastPolVertexes [ nPos2 ],
                          pastTexVertexes [ nPos1 ],
                          pastTexVertexes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon and texture edges, the X for each scanline defined
      // by the line vertex 1 ----> vertex 2
      mScanTexturedEdge ( m_pastBackwardEdges,
                          pastPolVertexes [ nPos1 ],
                          pastPolVertexes [ nPos2 ],
                          pastTexVertexes [ nPos1 ],
                          pastTexVertexes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // define the light level for the loaded light table
   // transform normalized light level to light table value
   DWORD
      dwLightLevel = (DWORD) ( fNormLightLevel * ( m_poLightTable->m_dwNumLevels - 1 ) );

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = ( nYMin < 0 ) ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // verify if clipping edge buffer exist, if yes use else use default
   tagInterval
      stDefaultClip;
   BYTE
      yNextLine = 1; // to go to next clip buffer element when scanning lines
   // if don't has a clip buffer (no portal clipping)
   if ( pastEdgeBuffer == NULL ) {
      // define the default clip
      stDefaultClip.nStart = 0;
      stDefaultClip.nEnd = m_nMaxX;
      pastEdgeBuffer = &stDefaultClip;
      // no line change
      yNextLine = 0;
   }
   else
      // find the right position to start buffer scan
      pastEdgeBuffer += nY;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalTexturedLine ( m_pastBackwardEdges [ nY ],
                                   m_pastForwardEdges [ nY ],
                                   *pastEdgeBuffer,
                                   (BYTE*) oTx, pLineAddress,
                                   dwLightLevel );
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalTexturedLine ( m_pastForwardEdges [ nY ],
                                   m_pastBackwardEdges [ nY ],
                                   *pastEdgeBuffer,
                                   (BYTE*) oTx, pLineAddress,
                                   dwLightLevel );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
}



/* Draw a affine texture mapped polygon with interpolated
shading at the vertexes (like Gouraud).
on frame buffer.
Make clip through the given edge buffer (portal).
pastPolVertexes is the array of polygon vertexes;
pastTexVertexes is the array of texture vertexes;
pafNormLights is the array of vertexes light levels (between 0 and 1);
One vertex in polygon has one vertex in texture and light levels array.
All vertexes must be in clockwise or anticlockwise order.
nNumVertex is the number of polygon vertexes.
oTx is the texture to use in polygon.
fNormLightLevel has the normalized polygon light level (between 0 and 1)
pastEdgeBuffer is the clip buffer to be used (in portal clip),
 if set to NULL only frame buffer clip is made (default NULL).
**********************************************************/
void CDraw::mIntLgtTexturedPolygon ( tagXY* pastPolVertexes,
                                      tagXY* pastTexVertexes,
                                      float* pafNormLights,
                                      INT nNumVertexes, CTexture& oTx,
                                      tagInterval* pastEdgeBuffer = NULL )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastPolVertexes [ 0 ].nY,   // y of the top vertex
      nYMax = pastPolVertexes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastPolVertexes [ ncVert ].nY > nYMax ) {
         nYMax = pastPolVertexes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastPolVertexes [ ncVert ].nY < nYMin ) {
         nYMin = pastPolVertexes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon, texture edges and lighting, the X for each scanline
      // defined by the line vertex 1 ----> vertex 2
      mScanIntLgtTexturedEdge ( m_pastForwardEdges,
                                 pastPolVertexes [ nPos1 ],
                                 pastPolVertexes [ nPos2 ],
                                 pastTexVertexes [ nPos1 ],
                                 pastTexVertexes [ nPos2 ],
                                 pafNormLights   [ nPos1 ],
                                 pafNormLights   [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon, texture edges and lighting, the X for each scanline
      // defined by the line vertex 1 ----> vertex 2
      mScanIntLgtTexturedEdge ( m_pastBackwardEdges,
                                 pastPolVertexes [ nPos1 ],
                                 pastPolVertexes [ nPos2 ],
                                 pastTexVertexes [ nPos1 ],
                                 pastTexVertexes [ nPos2 ],
                                 pafNormLights   [ nPos1 ],
                                 pafNormLights   [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = ( nYMin < 0 ) ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // verify if clipping edge buffer exist, if yes use else use default
   tagInterval
      stDefaultClip;
   BYTE
      yNextLine = 1; // to go to next clip buffer element when scanning lines
   // if don't has a clip buffer (no portal clipping)
   if ( pastEdgeBuffer == NULL ) {
      // define the default clip
      stDefaultClip.nStart = 0;
      stDefaultClip.nEnd = m_nMaxX;
      pastEdgeBuffer = &stDefaultClip;
      // no line change
      yNextLine = 0;
   }
   else
      // find the right position to start buffer scan
      pastEdgeBuffer += nY;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalIntLgtTexturedLine ( m_pastBackwardEdges [ nY ],
                                          m_pastForwardEdges [ nY ],
                                          *pastEdgeBuffer,
                                          (BYTE*) oTx, pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalIntLgtTexturedLine ( m_pastForwardEdges [ nY ],
                                          m_pastBackwardEdges [ nY ],
                                          *pastEdgeBuffer,
                                          (BYTE*) oTx, pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
}



///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// POLYGON EDGES SCAN PRIMITIVES
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////



/* Used by CDrawClip::mSetClipEdgeBuffer to define X limits
for clip edge buffer and by CDraw::mFlatPolygon and to define
the X coordenates in each scanline for the polygon edges.
pastEdge - fill the x coordenates for each scanline;
stV1 and stV2 - vertex between scan must be done.
Always stV1 is less (topper) than stV2 in Y axis.
**********************************************************/
void CDraw::mScanFlatEdge ( tagPolyEdge* pastEdge,
                            tagXY& stV1, tagXY& stV2 )
{
   // verify if Y are in same line, if yes exit
   if ( stV1.nY == stV2.nY )
      return;

   INT
      // define where X start in first scanline, use fix point 16.16
      nX = stV1.nX << 16,
      // X increase rate
      nXInc = ( ( stV2.nX - stV1.nX ) << 16 ) /
                ( stV2.nY - stV1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stV1.nY < 0 ) {
      nYTop = 0;
      // if top Y has been cliped then is necessary update initial X value
      nX -= stV1.nY * nXInc;
   }
   else
      nYTop = stV1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stV2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stV2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test
   do {

      pastEdge [ nYScan ].nX = nX >> 16; // X in frame buff, convert fix->int
      nX += nXInc;                       // new X in next scanline
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



/* Used by CDraw::mIntLgtPolygon:
pastEdge - fill the x coordenates for each scanline 
and the light for each scanline.
stV1 and stV2 - vertex between scan must be done.
yColor1 and yColor2 - the vertexes colors (palette indexes)
Always stV1 is less (topper) than stV2 in Y axis.
**********************************************************/
void CDraw::mScanIntLgtEdge ( tagPolyEdge* pastEdge,
                               tagXY& stV1, tagXY& stV2,
                               BYTE yColor1, BYTE yColor2 )
{
   // verify if Y are in same line, if yes exit
   if ( stV1.nY == stV2.nY )
      return;

   INT
      // define where X start in first scanline, use fix point 16.16
      nX = stV1.nX << 16,
      // X increase rate
      nXInc = ( ( stV2.nX - stV1.nX ) << 16 ) /
                ( stV2.nY - stV1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stV1.nY < 0 ) {
      nYTop = 0;
      // if top Y has been cliped then is necessary update initial X value
      nX -= stV1.nY * nXInc;
   }
   else
      nYTop = stV1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stV2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stV2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define the color parameters
   INT
      nColor = ( (INT) yColor1 ) << 16,                        // initial color
      nColorInc = ( ( (INT) yColor2 - yColor1 ) << 16 ) /      // increase rate
                    ( stV2.nY - stV1.nY );

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test
   do {

      pastEdge [ nYScan ].nX = nX >> 16; // X in frame buff, convert fix->int
      pastEdge [ nYScan ].nColor = nColor;   // color
      nX += nXInc;                       // new X in next scanline
      nColor += nColorInc;               // new color in next scanline
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}
   


/* Used by CDraw::mTexturedPolygon to define the X coordenates
in each scanline for the polygon and texture edges.
pastEdge - the x coordenates for each scanline in polygon and
           the x and y coordenates for each scanline in texture;
stP1 and stP2 - polygon vertex between scan must be done.
stT1 and stT2 - texture vertex between scan must be done.
Always stP1 is less (topper) than stP2 in Y axis.
Use fix point 16.16.
**********************************************************/
void CDraw::mScanTexturedEdge ( tagPolyEdge* pastEdge,
                                tagXY& stP1, tagXY& stP2,
                                tagXY& stT1, tagXY& stT2 )
{
   // verify if Y are in same line, if yes exit
   if ( stP1.nY == stP2.nY )
      return;

   INT
      // define where X start in first scanline
      nX = stP1.nX << 16,
      // define where texture X and Y start in first scanline
      nTexX = stT1.nX << 16,
      nTexY = stT1.nY << 16,
      // define X increase rate for each scanline
      nXInc = ( ( stP2.nX - stP1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      // define texture X and Y increase rates for each scan line
      nTexXInc = ( ( stT2.nX - stT1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      nTexYInc = ( ( stT2.nY - stT1.nY ) << 16 ) / ( stP2.nY - stP1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stP1.nY < 0 ) {
      nYTop = 0;
      // top Y has been cliped then is necessary update values
      nX    -= stP1.nY * nXInc;
      nTexX -= stP1.nY * nTexXInc;
      nTexY -= stP1.nY * nTexYInc;
   }
   else
      nYTop = stP1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stP2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stP2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test

   do {

      pastEdge [ nYScan ].nX = nX >> 16;  // X in frame buff, convert fix->int
      pastEdge [ nYScan ].stTex.nX = nTexX; // X in texture
      pastEdge [ nYScan ].stTex.nY = nTexY; // Y in texture
      nX += nXInc;                       // new X in next scanline
      nTexX += nTexXInc;                 // new texture X
      nTexY += nTexYInc;                 // new texture Y
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}
   


/* Used by CDraw::mIntLgtTexturedPolygon to define the X coordenates
in each scanline for the polygon, texture and light levels edges
pastEdge - the x coordenates for each scanline in polygon and
           the x and y coordenates for each scanline in texture;
stP1 and stP2 - polygon vertex between scan must be done.
stT1 and stT2 - texture vertex between scan must be done.
fL1 and fL2   - normalized (between 0 and 1) light level.
Always stP1 is less (topper) than stP2 in Y axis.
Use fix point 16.16.
**********************************************************/
void CDraw::mScanIntLgtTexturedEdge ( tagPolyEdge* pastEdge,
                                       tagXY& stP1,  tagXY& stP2,
                                       tagXY& stT1,  tagXY& stT2,
                                       float& fL1,   float& fL2 )
{
   // verify if Y are in same line, if yes exit
   if ( stP1.nY == stP2.nY )
      return;

   // convert normalized light level to the light level for current light table
   INT
      nL1 = (INT) ( fL1 * ( m_poLightTable->m_dwNumLevels - 1 ) ),
      nL2 = (INT) ( fL2 * ( m_poLightTable->m_dwNumLevels - 1 ) );

   INT
      // define where X start in first scanline
      nX = stP1.nX << 16,
      // define where texture X and Y start in first scanline
      nTexX = stT1.nX << 16,
      nTexY = stT1.nY << 16,
      // light level for the loaded light table (not normalized)
      nLightLevel = nL1 << 16,
      // define X increase rate for each scanline
      nXInc = ( ( stP2.nX - stP1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      // define texture X and Y increase rates for each scan line
      nTexXInc = ( ( stT2.nX - stT1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      nTexYInc = ( ( stT2.nY - stT1.nY ) << 16 ) / ( stP2.nY - stP1.nY ),
      // light level increase rate
      nLightLevelInc = ( ( nL2 - nL1 ) << 16 ) / ( stP2.nY - stP1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stP1.nY < 0 ) {
      nYTop = 0;
      // top Y has been cliped then is necessary update values
      nX          -= stP1.nY * nXInc;
      nTexX       -= stP1.nY * nTexXInc;
      nTexY       -= stP1.nY * nTexYInc;
      nLightLevel -= stP1.nY * nLightLevelInc;
   }
   else
      nYTop = stP1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stP2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stP2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test

   do {

      pastEdge [ nYScan ].nX = nX >> 16;  // X in frame buff, convert fix->int
      pastEdge [ nYScan ].stTex.nX = nTexX; // X in texture
      pastEdge [ nYScan ].stTex.nY = nTexY; // Y in texture
      pastEdge [ nYScan ].nColor = nLightLevel; // light level in polygon
      nX += nXInc;                       // new X in next scanline
      nTexX += nTexXInc;                 // new texture X
      nTexY += nTexYInc;                 // new texture Y
      nLightLevel += nLightLevelInc;     // new light level
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// POLYGON SCAN LINE CONVERSION PRIMITIVES
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////



/* Draw a horizontal line.
stEdgeIni and stEdgeEnd have the start and end X for the line.
uCol - is an array with 4 bytes. Each elements must have the
color. Example: to draw a line with color 10 create the array:
char chCol [] = { 10, 10, 10, 10 };
pLineAddress is the position in frame buffer where the line
must be drawn.
High optimized. Clip only X values.
**********************************************************/
void CDraw::mHorizontalLine ( tagPolyEdge& stEdgeIni,
                              tagPolyEdge& stEdgeEnd, UINT uCol,
                              PTRVAL pLineAddress )
{
   // clip x values
   // verify start
   if ( stEdgeIni.nX < 0 )
      stEdgeIni.nX = 0;
   else
      // start can be in right (line outside frame buffer)
      if ( stEdgeIni.nX > m_nMaxX )
         return;
   // verify end
   if ( stEdgeEnd.nX > m_nMaxX )
      stEdgeEnd.nX = m_nMaxX;
   else
      // end can be in left (line outside frame buffer)
      if ( stEdgeEnd.nX < 0 )
         return;

   // draw horizontal line

   // if compile with asm code
   #ifdef COMP_ASM
   ASM_HLINE_FLAT_2 ( pLineAddress + stEdgeIni.nX , stEdgeEnd.nX - stEdgeIni.nX + 1, uCol );
   #else
   // compile C only (no asm code)
   memset ( (void*) ( pLineAddress + stEdgeIni.nX ), uCol, stEdgeEnd.nX - stEdgeIni.nX + 1 );
   #endif
}



/* Draw a horizontal gouraud shaded like line.
Clip X values through given clip edge X limits.
stEdgeIni - left edge values (X and texture start)
stEdgeEnd - right edge values (X and texture end)
pLineAddress is the position in frame buffer where the
line must be drawn.
**********************************************************/
void CDraw::mHorizontalIntLgtLine ( tagPolyEdge& stEdgeIni,
                                     tagPolyEdge& stEdgeEnd,
                                     PTRVAL pLineAddress )
{
   // clip x values
   INT
      nFrXIni = stEdgeIni.nX,   // define left X value in frame buffer
      nFrXEnd = stEdgeEnd.nX;   // define right X value in frame buffer
   // verify start
   if ( stEdgeIni.nX < 0 )
      nFrXIni = 0;
   // verify end
   if ( stEdgeEnd.nX > m_nMaxX )
      nFrXEnd = m_nMaxX;

   // verify if line has no lenght
   if ( nFrXEnd <= nFrXIni )
      return;

   INT
      // define start color
      nColor = stEdgeIni.nColor,
      // color increase rate
      nColorInc = ( stEdgeEnd.nColor - stEdgeIni.nColor ) / ( stEdgeEnd.nX - stEdgeIni.nX );

   // if X has been clipped
   if ( stEdgeIni.nX < 0 )
      nColor -= stEdgeIni.nX * nColorInc;

   // define start and end address in frame buffer
   BYTE
      *pyCurr = (BYTE*) pLineAddress + nFrXIni,   // current address
      *pyEnd = (BYTE*) pLineAddress + nFrXEnd;    // end address

   // line draw loop
   do {

      // set the pixel
     *pyCurr = nColor >> 16;

      // update color
      nColor += nColorInc;
      // go to next pixel
      pyCurr ++;

   } while ( pyCurr <= pyEnd );  // while not found end of line
}



/* Draw a horizontal textured line.
Clip X values through given clip edge X limits.
stEdgeIni - left edge values (X and texture start)
stEdgeEnd - right edge values (X and texture end)
stLimits - the X limits to clip the line
pyTex - texture; pLineAddress - where draw line in frame buffer.
pLineAddress is the position in frame buffer where the
dwLightLevel - the light level in current light table.
line must be drawn.
**********************************************************/
void CDraw::mHorizontalTexturedLine ( tagPolyEdge& stEdgeIni,
                                      tagPolyEdge& stEdgeEnd,
                                      tagInterval& stLimits,
                                      BYTE* pyTex, PTRVAL pLineAddress,
                                      DWORD dwLightLevel )
{
   // clip x values
   INT
      nFrXIni = stEdgeIni.nX,   // define left X value in frame buffer
      nFrXEnd = stEdgeEnd.nX;   // define right X value in frame buffer
   // verify start
   if ( stEdgeIni.nX < stLimits.nStart )
      nFrXIni = stLimits.nStart;
   // verify end
   if ( stEdgeEnd.nX > stLimits.nEnd )
      nFrXEnd = stLimits.nEnd;

   // verify if line has no lenght
   if ( nFrXEnd <= nFrXIni )
      return;

   INT
      // define where texture X and Y start
      nTexX = stEdgeIni.stTex.nX,
      nTexY = stEdgeIni.stTex.nY,
      // texture X and Y increase rates
      nTexXInc = ( stEdgeEnd.stTex.nX - stEdgeIni.stTex.nX ) / ( stEdgeEnd.nX - stEdgeIni.nX ),
      nTexYInc = ( stEdgeEnd.stTex.nY - stEdgeIni.stTex.nY ) / ( stEdgeEnd.nX - stEdgeIni.nX );

   // if X has been clipped
   if ( stEdgeIni.nX < stLimits.nStart ) {
      nTexX -= ( stEdgeIni.nX - stLimits.nStart ) * nTexXInc;
      nTexY -= ( stEdgeIni.nX - stLimits.nStart ) * nTexYInc;
   }

   // define start and end address in frame buffer
   BYTE
      *pyCurr = (BYTE*) pLineAddress + nFrXIni,   // current address
      *pyEnd = (BYTE*) pLineAddress + nFrXEnd;    // end address

   // line draw loop
   do {

      // compile making lighting
      #ifdef COMP_LIGHTS
      // light it

      // C only
      *pyCurr = * ( m_poLightTable->m_payLightTable +

                    // ( nTexY >> 8 ) & 0xFF00 == MAX_COLORS * ( nTexY >> 16 )
                    * ( pyTex + ( ( nTexY >> 8 ) & 0xFF00 ) + ( nTexX >> 16 ) ) +
                   
                    ( dwLightLevel << MAX_COLORS_EXP ) ); // * 256  ==  << 8


      #else    // no lights


      // compile with asm code
      #ifdef COMP_ASM
      ASM_TRANSFER_TEXEL ( pyCurr, pyTex, nTexX, nTexY );
      #else
      // C only
      *pyCurr = * ( pyTex + ( ( nTexY >> 8 ) & 0xFF00 ) + ( nTexX >> 16 ) );
      #endif

      #endif

      // update texture coordenates
      nTexX += nTexXInc;
      nTexY += nTexYInc;
      // go to next pixel
      pyCurr ++;

   } while ( pyCurr <= pyEnd );  // while not found end of line
}



/* Draw a horizontal gouraud shaded (like) textured line.
Clip X values through given clip edge X limits.
stEdgeIni - left edge values (X and texture start)
stEdgeEnd - right edge values (X and texture end)
stLimits - the X limits to clip the line
pyTex - texture; pLineAddress - where draw line in frame buffer.
pLineAddress is the position in frame buffer where the
dwLightLevel - the light level in current light table.
line must be drawn.
**********************************************************/
void CDraw::mHorizontalIntLgtTexturedLine ( tagPolyEdge& stEdgeIni,
                                             tagPolyEdge& stEdgeEnd,
                                             tagInterval& stLimits,
                                             BYTE* pyTex,
                                             PTRVAL pLineAddress )
{
   // clip x values
   INT
      nFrXIni = stEdgeIni.nX,   // define left X value in frame buffer
      nFrXEnd = stEdgeEnd.nX;   // define right X value in frame buffer
   // verify start
   if ( stEdgeIni.nX < stLimits.nStart )
      nFrXIni = stLimits.nStart;
   // verify end
   if ( stEdgeEnd.nX > stLimits.nEnd )
      nFrXEnd = stLimits.nEnd;

   // verify if line has no lenght
   if ( nFrXEnd <= nFrXIni )
      return;

   INT
      // define where texture X and Y start
      nTexX = stEdgeIni.stTex.nX,
      nTexY = stEdgeIni.stTex.nY,
      // texture X and Y increase rates
      nTexXInc = ( stEdgeEnd.stTex.nX - stEdgeIni.stTex.nX ) / ( stEdgeEnd.nX - stEdgeIni.nX ),
      nTexYInc = ( stEdgeEnd.stTex.nY - stEdgeIni.stTex.nY ) / ( stEdgeEnd.nX - stEdgeIni.nX ),
      // define start light level
      nLightLevel = stEdgeIni.nColor,
      // light level increase rate
      nLightLevelInc = ( stEdgeEnd.nColor - stEdgeIni.nColor ) / ( stEdgeEnd.nX - stEdgeIni.nX );

   // if X has been clipped
   if ( stEdgeIni.nX < stLimits.nStart ) {
      nTexX -= ( stEdgeIni.nX - stLimits.nStart ) * nTexXInc;
      nTexY -= ( stEdgeIni.nX - stLimits.nStart ) * nTexYInc;
      nLightLevel -= ( stEdgeIni.nX - stLimits.nStart ) * nLightLevelInc;
   }

   // define start and end address in frame buffer
   BYTE
      *pyCurr = (BYTE*) pLineAddress + nFrXIni,   // current address
      *pyEnd = (BYTE*) pLineAddress + nFrXEnd;    // end address

   // line draw loop
   do {

      // compile making lighting

      // get texel
      // define texel with gouraud light
      #ifdef COMP_LIGHTS

      // compile with asm code
      #ifdef COMP_ASM
      ASM_TRANSFER_LIGHTED_TEXEL ( pyCurr, pyTex, nTexX, nTexY, m_poLightTable->m_payLightTable, nLightLevel );
      #else
      // C only
      *pyCurr = * ( m_poLightTable->m_payLightTable +

                    // ( nTexY >> 8 ) & 0xFF00 == MAX_COLORS * ( nTexY >> 16 )
                    * ( pyTex + ( ( nTexY >> 8 ) & 0xFF00 ) + ( nTexX >> 16 ) ) +
                   
                    ( ( nLightLevel >> 8 ) & 0xFF00 ) );  // === MAX_COLORS * ( nLightLevel >> 16 ) );
      #endif


      #else  // no lights


      // compile with asm code
      #ifdef COMP_ASM
      ASM_TRANSFER_TEXEL ( pyCurr, pyTex, nTexX, nTexY );
      #else
      // C only
      *pyCurr = * ( pyTex + ( ( nTexY >> 8 ) & 0xFF00 ) + ( nTexX >> 16 ) );
      #endif
 
      #endif

      // update texture coordenates
      nTexX += nTexXInc;
      nTexY += nTexYInc;
      nLightLevel += nLightLevelInc;
      // go to next pixel
      pyCurr ++;

   } while ( pyCurr <= pyEnd );  // while not found end of line
}
