/* Read PCX image files from version Z-Soft 5.0 with 256 colors.

TIPS:
- you can: read all file, only the bitmap or only the palette
- you store bitmaps in CTexture objects
- you store palettes in CMem objects
- load palettes with 256 colors only
- this class derive from CFile, see CFile for more info
*/



#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "file.h"

#include "texture.h"

#include "filepcx.h"



/* constructor - no file load
**********************************************************/
CFilePCX::CFilePCX ()
{
}



/* constructor - load PCX palette, if fail abort program
**********************************************************/
CFilePCX::CFilePCX ( CString oszFile, CMem& oPal )
{
   // try load palette, if fail abort
   if ( ! mLoadPalette ( oszFile, oPal ) )
      ABORT ( MSG_ERROR_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* constructor - load bitmap only (not palette),
if fail abort program.
bScale - if TRUE, scale texture to default texture size: TEXTURE_SIZE
defined in defs.h, if FALSE don't (default TRUE).
**********************************************************/
CFilePCX::CFilePCX ( CString oszFile, CTexture& oTx, BOOL bScale )
{
   // try load file, if fail abort
   if ( ! mLoadTexture ( oszFile, oTx, bScale ) )
      ABORT ( MSG_ERROR_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* constructor - load bitmap and palette, all PCX,
if fail abort program
bScale - if TRUE, scale texture to default texture size: TEXTURE_SIZE
defined in defs.h, if FALSE don't (default TRUE).
**********************************************************/
CFilePCX::CFilePCX ( CString oszFile, CTexture& oTx , CMem& oPal, BOOL bScale )
{
   // try load file, if fail abort
   if ( ! mLoad ( oszFile, oTx, oPal, bScale ) )
      ABORT ( MSG_ERROR_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* Load palette from file, return TRUE - ok, FALSE - error
**********************************************************/
BOOL CFilePCX::mLoadPalette ( CString oszFile, CMem& oPal )
{
   // try open PCX file, if fail exit
   if ( ! mOpen ( oszFile, "rb" ) )
      return FALSE;

   // goto start palette buffer in file (768 last bytes)
   mPosEnd ( 768 );

   oPal.mCreate ( 768 );

   // read file
   mRetMem ( 768, oPal );

   // shift each byte of palette 2 bits right, for conversion to 6 bits
   for ( DWORD ucP = 0; ucP < 768; ucP ++ )
      oPal.m_payMem [ ucP ] = oPal.m_payMem [ ucP ] >> 2;

   // close file
   mClose ();

   return TRUE;
}



/* Load bitmap from file.
bScale - if TRUE, scale texture to default texture size: TEXTURE_SIZE
defined in defs.h, if FALSE don't (default TRUE).
return TRUE - ok, FALSE - error
**********************************************************/
BOOL CFilePCX::mLoadTexture ( CString oszFile, CTexture& oTx, BOOL bScale )
{
   // try open PCX file, if fail exit
   if ( ! mOpen ( oszFile, "rb" ) )
      return FALSE;

   // read header
   tagHeaderPCX
      stH;
   mRetMem ( sizeof ( tagHeaderPCX ), (void*) &stH );

   // define image dimensions
   oTx.m_nWidth = stH.uXMax - stH.uXMin + 1;
   oTx.m_nHeight = stH.uYMax - stH.uYMin + 1;

   // alocate memory
   oTx.CMem::mCreate ( oTx.m_nWidth * oTx.m_nHeight );

   // define image size
   DWORD
      dwSize = oTx.m_nWidth * oTx.m_nHeight;

   // read image
   // jump first 128 bytes (header)
   mPosStart ( 128 );
   // points image data
   BYTE
      *pyNow = oTx.m_payMem,
      *pyEnd = oTx.m_payMem + dwSize;

   // while not end of image load
   while ( pyNow < pyEnd ) {

      // read number of color or number of ocurrences from one color
      BYTE
         yVal;
      mRetMem ( 1, (void*) &yVal );

      // if value read is less than 192 then be a color, draw bitmap
      if ( yVal < 192 ) {

         *pyNow = yVal;
         pyNow ++;
      }
      else { // else is a number of ocurrences of a color

         // compute number of ocurrences, only first 6 bits valid
         yVal &= 0x3F;
         // read next byte with the color number
         BYTE
            yColor;
         mRetMem ( 1, (void*) &yColor );
         // draw bitmap
         memset ( pyNow, yColor, yVal );
         // points new image area
         pyNow += yVal;
      }
   }

   // close file
   mClose ();

   // scale bitmap (if need)
   if ( bScale )
      oTx.mScale ( TEXTURE_SIZE, TEXTURE_SIZE );

   return TRUE;
}



/* Load whole PCX file, palette and bitmap,
bScale - if TRUE, scale texture to default texture size: TEXTURE_SIZE
defined in defs.h, if FALSE don't (default TRUE).
return TRUE - ok, FALSE - error
**********************************************************/
BOOL CFilePCX::mLoad ( CString oszFile, CTexture& oTx, CMem& oPal, BOOL bScale )
{
   // try load palette
   if ( mLoadPalette ( oszFile, oPal ) )
      // try load bitmap
      if ( mLoadTexture ( oszFile, oTx, bScale ) )
         return TRUE;

   return FALSE;
}
