/* Implements a light table. A light table is one essential component to
make light effects in polygons. It define the efective color (palette index,
in 256 color modes) to be used if the poly is shaded.
Acording the light intensity over one polygon it's texels color need change.
The change is acording the light intensity over the texel.
The light intensity of a pixel is called light level. The number of
light levels can change in table creation time. Not necessarily
much levels means better quality, good values are 16, 32,..., 256 -
probably more isn't better.

With this class is possible: create a Light table giving a palette,
save it to disk, or load.

The light table is one array with 2 dimensions.
along X axis (X2) is the complete palette colors
for one light level. Along Y axis (Y2) is all light levels for one
color.

FEATURES:
- suport for colored light sources and hilights.

TIPS:
- the number of light levels is defined acording the file loaded and is
  stored in member variable m_dwNumLevels.
- To use the light table follow the example:
  if you want to know the palette color of the color 50 with light level 6
     pal_index = * ( m_payLightTable + 50 + m_dwNumLevels * 6 )
- the recomended file extension for light table files is LTB
- to know with a CLightTable object have a light table compare:
  m_payLightTable == NULL, if TRUE object is empty.
- some lighting parameters can be changed in config.h and src/defs.h
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "file.h"

#include "texture.h"

#include "lighttab.h"



/* constructor - init empty table
**********************************************************/
CLightTable::CLightTable ()
{
   m_payLightTable = NULL;
   m_dwNumLevels = 0;
}



/* constructor - init table with a copy of other table
**********************************************************/
CLightTable::CLightTable ( CLightTable& oLgT )
{
   m_payLightTable = NULL;
   m_dwNumLevels = 0;

   mSetObj ( oLgT );
}



/* constructor - load table from disk
**********************************************************/
CLightTable::CLightTable ( CString oszFile )
{
   m_payLightTable = NULL;
   m_dwNumLevels = 0;

   // try load table, if fail abort
   if ( ! mCreateLoading ( oszFile ) )
      ABORT ( MSG_ERROR_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* constructor - make table from a palette
dwNumLevels - number of light levels
oPal - source palette
fNormR, fNormG, fNormB - the normalized (0...1) R,G,B for light source
**********************************************************/
CLightTable::CLightTable ( CMem& oPal, DWORD dwNumLevels,
                           float fNormR, float fNormG, float fNormB )
{
   m_payLightTable = NULL;
   m_dwNumLevels = 0;

   mCreate ( oPal, dwNumLevels, fNormR, fNormG, fNormB );
}



/* destructor - desalocate table, if alocated
**********************************************************/
CLightTable::~CLightTable ()
{
   mDestroy ();
}



/* Create table, loading from disk
oszFile - file to load
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CLightTable::mCreateLoading ( CString oszFile )
{
   mDestroy ();

   // try open file
   CFile
      oF;
   if ( ! oF.mOpen ( oszFile, "rb" ) )
      return FALSE;

   // try load signature
   CString
      oszSign ( 3 );
   if ( ! oF.mRetMem ( 2, (void*) oszSign ) )
      return FALSE;

   // compare char by char (problems returning strings from binary file read)
   if ( *( (char*) oszSign + 0 ) != *( (char*) LIGHT_TABLE_SIGN + 0 ) &&
        *( (char*) oszSign + 1 ) != *( (char*) LIGHT_TABLE_SIGN + 1 ) )
      return FALSE;

   // try load version
   DWORD
      dwVer;
   if ( ! oF.mRetMem ( 4, (void*) &dwVer ) )
      return FALSE;
   // compare
   if ( dwVer != LIGHT_TABLE_VERSION ) {
      printf ( MSG_ERROR_INCORRECT_VERSION );
      return FALSE;
   }

   // try load number of light levels
   if ( ! oF.mRetMem ( 4, (void*) &m_dwNumLevels ) )
      return FALSE;
   
   // alocate memory
   mAlocate ( m_dwNumLevels );

   // try load table
   if ( ! oF.mRetMem ( MAX_COLORS * m_dwNumLevels, (void*) m_payLightTable ) )
      return FALSE;

   return TRUE;
}



/* Make table from a palette
dwNumLevels - number of light levels
oPal - source palette
**********************************************************/
void CLightTable::mCreate ( CMem& oPal, DWORD dwNumLevels,
                            float fNormR, float fNormG, float fNormB )
{
   mDestroy ();

   // alocate memory
   mAlocate ( dwNumLevels );

   // number of colors
   DWORD
      dwNumColors = oPal.m_dwSize / 3;   

   #ifndef COMP_HILIGHTS

   // Compute the light table from dark to bitmap texture color.
   // Not used colored light sources.

   // for each light level
   for ( DWORD dwcL = 0; dwcL < dwNumLevels; dwcL ++ )

      // for each color
      for ( DWORD dwcC = 0; dwcC < dwNumColors; dwcC ++ ) {

         // the palette RGB components for the color
         DWORD
            dwRed   = * ( oPal.m_payMem + dwcC * 3 + 0 ),
            dwGreen = * ( oPal.m_payMem + dwcC * 3 + 1 ),
            dwBlue  = * ( oPal.m_payMem + dwcC * 3 + 2 );
         
         // define the lighted RGB
         dwRed =   (DWORD) ( (float) dwRed   / dwNumLevels * ( dwcL + 1 ) );
         dwGreen = (DWORD) ( (float) dwGreen / dwNumLevels * ( dwcL + 1 ) );
         dwBlue =  (DWORD) ( (float) dwBlue  / dwNumLevels * ( dwcL + 1 ) );

         // define lighted color (palette index)
         * ( m_payLightTable + dwcC + dwcL * MAX_COLORS ) = mSearchClosestColor ( oPal, dwRed, dwGreen, dwBlue );
      }
   #endif

   #ifdef COMP_HILIGHTS

   // Compute the light table from dark to bitmap texture color
   // (with light source color) and then to light source color.
   // Example:
   // dark: (0, 0, 0)
   // texture color: (30, 170, 5)
   // light source color: (0.2, 0.3, 0.4)
   //        part 1                                part 2
   // (0, 0, 0) -> (30 * 0.2, 170 * 0.3, 5 * 0.4) -> (0.2, 0.3, 0.4)
   //    0...dwNumColors/2              dwNumColors/2+1...dwNumColors

   // if only one light level, use only part 1 (ajust flags to avoid part 2)
   BOOL
      bOnlyOneLevel = FALSE;
   DWORD
      dwMax = dwNumLevels;
   if ( dwNumLevels == 1 ) {
      bOnlyOneLevel = TRUE;
      dwMax = 2;
   }

   // PART 1:
   // dark to texture color
   // for each light level
   for ( DWORD dwcL = 0; dwcL < dwMax / 2; dwcL ++ ) {

      // for each color
      DWORD
         dwNumColors = oPal.m_dwSize / 3;
      for ( DWORD dwcC = 0; dwcC < dwNumColors; dwcC ++ ) {

         // the palette RGB components for the color
         DWORD
            dwRed   = * ( oPal.m_payMem + dwcC * 3 + 0 ),
            dwGreen = * ( oPal.m_payMem + dwcC * 3 + 1 ),
            dwBlue  = * ( oPal.m_payMem + dwcC * 3 + 2 );
         
         // define the lighted RGB
         dwRed =   (DWORD) ( (float) dwRed   * fNormR / ( dwMax / 2 ) * ( dwcL + 1 ) );
         dwGreen = (DWORD) ( (float) dwGreen * fNormG / ( dwMax / 2 ) * ( dwcL + 1 ) );
         dwBlue =  (DWORD) ( (float) dwBlue  * fNormB / ( dwMax / 2 ) * ( dwcL + 1 ) );

         // define lighted color (palette index)
         * ( m_payLightTable + dwcC + dwcL * MAX_COLORS ) = mSearchClosestColor ( oPal, dwRed, dwGreen, dwBlue );
      }
   }

   // verify if using only one level, if no use part 2
   if ( ! bOnlyOneLevel ) {

      // PART 2:
      // texture color to light source color
      float
         fLightR = fNormR * PAL_RANGE,   // light color in palette range
         fLightG = fNormG * PAL_RANGE,
         fLightB = fNormB * PAL_RANGE;
      // for each light level for this color
      for ( DWORD dwcL = dwNumLevels / 2; dwcL < dwNumLevels; dwcL ++ )

         for ( DWORD dwcC = 0; dwcC < dwNumColors; dwcC ++ ) {

            // the palette RGB components for the color
            DWORD
               dwRed   = * ( oPal.m_payMem + dwcC * 3 + 0 ),
               dwGreen = * ( oPal.m_payMem + dwcC * 3 + 1 ),
               dwBlue  = * ( oPal.m_payMem + dwcC * 3 + 2 );

            // define the lighted RGB
            float
               fMidR = dwRed   * fNormR,   // the texture color with light color
               fMidG = dwGreen * fNormR,
               fMidB = dwBlue  * fNormR;
            dwRed   = (DWORD) ( fMidR + fLightR / ( dwNumLevels / 2 ) * ( dwcL - ( dwNumLevels / 2 ) + 1 ) );
            dwGreen = (DWORD) ( fMidG + fLightG / ( dwNumLevels / 2 ) * ( dwcL - ( dwNumLevels / 2 ) + 1 ) );
            dwBlue  = (DWORD) ( fMidB + fLightB / ( dwNumLevels / 2 ) * ( dwcL - ( dwNumLevels / 2 ) + 1 ) );

            // define lighted color (palette index)
            * ( m_payLightTable + dwcC + dwcL * MAX_COLORS ) = mSearchClosestColor ( oPal, dwRed, dwGreen, dwBlue );
         }
   }
   #endif
}        



/* Desalocate table, if alocated
**********************************************************/
void CLightTable::mDestroy ()
{
   // verify if it's alocated
   if ( m_payLightTable != NULL ) {
      delete m_payLightTable;
      m_payLightTable = NULL;
      m_dwNumLevels = 0;
   }
}



/* Init table with a copy of other object
(like CLightTable::constructor and CLightTable::operator =)
**********************************************************/
void CLightTable::mSetObj ( CLightTable& oLgT )
{
   mDestroy ();
   mAlocate ( oLgT.m_dwNumLevels );

   memcpy ( (void*) m_payLightTable, (void*) oLgT.m_payLightTable, MAX_COLORS * m_dwNumLevels );
}



/* like CLightTable::constructor and CLightTable::mSetObj
**********************************************************/
void CLightTable::operator = ( CLightTable& oLgT )
{
   mSetObj ( oLgT );
}



/* Save table to disk
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CLightTable::mSave ( CString oszFile )
{
   // try open file
   CFile
      oF;
   if ( ! oF.mOpen ( oszFile, "wb" ) )
      return FALSE;

   // try save signature
   if ( ! oF.mSetMem ( 2, (void*) LIGHT_TABLE_SIGN ) )
      return FALSE;

   // try save version
   DWORD
      dwVer = LIGHT_TABLE_VERSION;
   if ( ! oF.mSetMem ( 4, (void*) &dwVer ) )
      return FALSE;

   // try save number of light levels
   if ( ! oF.mSetMem ( 4, (void*) &m_dwNumLevels ) )
      return FALSE;

   // try save table
   if ( ! oF.mSetMem ( MAX_COLORS * m_dwNumLevels, (void*) m_payLightTable ) )
      return FALSE;

   return TRUE;
}



/* Make Light table memory alocation
**********************************************************/
void CLightTable::mAlocate ( DWORD dwNumLevels )
{
   // try alocation
   if ( ( m_payLightTable = new BYTE [ MAX_COLORS * dwNumLevels ] ) == NULL )
      ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

   m_dwNumLevels = dwNumLevels;
}



/* Scan the palette oPal searching for the closest color
with the given RGB.
The algorithm:
think in RGB like x, y, z coords and the palette colors
like points in space. The closest color is the closest point.
return the palette color found.
**********************************************************/
DWORD CLightTable::mSearchClosestColor ( CMem& oPal, DWORD dwRed,
                                         DWORD dwGreen, DWORD dwBlue )
{
   // distance between colors
   float
      fCloseDist = 100000;   // initial (impossible) big value
   DWORD
      dwCloseColor = 0;       // the closest color

   // for each color in palette
   DWORD
      dwNumColors = oPal.m_dwSize / 3;
   for ( DWORD dwcC = 0; dwcC < dwNumColors; dwcC ++ ) {
      
      // the palette RGB components for the color
      DWORD
         dwPalRed   = * ( oPal.m_payMem + dwcC * 3 + 0 ),
         dwPalGreen = * ( oPal.m_payMem + dwcC * 3 + 1 ),
         dwPalBlue  = * ( oPal.m_payMem + dwcC * 3 + 2 );

      // distance between palette color and the given color RGB
      INT
         nDistR = dwPalRed - dwRed,
         nDistG = dwPalGreen - dwGreen,
         nDistB = dwPalBlue - dwBlue;
      float
         fDist = // sqrt
                      ( nDistR * nDistR +
                        nDistG * nDistG +
                        nDistB * nDistB );

      // compare with the currently closest point
      if ( fDist <= fCloseDist ) {
         fCloseDist = fDist;
         dwCloseColor = dwcC;
      }
   }

   // the closest color
   return dwCloseColor;
}
