/* Implement a linked list with forward and backward links.

FEATURE:
- direct access: this feature provide direct access to an element in an
  random position, like an array. Is good to use if you need fast
  read of nodes in random positions (not sequence). Is implemented
  by an array alocated dynamicaly that have pointers to the node's list.
  Warning: direct access is not updated if you change the list nodes
  (insert a node or remove one). After this call again the direct access
  function: CList::mEnableDirectAccess

TIPS:
- CNode is the class of nodes for CList instances. You can derive from CNode
  your own classes.
- to know if list is empty use the comparison CList::mRetNumNodes () == 0
- the index of first node's list is 0
- the index of last node's list is CList::mRetNumNodes () - 1
- to deattach CList internal pointers to a list create a empty CList object
  and attach CList to this one. Example:
   oAttached - is the CList that you want to deattach
   {
      CList
         oEmpty;  (the empty list)
      oAttached = oEmpty;  (deattach oAttached to list and attach to empty)
   }
- to disable the direct access feature call CList::mDisableDirectAccess

CAUTION:
- if you remove or do other operations that change the nodes pointers the
  direct access feature is not automaticaly updated, you must re-run
  CList::mEnableDirectAccess. If you insert a new node or remove one you
  must after that call CList::mEnableDirectAccess.
- THE DESTRUCTOR DOES NOT DELETE (desalocate memory) THE NODES. YOU MUST
  CALL CList::mRemoveAll.

BUGS:
- if list have only 1 node and the extremes are linked this crach when
  removing this node (why? something related with the internal pointers
  of this class)
*/



#include "defs.h"

#include "node.h"

#include "list.h"



/* constructor - init pointers with NULL
**********************************************************/
CList::CList ()
{
   m_poStart = m_poEnd = m_poSelected = NULL;
   m_papoDirectAccess = NULL;
   m_dwNumNodes = 0;
}



/* constructor - init pointers com definicoes de outra List
**********************************************************/
CList::CList ( CList& oList )
{
   mSetObj ( oList );
}



/* destructor - do NOT remove nodes
**********************************************************/
CList::~CList ()
{
}



/* Remove all node's list (delete all of them desalocating memory)
**********************************************************/
void CList::mRemoveAll ()
{
   // while the first node is not NULL execute action
   while ( m_poStart != NULL ) {

      CNode
         *poNode = m_poStart;

      m_poStart = m_poStart->m_poNext;

      // desalocate node
      delete poNode;
   }

   mDisableDirectAccess ();

   m_poStart = m_poEnd = m_poSelected = NULL;
   m_dwNumNodes = 0;
}



/* Remove the first node's list
If bDelete is TRUE delete node, if FALSE only change list
pointers without desalocating memory used by the node.
Default bDelete = FALSE
return CNode pointer if node removed, but not deleted /
NULL - if node deleted or not found
**********************************************************/
CNode* CList::mRemoveStart ( BOOL bDelete )
{
   // if list is not empty
   if ( m_poStart != NULL ) {

      CNode
         *poNode = m_poStart;

      // verify if the list only have one node
      if ( m_poStart == m_poEnd )
         m_poStart = m_poEnd = m_poSelected = NULL;
      else {

         // verify if be selected the first node, if yes select next
         if ( m_poSelected == m_poStart )
            m_poSelected = m_poSelected->m_poNext;

         // the new start node will be the next node
         m_poStart = m_poStart->m_poNext;
         m_poStart->m_poPrevious = NULL;
      }

      m_dwNumNodes --;

      // verify if must desalocate node
      if ( bDelete ) {
         delete poNode;
         return NULL;
      }
      else
         return poNode;
   }

   return NULL;
}



/* Remove the last node's list
If bDelete is TRUE delete node, if FALSE only change list
pointers without desalocating memory used by the node.
Default bDelete = FALSE
return CNode pointer if node removed, but not deleted /
NULL - if node deleted or not found
**********************************************************/
CNode* CList::mRemoveEnd ( BOOL bDelete )
{
   // if the ;ist is not empty remove
   if ( m_poEnd != NULL ) {

      CNode
         *poNode = m_poEnd;

      // verify if the list has only one node
      if ( m_poStart == m_poEnd )
         m_poStart = m_poEnd = m_poSelected = NULL;
      else {

         // verify if is selected the last node, is yes select previous
         if ( m_poSelected == m_poEnd )
            m_poSelected = m_poSelected->m_poPrevious;

         // the new end will the the previous node
         m_poEnd = m_poEnd->m_poPrevious;
         m_poEnd->m_poNext = NULL;
      }

      m_dwNumNodes --;

      // verify if must desalocate node
      if ( bDelete ) {
         delete poNode;
         return NULL;
      }
      else
         return poNode;
   }

   return NULL;
}



/* Remove the selected note, select the previous
(or next if no previous)
If bDelete is TRUE delete node, if FALSE only change list
pointers without desalocating memory used by the node.
Default bDelete = FALSE
return CNode pointer if node removed, but not deleted /
NULL - if node deleted or not found
**********************************************************/
CNode* CList::mRemoveSelected ( BOOL bDelete )
{
   // if exist a selected node
   if ( m_poSelected != NULL ) {

      CNode
         *poNode = m_poSelected;

      // verify if the list only have one node
      if ( m_poStart == m_poEnd )
         m_poStart = m_poEnd = m_poSelected = NULL;
      else
         // verify if the selected node is the first
         if ( m_poStart == m_poSelected ) {

            // select the next node
            m_poStart = m_poSelected = m_poSelected->m_poNext;
            m_poSelected->m_poPrevious = NULL;
         }
         else
            // verify if the selected node is the last
            if ( m_poEnd == m_poSelected ) {

               // select the previous node
               m_poEnd = m_poSelected = m_poSelected->m_poPrevious;
               m_poSelected->m_poNext = NULL;
            }
            else {
               // if the selected node is in the midle of the list

               // change pointer's node
               m_poSelected->m_poPrevious->m_poNext = m_poSelected->m_poNext;
               m_poSelected->m_poNext->m_poPrevious = m_poSelected->m_poPrevious;

               m_poSelected = m_poSelected->m_poPrevious;
            }

      m_dwNumNodes --;

      // verify if must desalocate node
      if ( bDelete ) {
         delete poNode;
         return NULL;
      }
      else
         return poNode;
   }

   return NULL;
}



/* Remove the node that be in a given position (start in 0)
If bDelete is TRUE delete node, if FALSE only change list
pointers without desalocating memory used by the node.
Default bDelete = FALSE
return CNode pointer if node removed, but not deleted /
NULL - if node deleted or not found
**********************************************************/
CNode* CList::mRemoveNode ( DWORD dwIndex, BOOL bDelete )
{
   // node that will be removed
   CNode
      *poNode;

   // verify if index is out of list range, if yes return
   if ( dwIndex >= m_dwNumNodes )
      return NULL;

   // verify if exist direct access table, if yes use
   if ( m_papoDirectAccess != NULL )
      poNode = m_papoDirectAccess [ dwIndex ];
   else {
      // navigate list from start to position
      poNode = m_poStart;
      for ( DWORD dwcPos = 0; dwcPos < dwIndex; dwcPos ++ )

         // verify if find the end of the list
         if ( poNode->m_poNext == NULL )
            return NULL;           // node not found
         else
            poNode = poNode->m_poNext;
   }

   // verify if list have only one node
   if ( m_poStart == m_poEnd )
      m_poStart = m_poEnd = m_poSelected = NULL;
   else
      // verify if the selected node is the first
      if ( m_poStart == poNode ) {

         // set the start with next node
         m_poStart = poNode->m_poNext;

         m_poStart->m_poPrevious = NULL;

         // if the selected node is the node that will be desalocated
         if ( poNode == m_poSelected )
            // the selected continue the first node
            m_poSelected = m_poStart;
      }
      else
         // verify if the node is the last
         if ( m_poEnd == poNode ) {

            // set the last with previous node
            m_poEnd = m_poEnd->m_poPrevious;

            m_poEnd->m_poNext = NULL;

            // if the selected node is the last
            if ( poNode == m_poSelected )
               // the selected continue the last node
               m_poSelected = m_poEnd;
         }
         else {
            // if the selected node is in the midle of the list

            // change pointers's nodes
            poNode->m_poPrevious->m_poNext = poNode->m_poNext;
            poNode->m_poNext->m_poPrevious = poNode->m_poPrevious;

            // if the selected node is the removed node
            if ( poNode == m_poSelected )
               // selected node continue in same index
               m_poSelected = poNode->m_poPrevious;
         }

   m_dwNumNodes --;

   // verify if must desalocate node
   if ( bDelete ) {
      delete poNode;
      return NULL;
   }
   else
      return poNode;
}



/* Insert one node behind first node's list
**********************************************************/
void CList::mSetStart ( CNode* poNode )
{
   // verify if the list is not empty
   if ( m_poStart != NULL ) {
      m_poStart->m_poPrevious = poNode;
   }
   else
      // if list is empty
      m_poEnd = m_poSelected = poNode;

   poNode->m_poPrevious = NULL;
   poNode->m_poNext = m_poStart;

   m_poStart = poNode;

   m_dwNumNodes ++;
}



/* Insert one node after last list node (like CList:: operator =)
**********************************************************/
void CList::mSetEnd ( CNode* poNode )
{
   // verify if the list is not empty
   if ( m_poEnd != NULL ) {
      m_poEnd->m_poNext = poNode;
   }
   else
      // if list is empty
      m_poStart = m_poSelected = poNode;

   poNode->m_poPrevious = m_poEnd;
   poNode->m_poNext = NULL;

   m_poEnd = poNode;

   m_dwNumNodes ++;
}



/* Insert one node after the selected node
**********************************************************/
void CList::mSetSelected ( CNode* poNode )
{
   // verify if the list is empty
   if ( m_poStart == NULL ) {
      m_poStart = m_poEnd = m_poSelected = poNode;
      poNode->m_poNext = poNode->m_poPrevious = NULL;
   }
   else
      // if the selected is the last
      if ( m_poSelected == m_poEnd ) {

         poNode->m_poNext = NULL;
         m_poSelected->m_poNext = poNode;
         poNode->m_poPrevious = m_poSelected;
         m_poEnd = poNode;
      }
      else {
         // if selected is the first or one in the midle

         m_poSelected->m_poNext->m_poPrevious = poNode;
         poNode->m_poNext = m_poSelected->m_poNext;
         m_poSelected->m_poNext = poNode;
         poNode->m_poPrevious = m_poSelected;
      }

   m_dwNumNodes ++;
}



/* Insert one node behind the node specified by the index.
If index don't exist insert at end of list.
**********************************************************/
void CList::mSetNode ( DWORD dwIndex, CNode* poNodeIns )
{
   // node that is in index
   CNode
      *poNode;

   // verify if index is out of list range, if yes insert end
   if ( dwIndex >= m_dwNumNodes ) {
      mSetEnd ( poNodeIns );
      return;
   }

   // verify if exist direct access table, if yes use
   if ( m_papoDirectAccess != NULL )
      poNode = m_papoDirectAccess [ dwIndex ];
   else {
      // navigate list from start to position
      poNode = m_poStart;
      for ( DWORD dwcPos = 0; dwcPos < dwIndex; dwcPos ++ )

         // verify if find the end of the list, insert at end
         if ( poNode->m_poNext == NULL ) {
            mSetEnd ( poNodeIns );
            return;
         }
         else
            poNode = poNode->m_poNext;
   }

   // verify if the list is empty
   if ( m_poStart == NULL ) {
      m_poStart = m_poEnd = m_poSelected = poNodeIns;
      poNode->m_poNext = poNode->m_poPrevious = NULL;
   }
   else
      // if insertion is behind the first node
      if ( m_poStart == poNode ) {

         m_poStart->m_poPrevious = poNodeIns;
         poNodeIns->m_poNext = m_poStart;
         poNodeIns->m_poPrevious = NULL;
         m_poStart = poNodeIns;
      }
      else {
         // if insertion is behind the last position or in the midle
         poNode->m_poPrevious->m_poNext = poNodeIns;
         poNodeIns->m_poPrevious = poNode->m_poPrevious;
         poNodeIns->m_poNext = poNode;
         poNode->m_poPrevious = poNodeIns;
      }

   m_dwNumNodes ++;
}



/* Return a pointer to first node
**********************************************************/
CNode* CList::mRetStart ()
{
   return m_poStart;
}



/* Return a pointer to last node
**********************************************************/
CNode* CList::mRetEnd ()
{
   return m_poEnd;
}



/* Return a pointer to selected node
**********************************************************/
CNode* CList::mRetSelected ()
{
   return m_poSelected;
}



/* Return a pointer to the node specified by index
return !NULL - pointer to the node / NULL - fail, index don't
exist. (like CList::operator [])
**********************************************************/
CNode* CList::mRetNode ( DWORD dwIndex )
{
   // point node that will be returned
   CNode
      *poNode;

   // verify if index is out of list range, if yes return
   if ( dwIndex >= m_dwNumNodes )
      return NULL;

   // verify if direct access table exist
   if ( m_papoDirectAccess != NULL )
      return m_papoDirectAccess [ dwIndex ];
   else {
      // navigate list from start to position
      poNode = m_poStart;
      for ( DWORD dwcPos = 0; dwcPos < dwIndex; dwcPos ++ )

         // verify if find the end of the list
         if ( poNode->m_poNext == NULL )
            return NULL;
         else
            poNode = poNode->m_poNext;
   }

   return poNode;
}



/* Select the first node
**********************************************************/
void CList::mSelectStart ()
{
   m_poSelected = m_poStart;
}



/* Select the last node
**********************************************************/
void CList::mSelectEnd ()
{
   m_poSelected = m_poEnd;
}



/* Select the node that be in the given index return TRUE - done /
FALSE - fail, position don't exist (select NULL)
**********************************************************/
BOOL CList::mSelectNode ( DWORD dwIndex )
{
   // verify if index is out of list range, if yes return
   if ( dwIndex >= m_dwNumNodes )
      return FALSE;

   // navigate list from start to the given index
   m_poSelected = m_poStart;
   for ( DWORD dwcPos = 0; dwcPos < dwIndex; dwcPos ++ )

      // verify if find the end of the list
      if ( m_poSelected->m_poNext == NULL ) {
         m_poSelected = NULL;
         return FALSE;
      }
      else
         m_poSelected = m_poSelected->m_poNext;

   return TRUE;
}



/* Define selected node as one of the next nodes
return TRUE - done / FALSE - fail, position don't exist (select NULL).
Default offset 1.
**********************************************************/
BOOL CList::mSelectNext ( DWORD dwOffset )
{
   // seek nodes
   for ( DWORD dwcPos = 0; dwcPos < dwOffset; dwcPos ++ )

      // verify if find the end of the list
      if ( m_poSelected->m_poNext == NULL ) {
         m_poSelected = NULL;
         return FALSE;
      }
      else
         m_poSelected = m_poSelected->m_poNext;

   return TRUE;
}



/* Define selected node as one of the previous nodes
return TRUE - done / FALSE - fail, position don't exist (select NULL).
Default offset 1.
**********************************************************/
BOOL CList::mSelectPrevious ( DWORD dwOffset )
{
   // seek nodes
   for ( DWORD dwcPos = 0; dwcPos < dwOffset; dwcPos ++ )

      // verify if find the start of the list
      if ( m_poSelected->m_poPrevious == NULL ) {
         m_poSelected = NULL;
         return FALSE;
      }
      else
         m_poSelected = m_poSelected->m_poPrevious;

   return TRUE;
}



/* Change selection relative (acording) the position of the selected node
return TRUE - done / FALSE - fail, position don't exist (select NULL).
**********************************************************/
BOOL CList::mSelectOffset ( INT nOffset )
{
   // seek nodes
   if ( nOffset < 0 ) {

      nOffset = - nOffset;

      for ( INT ncPos = 0; ncPos < nOffset; ncPos ++ )

         // verify if find the end of the list
         if ( m_poSelected->m_poPrevious == NULL ) {
            m_poSelected = NULL;
            return FALSE;
         }
         else
            m_poSelected = m_poSelected->m_poPrevious;
   }
   else
      for ( INT ncPos = 0; ncPos < nOffset; ncPos ++ )

         // verify if find the start of the list
         if ( m_poSelected->m_poNext == NULL ) {
            m_poSelected = NULL;
            return FALSE;
         }
         else
            m_poSelected = m_poSelected->m_poNext;

   return TRUE;
}



/* Return TRUE in exist selection, FALSE if selection is NULL
**********************************************************/
BOOL CList::mRetExistSelection ()
{
   if ( m_poSelected == NULL )
      return FALSE;

   return TRUE;
}



/* Return the number of nodes in list
**********************************************************/
DWORD CList::mRetNumNodes ()
{
   return m_dwNumNodes;
}



/* Link the extreme pointers doing a round list
**********************************************************/
void CList::mLinkExtremes ()
{
   // verify if list is not empty
   if ( m_poStart != NULL ) {
      m_poStart->m_poPrevious = m_poEnd;
      m_poEnd->m_poNext = m_poStart;
   }
}



/* Unlink the extreme pointers (not doing a round list)
**********************************************************/
void CList::mUnlinkExtremes ()
{
   // verify if list is not empty
   if ( m_poStart != NULL )
      m_poStart->m_poPrevious = m_poEnd->m_poNext = NULL;
}



/* Criate direct access array (when criated nodes can be accessed
 so fast like in an array)
return TRUE - ok or list empty / FALSE, can't create (no memory)
**********************************************************/
BOOL CList::mEnableDirectAccess ()
{
   mDisableDirectAccess ();

   // verify if index is out of list empty, if yes return
   if ( m_dwNumNodes == 0 )
      return TRUE;

   // try alocate memory for direct access
   m_papoDirectAccess = new CNode* [ m_dwNumNodes ];
   if ( m_papoDirectAccess == NULL )
      return FALSE;

   // navigate list initialize array of pointers with points to node's list
   CNode
      *poNode = m_poStart;
   DWORD
      dwIndex = 0;
   // while end of list not found
   while ( poNode != NULL ) {

      m_papoDirectAccess [ dwIndex ] = poNode;

      // next node
      poNode = poNode->m_poNext;
      dwIndex ++;
   }

   return TRUE;
}



/* Desalocate direct access array
**********************************************************/
void CList::mDisableDirectAccess ()
{
   if ( m_papoDirectAccess != NULL ) {
      delete m_papoDirectAccess;
      m_papoDirectAccess = NULL;
   }
}



/* Define internal pointers to other object list, NOT copy list,
only atach.
(like CList::constructor, CList::operator =)
**********************************************************/
void CList::mSetObj ( CList& oList )
{
   m_poStart = oList.mRetStart ();
   m_poEnd = oList.mRetEnd ();
   m_poSelected = oList.mRetSelected ();
   m_dwNumNodes = oList.mRetNumNodes ();

   mDisableDirectAccess ();

   m_papoDirectAccess = oList.m_papoDirectAccess;
}



/* 1 - like CList::constructor and CList::mSetObj
I DESRECOMEND YOU USE THIS BECAUSE CAN CAUSE MANY
CONFUSION WITH THE C = OPERATOR.
**********************************************************/
void CList::operator = ( CList& oList )
{
   mSetObj ( oList );
}



/* 2 - like CList::mSetEnd. I DESRECOMEND YOU USE THIS
BECAUSE CAN CAUSE MANY CONFUSION WITH THE C = OPERATOR.
**********************************************************/
void CList::operator = ( CNode* poNode )
{
   mSetEnd ( poNode );
}



/* like CList::mRetNode
**********************************************************/
CNode* CList::operator [] ( DWORD dwIndex )
{
   return mRetNode ( dwIndex );
}
