/* Implement a discrete function. Make an association between one key and
one value. The key is a double word and the value is a object derived
from CNode, alocated dynamicaly by programmer.

FEATURES:
- CMap use 3 heuristics to improve fast search for a given key:
  "cache", "indexing" and binary-tree search.
- "cache": the last key added or returned have your position
  saved in internal variables. If you call this key the map know
  previously where is the value, speeding the answer.
- "indexing": verify if the key and its index in keys array is
  the same. If yes return imediatly the value. To use this facility
  start key count in any value an increate next keys to be inserted by one.
  Example: first key - 3, second - 4, ...

TIPS:
- use this class to do relations between 2 objects, like an identifier
  an it's value. Ex.: texture ID (as key) and the texture (as value).
  In this case the texture must be stored in a class derived for
  CTexture and CNode because values must be derived from CNode.
- if you want to know if some key exist use the comparison:
  CMap::mRetValue ( KEY_TEST )==NULL. Keys that not exist cause return NULL,
  KEY_TEST is the key you want to test (a double word value).

CAUTION:
- you can add equal keys to distinct values, only one value would be
  visible (the first found). To make the others visible you must remove
  this one. The visible value is not necessarily the last value inserted.
*/



#include "defs.h"

#include "node.h"

#include "nodeint.h"

#include "list.h"

#include "map.h"



/* constructor - init empty map
**********************************************************/
CMap::CMap ()
{
   // reset "cache"
   m_dwLastPos = m_dwLastKey = 0;

   m_dwLessKey = 4294967295;    // maximum key value
}



/* destructor - empty map
**********************************************************/
CMap::~CMap ()
{
   mRemoveAll ();
}



/* Remove all keys and values
**********************************************************/
void CMap::mRemoveAll ()
{
   // reset "cache"
   m_dwLastPos = m_dwLastKey = 0;

   m_dwLessKey = 4294967295;    // maximum key value

   m_oKeys.mRemoveAll ();
   m_oValues.mRemoveAll ();
}



/* Remove the key (dwKey) and the value associated.
dDeleteValue - TRUE - desalocate value memory / FALSE - no (default TRUE)
return TRUE - node removed / FALSE - not removed, key not found
**********************************************************/
BOOL CMap::mRemoveKey ( DWORD dwKey, BOOL dDeleteValue )
{
   // define where is the node that will be removed, it's index
   DWORD
      dwPos = 0;
   CNodeInt
      *poKey = (CNodeInt*) m_oKeys [ dwPos ];
   do {
      // end of map, key not found
      if ( poKey == NULL )
         return FALSE;

      // verify if found the key
      if ( (DWORD) poKey->m_nValue == dwKey )
         break;

      // try next key
      dwPos ++;
      poKey = (CNodeInt*) poKey->m_poNext;

   } while ( 1 );

   // remove key and value
   delete (CNode*) m_oKeys.mRemoveNode ( dwPos );
   // if need desalocate value memory
   if ( dDeleteValue )
      delete (CNode*) m_oValues.mRemoveNode ( dwPos );
   else
      // only update list nodes, no delete
      m_oValues.mRemoveNode ( dwPos );

   // refresh direct access
   m_oKeys.mEnableDirectAccess ();
   m_oValues.mEnableDirectAccess ();

   // reset "cache"
   m_dwLastPos = m_dwLastKey = 0;

   // verify if removed the less key, if yes update less key
   if ( dwKey == m_dwLessKey )
      // if have some keys
      if ( m_oKeys.mRetNumNodes () != 0 )
         // less key is the first key
         m_dwLessKey = (DWORD) ( (CNodeInt*) m_oKeys.mRetStart ())->m_nValue;
      else
         m_dwLessKey = 4294967295;    // maximum key value

   return TRUE;
}



/* Return the key of a given position defined in dwKey
**********************************************************/
DWORD CMap::mRetKey ( DWORD dwIndex )
{
   return ( (CNodeInt*) m_oKeys.mRetNode ( dwIndex ) ) -> m_nValue;
}



/* Insert one key (dwKey) and the it's value poValue.
return TRUE - ok / FALSE - error (impossible alocate memory)
**********************************************************/
BOOL CMap::mAddKey ( DWORD dwKey, CNode* poValue )
{
   // alocate memory for key
   CNodeInt
      *poNewKey = new CNodeInt ( dwKey );
   // verify if alocation ok, if no return
   if ( poNewKey == NULL )
      return FALSE;

   // scan key list searching for best position to insert node
   // define position (make map be sorted by key)
   DWORD
      dwPos = 0;     // the position
   CNodeInt
      *poKeyScan = (CNodeInt*) m_oKeys [ dwPos ];
   do {
      // end of map, best position is end
      if ( poKeyScan == NULL )
         break;

      // Verify if found a mapped key greater than the given key,
      // if yes this will be the position for new key.
      if ( (DWORD) poKeyScan->m_nValue > dwKey )
         break;

      // try next key
      dwPos ++;
      poKeyScan = (CNodeInt*) poKeyScan->m_poNext;

   } while ( 1 );

   // insert nodes (node key and node value)
   m_oKeys.mSetNode ( dwPos, poNewKey );
   m_oValues.mSetNode ( dwPos, poValue );

   // refresh direct access
   m_oKeys.mEnableDirectAccess ();
   m_oValues.mEnableDirectAccess ();

   // update "cache"
   m_dwLastPos = dwPos;
   m_dwLastKey = dwKey;

   // verify if this key is the less, if yes update less key
   if ( dwKey < m_dwLessKey )
      m_dwLessKey = dwKey;

   return TRUE;
}



/* Return number of values in map
**********************************************************/
DWORD CMap::mRetSize ()
{
   return m_oValues.mRetNumNodes ();
}



/* Return the value associated of a key (dwKey).
return NULL if no key found or a pointer to value.
**********************************************************/
CNode* CMap::mRetValue ( DWORD dwKey )
{
   // search key by "cache"
   // verify if "cache" have the key, if yes do not need list search
   if ( m_dwLastPos != 0 && m_dwLastKey == dwKey ) //!=0 prevent error 1st
      return m_oValues.mRetNode ( m_dwLastPos );

   // define number of nodes in map
   DWORD
      dwNumNodes = m_oKeys.mRetNumNodes ();

   // verify if map is empty, if yes return
   if ( dwNumNodes == 0 )
      return NULL;

   // search key by "indexing"
   // verify is key is not greater than number of nodes
   DWORD
      dwKeyPos = dwKey - m_dwLessKey;
   if ( dwKeyPos < dwNumNodes )
      // verify if key in array is the same as the given key
      if ( dwKey == (DWORD) ( (CNodeInt*) m_oKeys.mRetNode ( dwKeyPos ) )->m_nValue ) {
         // update cache
         m_dwLastPos = dwKeyPos;
         m_dwLastKey = dwKey;
         // return pointer to value
         return m_oValues.mRetNode ( dwKeyPos );
      }

   // seach key position using binary tree search
   DWORD
      dwPosIni = 0,                // position of first key of segment
      dwPosMax = dwNumNodes - 1;   // position of last key of segment
   while ( dwPosIni <= dwPosMax ) {

      DWORD
         // position of actual key (in test): middle of segment
         dwPosCur = ( dwPosIni + dwPosMax ) >> 1, // >>1 equal /2, but faster
         // the current key in test
         dwKeyCur = (DWORD) ( (CNodeInt*) m_oKeys.mRetNode ( dwPosCur ) )->m_nValue;

      // verify if middle key and given key (dwKey) are equals: key found
      if ( dwKey == dwKeyCur ) {
         // update cache
         m_dwLastPos = dwPosCur;
         m_dwLastKey = dwKey;
         // return pointer to value
         return m_oValues.mRetNode ( dwPosCur );
      }
      else
         // if start and end of segment are the same return (no key found)
         if ( dwPosIni == dwPosMax )
            return NULL;
         else
            // split map up current position or down current position
            if ( dwKey > dwKeyCur )
               dwPosIni = dwPosCur + 1;
            else
               dwPosMax = dwPosCur - 1;
   }

   // no key found
   return NULL;
}



/* Return the value associated of a key (dwKey).
return NULL if no key found or pointer to value.
**********************************************************/
CNode* CMap::operator [] ( DWORD dwKey )
{
   return mRetValue ( dwKey );
}
