/* Math functions.

TIPS:
- this object need to much memory and time to compute the 
  internal lockup tables of sen/cos. Don't alocate the tables if
  you don't need and do not declare one object instance of this
  class inside a fast inner loop. */



#include <string.h>
#include <math.h>

#include "defs.h"

#include "node.h"

#include "nodexyz.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "map.h"

#include "math.h"

#include "refsys.h"

#include "mesh.h"



/* constructor - init math functions. If bBuildTrigTables
is TRUE build the sen/cos tables, else no (default FALSE)
**********************************************************/
CMath::CMath ( BOOL bBuildTrigTables )
{
   m_pafSin = m_pafCos = NULL;

   if ( ! mCreate ( bBuildTrigTables ) )
      ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
}



/* destructor - empty object, if need
**********************************************************/
CMath::~CMath ()
{
   mDestroy ();
}



/* Init math functions. If bBuildTrigTables
is TRUE build the sen/cos tables, else no (default FALSE).
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CMath::mCreate ( BOOL bBuildTrigTables )
{
   mDestroy ();
   
   // make the sen/cos tables
   if ( bBuildTrigTables ) {

      // verify if need alocate the lockup tables
      if ( m_pafSin == NULL ) {
         m_pafSin = new float [ ROUND ];
         if ( m_pafSin == NULL )
            return FALSE;
      }
      if ( m_pafCos == NULL ) {
         m_pafCos = new float [ ROUND ];
         if ( m_pafCos == NULL )
            return FALSE;
      }

      // sin and cos lockup tables
      for ( INT ncD = 0; ncD < ROUND; ncD ++ )  {
         m_pafSin [ ncD ] = sin ( ncD * PI / ROUND * 2 );
         m_pafCos [ ncD ] = cos ( ncD * PI / ROUND * 2 );
      }
   }

   return TRUE;
}



/* Desalocate lockup tables, if alocated
**********************************************************/
void CMath::mDestroy ()
{
   if ( m_pafSin != NULL ) {
      delete m_pafSin;
      m_pafSin = NULL;
   }
   if ( m_pafCos != NULL ) {
      delete m_pafCos;
      m_pafCos = NULL;
   }
}



/* Sort 3 values descenting.
fA, fB, fC - the values to sort
return the sorted values in the given variables
**********************************************************/
void CMath::mSort ( float &fA, float fB, float fC )
{
   float
      fTempA = fA,  // the less
      fTempB = fB,  // the midle
      fTempC = fC;  // the great

   if ( fA > fB )
      if ( fB > fC ) {
         fA = fTempC;
         fB = fTempB;
         fC = fTempA;
      }
      else
         if ( fA < fC ) {
            fA = fTempB;
            fB = fTempA;
            fC = fTempC;
         }
         else {
            fA = fTempB;
            fB = fTempC;
            fC = fTempA;
         }
   else
      if ( fA > fC ) {
         fA = fTempC;
         fB = fTempA;
         fC = fTempB;
      }
      else
         if ( fB < fC ) {
            fA = fTempA;
            fB = fTempB;
            fC = fTempC;
         }
         else {
            fA = fTempA;
            fB = fTempC;
            fC = fTempB;
         }
}



/* Return the maximal distance between two points.
The maximal distance is the major component (X or Y or Z) for
the vector between the given points.
stA, stB - the given points
**********************************************************/
float CMath::mRetMaximalDistance ( tagfXYZ& stA, tagfXYZ& stB )
{
   // find the vector between the points
   tagfXYZ
      stDist;
   mSubVector ( stA, stB, stDist );

   // make all components positive
   if ( stDist.fX < 0 )
      stDist.fX *= -1;
   if ( stDist.fY < 0 )
      stDist.fY *= -1;
   if ( stDist.fZ < 0 )
      stDist.fZ *= -1;

   // find the greatters component (x,y,z)
   if ( stDist.fX >= stDist.fY && stDist.fX >= stDist.fZ )
      return stDist.fX;
   if ( stDist.fY > stDist.fX && stDist.fY > stDist.fZ )
      return stDist.fY;
   return stDist.fZ;
}



/* Substract 2 vectors (make vector from 2 points): AB = B - A
**********************************************************/
void CMath::mSubVector ( tagfXYZ& stA, tagfXYZ& stB, tagfXYZ& stAnswer )
{
   stAnswer.fX = stB.fX - stA.fX;
   stAnswer.fY = stB.fY - stA.fY;
   stAnswer.fZ = stB.fZ - stA.fZ;
}



/* Normalize a given vector.
stVector - vector to normalize
return stVector - vector normalized
**********************************************************/
void CMath::mNormalize ( tagfXYZ& stVector )
{
   // normalize vector (make length 1)
   float
      fModulus = sqrt ( stVector.fX * stVector.fX + stVector.fY * stVector.fY + stVector.fZ * stVector.fZ );
   stVector.fX /= fModulus;
   stVector.fY /= fModulus;
   stVector.fZ /= fModulus;
}



/* Make the cross product between two vectors.
Stand math notation: V = V1 x V2
stVector1 and stVector2 - the given vectors
stAnswer - the result vector
**********************************************************/
void CMath::mCrossProduct ( tagfXYZ& stVector1, tagfXYZ& stVector2, tagfXYZ& stAnswer )
{
   stAnswer.fX = stVector1.fY * stVector2.fZ - stVector1.fZ * stVector2.fY;
   stAnswer.fY = stVector1.fZ * stVector2.fX - stVector1.fX * stVector2.fZ;
   stAnswer.fZ = stVector1.fX * stVector2.fY - stVector1.fY * stVector2.fX;
}



/* Make the dot product between two vectors.
Stand math notation: V = V1 * V2
stVector1 and stVector2 - the given vectors
return the dot product
**********************************************************/
float CMath::mDotProduct ( tagfXYZ& stVector1, tagfXYZ& stVector2 )
{
   return stVector1.fX * stVector2.fX +
          stVector1.fY * stVector2.fY +
          stVector1.fZ * stVector2.fZ;
}



/* Compute the distance between a point and a plan.
stPoint - point to test
stPlanNormal - the plan normal normalized vector
fPlanD - the plan D component.
The distance can be negative if the point is behind plan
(acording normal vector).
return TRUE - yes / FALSE - no
**********************************************************/
float CMath::mDistancePointPlan ( tagfXYZ& stPoint,
                                  tagfXYZ& stPlanNormal,
                                  float fPlanD )
{
   return stPlanNormal.fX * stPoint.fX +
          stPlanNormal.fY * stPoint.fY +
          stPlanNormal.fZ * stPoint.fZ -
          fPlanD;
}



/* Compute the poly plan equation:
First compute polygon normal vector and then the D component.
 Plan equation: AX + BY + CZ + D = 0,   where
 (A, B, C) is a plan normal and unitary vector
 D = AX' + BY' + CZ',   (X', Y', Z') is a point inside plan
 (X, Y, Z) is a point to test
oVertexes - is a list of x,y,z vertexes
poPol - point the poly inside the plan (change it's plan data)
**********************************************************/
void CMath::mDefinePolyPlanEquation ( CList& oVertexes, CNodePolygon* poPol )
{
   // verify if enought vertexes are defined in mesh
   if ( poPol->m_oPolygon.m_dwNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // define 3 vertexes inside polygon plan
   CNodeXYZ 
      *poVert0 = (CNodeXYZ*) oVertexes.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ 0 ] ),
      *poVert1 = (CNodeXYZ*) oVertexes.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ 1 ] ),
      *poVert2 = (CNodeXYZ*) oVertexes.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ 2 ] );

   // define 2 vectors inside the poly plan
   tagfXYZ
      stVector1,
      stVector2;
   stVector1.fX = poVert0->m_stXYZ.fX - poVert1->m_stXYZ.fX;
   stVector1.fY = poVert0->m_stXYZ.fY - poVert1->m_stXYZ.fY;
   stVector1.fZ = poVert0->m_stXYZ.fZ - poVert1->m_stXYZ.fZ;
   stVector2.fX = poVert0->m_stXYZ.fX - poVert2->m_stXYZ.fX;
   stVector2.fY = poVert0->m_stXYZ.fY - poVert2->m_stXYZ.fY;
   stVector2.fZ = poVert0->m_stXYZ.fZ - poVert2->m_stXYZ.fZ;

   // compute normal vector
   mCrossProduct ( stVector1, stVector2, poPol->m_oPolygon.m_stNormal );

   // normalize vector (make length 1)
   mNormalize ( poPol->m_oPolygon.m_stNormal );

   // compute plan D component
   poPol->m_oPolygon.m_fD = mDotProduct ( poPol->m_oPolygon.m_stNormal, poVert0->m_stXYZ );
}



/* Rotate a point - is inline - see the header file
**********************************************************/
void CMath::mRotate ( BYTE yAX, BYTE yAY, BYTE yAZ,
                      float& fX, float& fY, float& fZ )
{
   // rotated temp components
   float
      fXAux,
      fYAux,
      fZAux;

   fXAux = fX    * * ( m_pafCos + yAZ ) + fY    * * ( m_pafSin + yAZ );
   fYAux = fY    * * ( m_pafCos + yAZ ) - fX    * * ( m_pafSin + yAZ );
   fY    = fYAux * * ( m_pafCos + yAX ) + fZ    * * ( m_pafSin + yAX );
   fZAux = fZ    * * ( m_pafCos + yAX ) - fYAux * * ( m_pafSin + yAX );
   fX    = fXAux * * ( m_pafCos + yAY ) + fZAux * * ( m_pafSin + yAY );
   fZ    = fZAux * * ( m_pafCos + yAY ) - fXAux * * ( m_pafSin + yAY );
}



// REVIEW - does not work for all cases (?):
/* Verify if a point is inside or outside a given polygon.
stPoint - point to test
poPol - points the polygon
oVertexes - the list with polygon vertexes (x y z)
return TRUE - yes / FALSE - no
**********************************************************/
BOOL CMath::mPointInsidePoly ( tagfXYZ& stPoint,
                               CNodePolygon *poPol,
                               CList& oVertexes )
{
   // number of vertexes in poly
   DWORD
      dwNumVerts = poPol->m_oPolygon.m_dwNumVertexes;

   BOOL
      // flag if point is inside (TRUE) or outside poly (FALSE)
      bPointInside = TRUE,
      // flag dot-product sign (if all have the same sign the point is inside poly)
      bFlagDot = TRUE;       // TRUE - positive, FALSE - negative
   
   // point poly vertex index array
   DWORD
      *pdwVertNdx = poPol->m_oPolygon.m_padwVertexes;

   // vertexes to work with         
   CNodeXYZ
      *poVert1 = (CNodeXYZ*) oVertexes.mRetNode ( *pdwVertNdx ),
      *poVert2 = NULL;

   // for each vertex in poly make the test (scan all poly segments)
   for ( DWORD dwcV = 0; dwcV < dwNumVerts; dwcV ++ ) {

      // define next vertex
      if ( dwcV == dwNumVerts - 1 )  // if last index
         pdwVertNdx = poPol->m_oPolygon.m_padwVertexes;       // use first
      else     
         pdwVertNdx ++;                            // next
      // point it
      poVert2 = (CNodeXYZ*) oVertexes.mRetNode ( *pdwVertNdx );

      // define 2 vectors using the vertexes and the point
      tagfXYZ
         stVector1,
         stVector2;
      stVector1.fX = poVert1->m_stXYZ.fX - stPoint.fX;
      stVector1.fY = poVert1->m_stXYZ.fY - stPoint.fY;
      stVector1.fZ = poVert1->m_stXYZ.fZ - stPoint.fZ;
      stVector2.fX = poVert1->m_stXYZ.fX - poVert2->m_stXYZ.fX;
      stVector2.fY = poVert1->m_stXYZ.fY - poVert2->m_stXYZ.fY;
      stVector2.fZ = poVert1->m_stXYZ.fZ - poVert2->m_stXYZ.fZ;

      // compute dot-product
      float
         fDotProd = mDotProduct ( stVector1, stVector2 );

      // verify if is the first dot calculated (only update signal flags)
      if ( dwcV == 0 && fDotProd < 0 )
         bFlagDot = FALSE;
      else  // if not the first, make sign comparisons
         if ( ( bFlagDot == TRUE  && fDotProd < 0 ) ||
              ( bFlagDot == FALSE && fDotProd >= 0 ) ) {
            bPointInside = FALSE;
            break;
         }

      // update who is the first vertex
      poVert1 = poVert2;
   }

   return bPointInside;
}
   
   
   
/* Set the reference system base to a given view direction
stDirection - the view direction (set the Z vector)
oRefSys - the reference system to change
**********************************************************/
void CMath::mSetViewBase ( tagfXYZ& stDirection, CReferenceSystem& oRefSys )
{
   oRefSys.mResetBase ();

   // set J vector
   mCrossProduct ( oRefSys.m_stK, stDirection, oRefSys.m_stJ );

   // correct the K vector
   mCrossProduct ( stDirection, oRefSys.m_stJ, oRefSys.m_stK );

   // set the I vector
   oRefSys.m_stI.fX = stDirection.fX;
   oRefSys.m_stI.fY = stDirection.fY;
   oRefSys.m_stI.fZ = stDirection.fZ;
}   
