/* Store and manage a set of components.
Intances of this class implement scenery DEFINES and OBJECTS.

Read doc/3dfc.txt for more information about components.         */



#include <string.h>

#include "defs.h"

#include "mem.h"

#include "array.h"

#include "string.h"

#include "node.h"

#include "list.h"
       
#include "polygon.h"

#include "nodepol.h"

#include "nodeint.h"

#include "map.h"

#include "math.h"

#include "mesh.h"

#include "sector.h"

#include "refsys.h"

#include "camsys.h"

#include "nodesec.h"

#include "lighttab.h"

#include "eventtab.h"

#include "scenery.h"

#include "object.h"

#include "nodeobj.h"

#include "comp/nodecomp.h"

#include "comp/cokbcont.h"

#include "comp/codata.h"

#include "comp/cocamsys.h"

#include "comp/comesh.h"

#include "comp/colsomni.h"

#include "comp/cosecset.h"



/* constructor - init empty object
**********************************************************/
CObject::CObject ()
{
   m_dwMyID = 0;
   m_poScen = NULL;
   m_poMySec = NULL;
}



/* destructor - desalocate object data, if alocated
**********************************************************/
CObject::~CObject ()
{
   mDestroy ();
}



/* Create object - set variables, init empty object
dwMyID - object ID (uniq in all scenery)
poScen - scenery pointer
poMySec - sector where the object is inside
**********************************************************/
void CObject::mCreate ( DWORD dwMyID, CScenery* poScen, CNodeSector* poMySec )
{
   mDestroy ();

   m_dwMyID = dwMyID;
   m_poScen = poScen;
   m_poMySec = poMySec;
}



/* Desalocate object data, if alocated
**********************************************************/
void CObject::mDestroy ()
{
   m_oComponents.mRemoveAll ();

   m_poScen = NULL;
   m_dwMyID = 0;
   m_poMySec = NULL;
}



/* Init object with a copy of other. Does not reset the
actual object contents, only add more data.
oSrc - the source object
bInitAll - T - update the ID, scenery pointer and reference
               system,
           F - no. (default FALSE)
**********************************************************/
void CObject::mAppendObj ( CObject &oSrc, BOOL bInitAll )
{
   // update extra variables
   if ( bInitAll ) {
      m_dwMyID = oSrc.m_dwMyID;
      m_poMySec = oSrc.m_poMySec;
      m_poScen = oSrc.m_poScen;
   }

   // add components
   // for each component in object
   for ( DWORD dwcC = 0; dwcC < oSrc.m_oComponents.mRetSize (); dwcC ++ ) {

      // get the slot with data
      DWORD
         dwSlotID = oSrc.m_oComponents.mRetKey ( dwcC );
      // get the component
      CNodeComponent
         *poSrcComp = (CNodeComponent*) oSrc.m_oComponents.mRetValue ( dwSlotID );

      // copy component
      switch ( poSrcComp->m_dwMyTypeID ) {

         case TYPE_UNKNOWN:
            mCreateComponent ( dwSlotID, poSrcComp->m_dwMyTypeID, 0, 0, 0 );
            break;

         case TYPE_KBCONTROL:
            mCreateComponent ( dwSlotID, poSrcComp->m_dwMyTypeID, 0, 0, 0 );
            break;

         case TYPE_DATA:
            mCreateComponent ( dwSlotID, ( (CComponentData*) poSrcComp )->m_dwMyTypeID,
                                         ( (CComponentData*) poSrcComp )->m_dwCastID,
                                         ( (CComponentData*) poSrcComp )->m_dwNumElements,
                                         0 );
            break;

         case TYPE_CAMSYS:
            {
               // create with trig tables
               mCreateComponent ( dwSlotID, poSrcComp->m_dwMyTypeID, 0, 1, 0 );
               CCameraSystem
                  *poDestCam = &( (CComponentCameraSystem*) mRetComponent ( dwSlotID ) )->m_oCamSys;
               poDestCam->mSetObj ( ( (CComponentCameraSystem*) poSrcComp )->m_oCamSys );
            }
            break;

         case TYPE_MESH:
            {
               mCreateComponent ( dwSlotID, poSrcComp->m_dwMyTypeID, 0, 0, 0 );
               CMesh
                  *poDestMesh = &( (CComponentMesh*) mRetComponent ( dwSlotID ) )->m_oMesh;
               poDestMesh->mBeginAddMesh ();
               poDestMesh->mAppendObj ( ( (CComponentMesh*) poSrcComp )->m_oMesh );
               poDestMesh->mEndAddMesh ();
            }
            break;

         case TYPE_SECTORSSET:
            {
               mCreateComponent ( dwSlotID, poSrcComp->m_dwMyTypeID, 0, 0, 0 );
               mActivateComponent ( dwSlotID, 1, (PTRVAL) &( (CComponentSectorsSet*) poSrcComp )->m_oSectors, 0 );
               mActivateComponent ( dwSlotID, 3, 0, 0 );
            }
            break;

         case TYPE_LIGHT_SOURCE_OMNI:
            {
               mCreateComponent ( dwSlotID, poSrcComp->m_dwMyTypeID, 0, 0, 0 );
               mActivateComponent ( dwSlotID, 1, (PTRVAL) poSrcComp, 0 );
            }
            break;
      }
   }
}



/* Add one component to the object.
dwSlotID - object slot to fill with component
dwTypeID - component type to add.
pData1, pData2, pData3 - data passed to component (the use is acording
 the specific component).
If the component type is unknown return FALSE.
Return value depend acording the component.
**********************************************************/
PTRVAL CObject::mCreateComponent ( DWORD dwSlotID, DWORD dwTypeID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   // acording the type create the specific component
   CNodeComponent
      *poComp = NULL;
   switch ( dwTypeID ) {

      case TYPE_LET_SLOT_EMPTY:  // just clear the slot
         return mDestroyComponent ( dwSlotID, pData1, pData2, pData3 );
         break;

      case TYPE_UNKNOWN:
         poComp = (CNodeComponent*) new CNodeComponent;
         break;

      case TYPE_KBCONTROL:
         poComp = (CNodeComponent*) new CComponentKbControl;
         break;

      case TYPE_DATA:
         poComp = (CNodeComponent*) new CComponentData;
         break;

      case TYPE_CAMSYS:
         poComp = (CNodeComponent*) new CComponentCameraSystem;
         break;

      case TYPE_MESH:
         poComp = (CNodeComponent*) new CComponentMesh;
         break;

      case TYPE_SECTORSSET:
         poComp = (CNodeComponent*) new CComponentSectorsSet;
         break;

      case TYPE_LIGHT_SOURCE_OMNI:
         poComp = (CNodeComponent*) new CComponentLightSourceOmni;
         break;

      default:
         return FALSE;
         break;
   }

   // verify success when alocating component
   if ( poComp == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;         // slot mantain old component
   }

   // REVIEW super bug mDestroyComponent ( dwSlotID, pData1, pData2, pData3 );

   // update component with default values
   ( (CNodeComponent*) poComp )->m_poMyObj = this;
   ( (CNodeComponent*) poComp )->m_dwMySlotID = dwSlotID;

   // update object adding the component
   if ( ! m_oComponents.mAddKey ( dwSlotID, (CNode*) poComp ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // call component initialization
   PTRVAL
      pRet = TRUE;
   switch ( dwTypeID ) {

      case TYPE_UNKNOWN:
         pRet = ( (CNodeComponent*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;

      case TYPE_KBCONTROL:
         pRet = ( (CComponentKbControl*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;

      case TYPE_DATA:
         pRet = ( (CComponentData*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;

      case TYPE_CAMSYS:
         pRet = ( (CComponentCameraSystem*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;

      case TYPE_MESH:
         pRet = ( (CComponentMesh*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;

      case TYPE_SECTORSSET:
         pRet = ( (CComponentSectorsSet*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;

      case TYPE_LIGHT_SOURCE_OMNI:
         pRet = ( (CComponentLightSourceOmni*) poComp )->mCreate ( pData1, pData2, pData3 );
         break;
   }

   return pRet;
}



/* Remove the component of the given type from the object.
dwSlotID - object slot where is the component to be removed
pData1, pData2, pData3 - data passed to component (the use is acording
 the specific component).
Don't forget to call mBeginAddObject and mEndAddObject!!
Return value depend acording the component.
**********************************************************/
PTRVAL CObject::mDestroyComponent ( DWORD dwSlotID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   CNodeComponent
      *poComp = (CNodeComponent*) m_oComponents.mRetValue ( dwSlotID );

   // if slot is empty
   if ( poComp == NULL )
      return FALSE;

   // destroy component
   PTRVAL
      pRet = TRUE;
   switch ( poComp->m_dwMyTypeID ) {

      case TYPE_UNKNOWN:
         pRet = ( (CNodeComponent*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;

      case TYPE_KBCONTROL:
	 pRet = ( (CComponentKbControl*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;

      case TYPE_DATA:
         pRet = ( (CComponentData*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;

      case TYPE_CAMSYS:
         pRet = ( (CComponentCameraSystem*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;

      case TYPE_MESH:
         pRet = ( (CComponentMesh*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;

      case TYPE_SECTORSSET:
         pRet = ( (CComponentSectorsSet*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;

      case TYPE_LIGHT_SOURCE_OMNI:
         pRet = ( (CComponentLightSourceOmni*) poComp )->mDestroy ( pData1, pData2, pData3 );
         break;
   }
   
   // desalocate
   m_oComponents.mRemoveKey ( dwSlotID );

   return pRet;
}



/* Execute one specific object component.
dwSlotID - slot where component to run is
pData1, pData2, pData3 - data passed to component (the use is acording
 the specific component)
Return value depend acording the component.
**********************************************************/
PTRVAL CObject::mActivateComponent ( DWORD dwSlotID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   // the component
   CNodeComponent
      *poComp = (CNodeComponent*) m_oComponents.mRetValue ( dwSlotID );

   // slot empty
   if ( poComp == NULL )
      return FALSE;

   // activate component
   PTRVAL
      pRet = TRUE;
   switch ( poComp->m_dwMyTypeID ) {

      case TYPE_UNKNOWN:
         pRet = ( (CNodeComponent*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;

      case TYPE_KBCONTROL:
         pRet = ( (CComponentKbControl*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;

      case TYPE_DATA:
         pRet = ( (CComponentData*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;

      case TYPE_CAMSYS:
         pRet = ( (CComponentCameraSystem*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;

      case TYPE_MESH:
         pRet = ( (CComponentMesh*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;

      case TYPE_SECTORSSET:
         pRet = ( (CComponentSectorsSet*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;

      case TYPE_LIGHT_SOURCE_OMNI:
         pRet = ( (CComponentLightSourceOmni*) poComp )->mActivate ( pData1, pData2, pData3 );
         break;
   }

   return pRet;
}



/* Return a pointer to the component in a given slot.
return CNodeComponent pointer or NULL if slot empty
**********************************************************/
CNodeComponent* CObject::mRetComponent ( DWORD dwSlotID )
{
   PTRVAL
      pComp = (PTRVAL) m_oComponents.mRetValue ( dwSlotID );
   if ( pComp == 0 )
      return NULL;
   
   return (CNodeComponent*) pComp;
}
