/* Implements a 3D reference system with some transformations.
The reference system have one origin and 3 vectors (I, J, K) with module 1.
Example:
             J
            |
            |  / K
            | /
            |/
            *------
             O     I
*/



#include <string.h>

#include <math.h>

#include "defs.h"

#include "node.h"

#include "nodexyz.h"

#include "nodexy.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "math.h"

#include "refsys.h"



/* constructor - init default reference system
bBuildTrigTables - make lookup sin/cos tables (default FALSE)
**********************************************************/
CReferenceSystem::CReferenceSystem ( BOOL bBuildTrigTables )
{
   mCreate ( bBuildTrigTables );
}



/* destructor
**********************************************************/
CReferenceSystem::~CReferenceSystem ()
{
}



/* Init reference system and, if bBuildTrigTables is TRUE
create sin/cos lookup tables
**********************************************************/
void CReferenceSystem::mCreate ( BOOL bBuildTrigTables )
{
   // create lockup tables
   m_oM.mCreate ( bBuildTrigTables );
   
   mResetPos ();
   mResetBase ();
}



/* Destroy object
**********************************************************/
void CReferenceSystem::mDestroy ()
{
}



/* Set position to default value: space origin
**********************************************************/
void CReferenceSystem::mResetPos ()
{
   m_stPos.fX = m_stPos.fY = m_stPos.fZ = 0;
}



/* Set reference system base to default value.
**********************************************************/
void CReferenceSystem::mResetBase ()
{
   // set I
   m_stI.fX = 1;
   m_stI.fY = 0;
   m_stI.fZ = 0;
   // set J
   m_stJ.fX = 0;
   m_stJ.fY = 1;
   m_stJ.fZ = 0;
   // set K
   m_stK.fX = 0;
   m_stK.fY = 0;
   m_stK.fZ = -1;
}



/* Init reference system with a copy of other object
(like CReferenceSystem::constructor and CReferenceSystem::operator =)
**********************************************************/
void CReferenceSystem::mSetObj ( CReferenceSystem& oRS )
{
   // set position
   m_stPos.fX = oRS.m_stPos.fX;
   m_stPos.fY = oRS.m_stPos.fY;
   m_stPos.fZ = oRS.m_stPos.fZ;
   // set I
   m_stI.fX = oRS.m_stI.fX;
   m_stI.fY = oRS.m_stI.fY;
   m_stI.fZ = oRS.m_stI.fZ;
   // set J
   m_stJ.fX = oRS.m_stJ.fX;
   m_stJ.fY = oRS.m_stJ.fY;
   m_stJ.fZ = oRS.m_stJ.fZ;
   // set K
   m_stK.fX = oRS.m_stK.fX;
   m_stK.fY = oRS.m_stK.fY;
   m_stK.fZ = oRS.m_stK.fZ;
}



/* like CReferenceSystem::constructor and CReferenceSystem::mSetObj
**********************************************************/
void CReferenceSystem::operator = ( CReferenceSystem& oRS )
{
   mSetObj ( oRS );
}



/* Transform this system acording the given system.
Update camera position and orientation parameters.
**********************************************************/
void CReferenceSystem::mTransform ( CReferenceSystem& oR )
{
   // transform position
   m_stPos.fX += oR.m_stPos.fX;
   m_stPos.fY += oR.m_stPos.fY;
   m_stPos.fZ += oR.m_stPos.fZ;

   // transform I
   float
      fXa = m_stI.fX * oR.m_stI.fX + m_stI.fY * oR.m_stI.fX + m_stI.fZ * oR.m_stI.fX,
      fYa = m_stI.fX * oR.m_stI.fY + m_stI.fY * oR.m_stI.fY + m_stI.fZ * oR.m_stI.fY,
      fZa = m_stI.fX * oR.m_stI.fZ + m_stI.fY * oR.m_stI.fZ + m_stI.fZ * oR.m_stI.fZ;
   m_stI.fX = fXa;
   m_stI.fY = fYa;
   m_stI.fZ = fZa;
   // transform J
   fXa = m_stJ.fX * oR.m_stJ.fX + m_stJ.fY * oR.m_stJ.fX + m_stJ.fZ * oR.m_stJ.fX;
   fYa = m_stJ.fX * oR.m_stJ.fY + m_stJ.fY * oR.m_stJ.fY + m_stJ.fZ * oR.m_stJ.fY;
   fZa = m_stJ.fX * oR.m_stJ.fZ + m_stJ.fY * oR.m_stJ.fZ + m_stJ.fZ * oR.m_stJ.fZ;
   m_stJ.fX = fXa;
   m_stJ.fY = fYa;
   m_stJ.fZ = fZa;
   // transform K
   fXa = m_stK.fX * oR.m_stK.fX + m_stK.fY * oR.m_stK.fX + m_stK.fZ * oR.m_stK.fX;
   fYa = m_stK.fX * oR.m_stK.fY + m_stK.fY * oR.m_stK.fY + m_stK.fZ * oR.m_stK.fY;
   fZa = m_stK.fX * oR.m_stK.fZ + m_stK.fY * oR.m_stK.fZ + m_stK.fZ * oR.m_stK.fZ;
   m_stK.fX = fXa;
   m_stK.fY = fYa;
   m_stK.fZ = fZa;
}



/* Apply the reference system base to a point stPoint
**********************************************************/
void CReferenceSystem::mApplyBase ( tagfXYZ& stPoint )
{
   float
      fXa = stPoint.fX * m_stI.fX + stPoint.fY * m_stI.fY + stPoint.fZ * m_stI.fZ,
      fYa = stPoint.fX * m_stJ.fX + stPoint.fY * m_stJ.fY + stPoint.fZ * m_stJ.fZ,
      fZa = stPoint.fX * m_stK.fX + stPoint.fY * m_stK.fY + stPoint.fZ * m_stK.fZ;
   stPoint.fX = fXa;
   stPoint.fY = fYa;
   stPoint.fZ = fZa;
}



/* Rotate reference system base using the transpose matrix.
(make Descent like behavior)
**********************************************************/
void CReferenceSystem::mRotateTransBase ( BYTE yAX, BYTE yAY, BYTE yAZ )
{
   m_oM.mRotate ( yAX, yAY, yAZ, m_stI.fX, m_stJ.fX, m_stK.fX );
   m_oM.mRotate ( yAX, yAY, yAZ, m_stI.fY, m_stJ.fY, m_stK.fY );
   m_oM.mRotate ( yAX, yAY, yAZ, m_stI.fZ, m_stJ.fZ, m_stK.fZ );
}



/* Rotate reference system base
**********************************************************/
void CReferenceSystem::mRotateBase ( BYTE yAX, BYTE yAY, BYTE yAZ )
{
   m_oM.mRotate ( yAX, yAY, yAZ, m_stI.fX, m_stI.fY, m_stI.fZ );
   m_oM.mRotate ( yAX, yAY, yAZ, m_stJ.fX, m_stJ.fY, m_stJ.fZ );
   m_oM.mRotate ( yAX, yAY, yAZ, m_stK.fX, m_stK.fY, m_stK.fZ );
}
