/* Implement a portal engine core for scenery indoor renderization.

Has the main visualization pipeline: make transformations from universe to
camera reference system, make poly clip, draw polygon mesh on frame buffer,
call recursively functions to draw other visible objects and sectors.

Portal is a polygon where other polygons are clipped, used to define visible
sectors from where the camera be.

Please refer doc/notes.txt for more info about the view pipeline and
engine details.

The main function is mRenderMesh. It call:
   For each mesh
      mApplyCamera - transform SRU -> SRC (and scale mesh)
      mAjust - project vertexes
      For each poly:
         mDefinePolygonData - set poly vertexes, light info,...
         mPassBackFaceCulling - try back face culling
         mGetPolyTextureData - get texture for the poly
         mMakeZNearPlaneClip  - clip poly to near Z
and draw a poly mesh.

WARNING:   
- polygons vertexes must be in anticlockwise order
*/



#include <string.h>

#include <math.h>

#include "defs.h"

#include "mem.h"

#include "array.h"

#include "string.h"

#include "texture.h"

#include "file.h"

#include "filerem.h"

#include "filepcx.h"

#include "lighttab.h"

#include "draw.h"

#include "node.h"

#include "nodelii.h"

#include "list.h"

#include "polygon.h"

#include "nodeflo.h"

#include "nodepol.h"

#include "nodexyz.h"

#include "nodexy.h"

#include "map.h"

#include "math.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"

#include "refsys.h"

#include "camsys.h"

#include "object.h"

#include "nodeint.h"

#include "nodetex.h"

#include "nodeobj.h"

#include "nodelgt.h"

#include "filetxe.h"

#include "eventtab.h"

#include "scenery.h"

#include "render.h"

#include "comp/nodecomp.h"

#include "comp/cokbcont.h"

#include "comp/codata.h"

#include "comp/cocamsys.h"

#include "comp/comesh.h"



/* constructor - init empty
**********************************************************/
CRender::CRender ()
{
   m_pastPolVertexCoords = NULL;
   m_dwMaxVertArray = 0;
   m_pastPolZClipTex = m_pastPolZClipCoords = NULL;
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   m_pafPolVertexLights = NULL;
   m_pafPolZClipLight = NULL;
   #endif
   m_dwMaxZClipArray = 0;

   mDestroy ();
}



/* constructor - init fly system
**********************************************************/
CRender::CRender ( CScenery& oScen, INT nResX, INT nResY, CMem& oFrBuff )
{
   m_pastPolVertexCoords = NULL;
   m_dwMaxVertArray = 0;
   m_pastPolZClipTex = m_pastPolZClipCoords = NULL;
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   m_pafPolVertexLights = NULL;
   m_pafPolZClipLight = NULL;
   #endif
   m_dwMaxZClipArray = 0;

   mCreate ( oScen, nResX, nResY, oFrBuff );
}



/* destructor
**********************************************************/
CRender::~CRender ()
{
   mDestroy ();
}



/* Init render engine.
oScen define the scenery to render
nResX define the X resolution to frame buffer
nResY define Y resolution to frame buffer
oFrBuffer define the frame buffer (where draw)
oCamSys the camera reference system (camera status in space)
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CRender::mCreate ( CScenery& oScen, INT nResX, INT nResY, CMem& oFrBuff )
{
   mDestroy ();

   m_poScen = &oScen;      // define scenery pointer

   // REVIEW
   // define texture vertexes, used in textures of size 256x256
   // polys with 3 vertexes
   m_astTexVertexes3 [ 0 ].nX = 0;
   m_astTexVertexes3 [ 0 ].nY = 0;
   m_astTexVertexes3 [ 1 ].nX = 0;
   m_astTexVertexes3 [ 1 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes3 [ 2 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes3 [ 2 ].nY = 0;
   // polys with 4 vertexes
   m_astTexVertexes4 [ 0 ].nX = 0;
   m_astTexVertexes4 [ 0 ].nY = 0;
   m_astTexVertexes4 [ 1 ].nX = 0;
   m_astTexVertexes4 [ 1 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 2 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 3 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 3 ].nY = 0;
   // polys with 5 vertexes
   m_astTexVertexes5 [ 0 ].nX = 0;
   m_astTexVertexes5 [ 0 ].nY = 0;
   m_astTexVertexes5 [ 1 ].nX = 0;
   m_astTexVertexes5 [ 1 ].nY = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes5 [ 2 ].nX = TEXTURE_MAXDIM / 2;
   m_astTexVertexes5 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 3 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 3 ].nY = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes5 [ 4 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 4 ].nY = 0;
   // polys with 6 vertexes
   m_astTexVertexes6 [ 0 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 1 / 4 );
   m_astTexVertexes6 [ 0 ].nY = 0;
   m_astTexVertexes6 [ 1 ].nX = 0;
   m_astTexVertexes6 [ 1 ].nY = TEXTURE_MAXDIM / 2;
   m_astTexVertexes6 [ 2 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 1 / 4 );
   m_astTexVertexes6 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 3 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes6 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 4 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 4 ].nY = TEXTURE_MAXDIM / 2;
   m_astTexVertexes6 [ 5 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes6 [ 5 ].nY = 0;
   // polys with 7 vertexes (not very precise)
   m_astTexVertexes7 [ 0 ].nX = TEXTURE_MAXDIM / 2;
   m_astTexVertexes7 [ 0 ].nY = 0;
   m_astTexVertexes7 [ 1 ].nX = 0;
   m_astTexVertexes7 [ 1 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes7 [ 2 ].nX = 0;
   m_astTexVertexes7 [ 2 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 3 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes7 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 4 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 4 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 5 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 5 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 6 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 6 ].nY = TEXTURE_MAXDIM / 3;
   // polys with 8 vertexes
   m_astTexVertexes8 [ 0 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 0 ].nY = 0;
   m_astTexVertexes8 [ 1 ].nX = 0;
   m_astTexVertexes8 [ 1 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 2 ].nX = 0;
   m_astTexVertexes8 [ 2 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 3 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 4 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 4 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 5 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 5 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 6 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 6 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 7 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 7 ].nY = 0;

   // verify Y resolution limitation
   if ( nResY > MAX_RESY )
      ABORT ( MSG_ERROR_OVER_RESY_LIMIT, __FILE__, __LINE__ )

   CLightTable // empty light table
      oLgT;
   return m_oDraw.mCreate ( oLgT, nResX, nResY, oFrBuff );
}



/* Destroy: make pointers point to NULL, free memory.
**********************************************************/
void CRender::mDestroy ()
{
   m_poScen = NULL;

   m_dwMaxVertArray = 0;
   m_dwMaxZClipArray = 0;

   // free array memory
   if ( m_pastPolVertexCoords != NULL ) {
      delete m_pastPolVertexCoords;
      m_pastPolVertexCoords = NULL;
   }

   // free array memory for clipped vertexes
   if ( m_pastPolZClipCoords != NULL ) {
      delete m_pastPolZClipCoords;
      m_pastPolZClipCoords = NULL;
   }

   // free array memory for clipped vertexes
   if ( m_pastPolZClipTex != NULL ) {
      delete m_pastPolZClipTex;
      m_pastPolZClipTex = NULL;
   }

   #ifdef COMP_RENDER_INTERPOLATED_LIGHT

   // free array with light levels
   if ( m_pafPolVertexLights != NULL ) {
      delete m_pafPolVertexLights;
      m_pafPolVertexLights = NULL;
   }

   // free array memory for clipped light information
   if ( m_pafPolZClipLight != NULL ) {
      delete m_pafPolZClipLight;
      m_pafPolZClipLight = NULL;
   }

   #endif

   m_oCamVert.mRemoveAll ();
   m_oPrjVert.mRemoveAll ();

   m_oDraw.mDestroy ();
}



/* Make the scenery image in frame buffer. Call the sector
and object render functions.
poCamera point the camera object to be used in visualization.
**********************************************************/
void CRender::mRenderFrame ( CNodeObject* poCamera )
{
   // the initial clip edge buffer
   tagInterval
      astClipEdgeBuffer [ MAX_RESY ];
   m_oDraw.mResetClipEdgeBuffer ( astClipEdgeBuffer );

   // the initial clip bounding rectangle
   tagRectangle
      stClipBox;
   stClipBox.nStartX = 0;
   stClipBox.nEndX = m_oDraw.m_nMaxX;
   stClipBox.nStartY = 0;
   stClipBox.nEndY = m_oDraw.m_nMaxY;

   // clear the frame if is the case
   if ( m_poScen->m_nClearBackground != -1 )
      m_oDraw.mClear ( m_poScen->m_nClearBackground );

   // REVIEW, eh necessario?
   // define the camera system to be used in this object
   /*CCameraSystem
      oCamObj;
   // update this camera with the viewer camera, if is not the viewer
   if ( dwObjectID != m_poScen->m_dwCameraID ) {
      // set the camera as the object parameters
      oCamObj.mSetObj ( ( (CNodeObject*) m_poScen->m_oRoot.m_oObjects.mRetValue ( dwObjectID ) )->m_oRefSys );
      oCamObj.mTransform ( *poCamSys );
   }*/

   // get the reference system
   CCameraSystem
      *poCamSys = &( (CComponentCameraSystem*) poCamera->m_oObject.mRetComponent ( SLOT_REFSYS ) )->m_oCamSys;
   // make frame buffer
   mRenderSector ( poCamera->m_oObject.m_poMySec, poCamSys, 0,
                   astClipEdgeBuffer, stClipBox );
}



/* Sector rendering: render sector mesh and sector objects.
poSector - the sector to render
poCamSys - define the camera to be used in view transformations
yLevel the actual recursion level
pastClipEdgeBuffer have the current clip buffer to be used in poly draw,
stClipBox  the clip square (if all vertexes of the poly are outside, no draw
**********************************************************/
void CRender::mRenderSector ( CNodeSector *poSector,
                              CCameraSystem *poCamSys, BYTE yLevel,
                              tagInterval* pastClipEdgeBuffer,
                              tagRectangle& stClipBox )
{
   // REVIEW
   // loop bug (normal vector remove error)
   if ( yLevel > 4 )
      return;

   // verify if sector not found
   if ( poSector == NULL )
      ABORT ( MSG_ERROR_SECTOR_NOT_FOUND, __FILE__, __LINE__ )

   mRenderMesh ( poSector->m_oSector, poCamSys,
                 pastClipEdgeBuffer, stClipBox, yLevel );

   // search for objects to render (objects inside this sector)
   DWORD
      dwNumObjects = poSector->m_oSector.m_oObjects.mRetSize ();
   // for each object in sector
   for ( DWORD dwcO = 0; dwcO < dwNumObjects; dwcO ++ ) {
      
      // the object ID
      DWORD
         dwObjID = poSector->m_oSector.m_oObjects.mRetKey ( dwcO );
      // the object inside this sector
      CNodeObject
         *poObj = (CNodeObject*) poSector->m_oSector.m_oObjects.mRetValue ( dwObjID );

      // define the camera system to be used in this object
      CCameraSystem
         oCamObj;
      // update this camera with the viewer camera, if is not the viewer
      if ( poObj != m_poScen->m_poCamera ) {
         // get the reference system
         CCameraSystem
            *poObjCamSys = &( (CComponentCameraSystem*) poObj->m_oObject.mRetComponent ( SLOT_REFSYS ) )->m_oCamSys;
         // set the camera as the object parameters
         oCamObj.mSetObj ( *poObjCamSys );
         oCamObj.mTransform ( *poCamSys );
      }

      // get the object body
      CComponentMesh
         *poBody = (CComponentMesh*) poObj->m_oObject.mRetComponent ( SLOT_BODY );

      // verify if it has a body to be rendered
      if ( poBody != NULL ) {

         // get the body mesh
         CMesh
           *poMesh = &( (CComponentMesh*) poObj->m_oObject.mRetComponent ( SLOT_BODY ) )->m_oMesh;

         mRenderMesh ( *poMesh, &oCamObj,
                       pastClipEdgeBuffer, stClipBox, yLevel );
      }
   }
}



/* Draw a poly mesh: transform to SRC, project and
ajust to frame buffer coordenates, draw the polygons.
If a portal is visible call recursively the draw function for
the respective sector or object.
**********************************************************/
void CRender::mRenderMesh ( CMesh& oMesh, CCameraSystem *poCamSys,
                            tagInterval* pastClipEdgeBuffer,
                            tagRectangle& stClipBox, BYTE yLevel )
{
   // transform sector mesh from universe to camera
   mApplyCamera ( poCamSys, oMesh.m_oVertexes );

   // project mesh to 2D and ajust to viewport reference system
   mAjust ( poCamSys );

   // draw poly mesh making portal clip
   DWORD
      dwPolyNdx = 0;  // the index of current poly beeing draw
   // for each polygon in mesh
   oMesh.m_oPolygons.mSelectStart ();
   while ( oMesh.m_oPolygons.mRetExistSelection () ) {

      // the current polygon to define
      CNodePolygon
         *poPol = (CNodePolygon*) oMesh.m_oPolygons.mRetSelected ();

      // define poly vertex coordenates projected, make frame buffer clip
      // if some vertex is inside the frame buffer -> draw poly, else no
      DWORD
         dwBehind;    // count number of vertexes behind near Z clip plane
      if ( mDefinePolygonData ( poPol, oMesh.m_oVertLight, stClipBox, dwBehind ) )

         // make normal vector visible cut
         if ( mPassBackFaceCulling () ) {

            // define polygon texture and texture coodenates
            CTexture
               *poTx = NULL;
            tagXY
               *pastTexVertexes = NULL;
            mGetPolyTextureData ( poPol, &poTx, &pastTexVertexes ); 

            // point the poly data to be drawn
            tagXY
               *pstPolVerDraw = NULL,  // the vertexes in frame buffer
               *pstPolTexDraw = NULL;  // the texture vertexes in texture
            #ifdef COMP_RENDER_INTERPOLATED_LIGHT
            float
               *pfPolLightDraw = NULL; // the vertexes light levels
            #endif
            DWORD
               dwPolNumVerDraw;        // the number of vertexes

            // define if the polygon is clipped in Z plan
            // verify if poly not intercect the Z plan
            if ( dwBehind == 0 ) {

               // define the polygon to draw as the current polygon
               pstPolVerDraw = m_pastPolVertexCoords;
               dwPolNumVerDraw = poPol->m_oPolygon.m_dwNumVertexes;

               // if texture defined
               if ( poTx != NULL )
                  pstPolTexDraw = pastTexVertexes;

               // define light levels (if compiled)
               #ifdef COMP_RENDER_INTERPOLATED_LIGHT
               pfPolLightDraw = m_pafPolVertexLights;
               #endif
            }
            else { // if the poly intersect the clip Z plan make poly cut

               dwPolNumVerDraw = mMakeZNearPlaneClip ( poPol, pastTexVertexes, poCamSys, pastClipEdgeBuffer, stClipBox );

               // define the polygon to draw as the current polygon
               pstPolVerDraw = m_pastPolZClipCoords;

               // if texture defined
               if ( poTx != NULL )
                  pstPolTexDraw = m_pastPolZClipTex;

               // define light levels (if compiled)
               #ifdef COMP_RENDER_INTERPOLATED_LIGHT
               pfPolLightDraw = m_pafPolZClipLight;
               #endif
            }

            // verify if the polygon is a portal
            if ( poPol->m_oPolygon.m_stPortal.yCode != 0 ) {

               // if yes make the new clip edge buffer
               // have start and end X limits per scanline (the edge buffer)
               tagInterval
                  astNewBuffer [ MAX_RESY ];
               // copy the current buffer to new (will be used to define new)
               memcpy ( (void*) astNewBuffer, (void*) pastClipEdgeBuffer, sizeof ( tagInterval ) * m_oDraw.m_nResY );
               // define new clip box (rectangle around clip poly)
               tagRectangle
                  stNewBox;
               stNewBox.nStartX = stClipBox.nStartX;
               stNewBox.nStartY = stClipBox.nStartY;
               stNewBox.nEndX = stClipBox.nEndX;
               stNewBox.nEndY = stClipBox.nEndY;

               // verify if need draw a new sector
               // Set the new clip edge buffer using the clip polygon (portal)
               // if the clip polygon is totaly clipped (return FALSE) the destin
               // sector isnt visible and to not call sector draw, recursion end
               if ( m_oDraw.mSetClipEdgeBuffer ( pstPolVerDraw, dwPolNumVerDraw, astNewBuffer, stNewBox ) ) {

                  // the sector is visible:  call render sector again
                  mRenderSector ( poPol->m_oPolygon.m_stPortal.poSector, poCamSys, yLevel + 1, astNewBuffer, stNewBox );

                  // REVIEW (guardar numa "cache")
                  oMesh.m_oPolygons.mSelectNode ( dwPolyNdx );
                  // transform sector mesh from universe to camera
                  mApplyCamera ( poCamSys, oMesh.m_oVertexes );
                  // project mesh to 2D
                  mAjust ( poCamSys );
               }
            }
            else // if poly is not a portal draw texture (bug: "global" projected
                 // member variables are changes after other sectors draw->
                 // impossible draw a textured portal now)
               // draw textured polygon
               // if texture defined
               if ( poTx != NULL ) {

                  // set the light table to be used in poly draw
                  CNodeLightTable
                     *poNodeLgT = (CNodeLightTable*) m_poScen->m_oLightTables.mRetValue ( poPol->m_oPolygon.m_dwLightID );
                  // verify if not found
                  if ( poNodeLgT == NULL )
                     ABORT ( MSG_ERROR_NO_LIGHTTABLE, __FILE__, __LINE__ )
                  m_oDraw.m_poLightTable = &poNodeLgT->m_oLgT;

                  #ifdef COMP_RENDER_INTERPOLATED_LIGHT
                  m_oDraw.mIntLgtTexturedPolygon ( pstPolVerDraw, pstPolTexDraw,
                                                   pfPolLightDraw, dwPolNumVerDraw,
                                                   *poTx, pastClipEdgeBuffer );
                  #else
                  // draw it with flat textured shading
                  m_oDraw.mTexturedPolygon ( pstPolVerDraw, pstPolTexDraw, dwPolNumVerDraw,
                                             *poTx, pastClipEdgeBuffer, poPol->m_oPolygon.m_fLight );
                  #endif
                  //m_oDraw.mFlatPolygon ( pstPolVerDraw, dwPolNumVerDraw, 168+poPol->m_oPolygon.m_dwTextureID );
               }
         }

      // next polygon to draw
      oMesh.m_oPolygons.mSelectNext ();
      dwPolyNdx ++;
   }
}



/* Make universe reference system to view reference system 
transformation. Apply the reference system base and origin in oSrc
to CRender::m_oCamVert
The given list must have CNodeXYZ nodes.
poCamSys is the camera to be used in transformation.
**********************************************************/
void CRender::mApplyCamera ( CCameraSystem *poCamSys, CList& oSrc )
{
   DWORD
      dwNumVert = oSrc.mRetNumNodes ();  // number of nodes

   // verify if need realocate memory for destin list (not enough nodes)
   if ( dwNumVert > m_oCamVert.mRetNumNodes () ) {

      // disable direct access feature
      m_oCamVert.mDisableDirectAccess ();

      // define how many nodes must be added
      DWORD
         dwNew = dwNumVert - m_oCamVert.mRetNumNodes ();
      // alocate more nodes for destin
      for ( DWORD dwcN = 0; dwcN < dwNew; dwcN ++ ) {

         CNodeXYZ
            *poNew = new CNodeXYZ;
         // verify alocation error
         if ( poNew == NULL )
            ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

         m_oCamVert.mSetEnd ( (CNode*) poNew );
      }

      // enable direct access feature
      m_oCamVert.mEnableDirectAccess ();
   }

   // transform vertexes (move and rotate)
   tagfXYZ
      stPos;
   stPos.fX = poCamSys->m_stPos.fX;
   stPos.fY = poCamSys->m_stPos.fY;
   stPos.fZ = poCamSys->m_stPos.fZ;
   // select first vertex of lists
   oSrc.mSelectStart ();
   m_oCamVert.mSelectStart ();

   // scan all source list transforming to destin list
   while ( oSrc.mRetExistSelection () ) {

      // point vertexes
      CNodeXYZ
         *poVSrc = (CNodeXYZ*) oSrc.mRetSelected ();
      CNodeXYZ
         *poVDest = (CNodeXYZ*) m_oCamVert.mRetSelected ();

      // apply reference system origin (move)
      poVDest->m_stXYZ.fX = ( poVSrc->m_stXYZ.fX - stPos.fX ) * RENDERING_SCALE_FACTOR;
      poVDest->m_stXYZ.fY = ( poVSrc->m_stXYZ.fY - stPos.fY ) * RENDERING_SCALE_FACTOR;
      poVDest->m_stXYZ.fZ = ( poVSrc->m_stXYZ.fZ - stPos.fZ ) * RENDERING_SCALE_FACTOR;

      // apply reference system base (rotate)
      tagfXYZ
         stRot;   // temp result
      stRot.fX = ( poVDest->m_stXYZ.fX * poCamSys->m_stI.fX ) +
                 ( poVDest->m_stXYZ.fY * poCamSys->m_stI.fY ) +
                 ( poVDest->m_stXYZ.fZ * poCamSys->m_stI.fZ );
      stRot.fY = ( poVDest->m_stXYZ.fX * poCamSys->m_stJ.fX ) +
                 ( poVDest->m_stXYZ.fY * poCamSys->m_stJ.fY ) +
                 ( poVDest->m_stXYZ.fZ * poCamSys->m_stJ.fZ );
      stRot.fZ = ( poVDest->m_stXYZ.fX * poCamSys->m_stK.fX ) +
                 ( poVDest->m_stXYZ.fY * poCamSys->m_stK.fY ) +
                 ( poVDest->m_stXYZ.fZ * poCamSys->m_stK.fZ );
      poVDest->m_stXYZ.fX = stRot.fX;
      poVDest->m_stXYZ.fY = stRot.fY;
      poVDest->m_stXYZ.fZ = stRot.fZ;

      // next node
      oSrc.mSelectNext ();
      m_oCamVert.mSelectNext ();
   }
}



/* Make de projection from a 3D mesh to a 2D mesh and ajust
to viewport.
Use m_oCamVert as input and m_oPrjVert as output.
poCamSys is the camera to be used in transformation.
**********************************************************/
void CRender::mAjust ( CCameraSystem *poCamSys )
{
   // verify if need grow projected vertexes list (insert new nodes)
   if ( m_oPrjVert.mRetNumNodes () < m_oCamVert.mRetNumNodes () ) {

      // number of nodes to insert
      DWORD
         dwNumNew = m_oCamVert.mRetNumNodes () - m_oPrjVert.mRetNumNodes ();

      // no direct access (need to allow list pointers change)
      m_oPrjVert.mDisableDirectAccess ();

      // add vertexes
      for ( DWORD dwcD = 0; dwcD < dwNumNew; dwcD ++ ) {
         CNodeXY
            *poNode = new CNodeXY;
         if ( poNode == NULL )
            ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
         // add to vertex list
         m_oPrjVert.mSetEnd ( poNode );
      }

      // direct access (for fast random access)
      m_oPrjVert.mEnableDirectAccess ();
   }

   // for each vertex
   m_oCamVert.mSelectStart ();
   m_oPrjVert.mSelectStart ();
   while ( m_oCamVert.mRetExistSelection () ) {

      // vertex to project
      CNodeXYZ
         *poVert = (CNodeXYZ*) m_oCamVert.mRetSelected ();
      // projected vertex
      CNodeXY
         *poXY = (CNodeXY*) m_oPrjVert.mRetSelected ();

      // project
      // verify if vertex is in front of the projection plane
      if ( poVert->m_stXYZ.fZ > - poCamSys->m_fDistCP ) {

         // make standard projection
         poXY->m_stXY.nX = (INT) ( (float) poVert->m_stXYZ.fX / ( (float) poVert->m_stXYZ.fZ / poCamSys->m_fDistCP + 1 ) );
         poXY->m_stXY.nY = (INT) ( (float) poVert->m_stXYZ.fY / ( (float) poVert->m_stXYZ.fZ / poCamSys->m_fDistCP + 1 ) );
      }
      else {
         // REVIEW - soh projeta se positivo, senao nem aloca memoria,
         //     seta flag no pol...? ou aloca e nao usa, seta flag X=-1
         // make interpolated projection
         poXY->m_stXY.nX = (INT) ( (float) poVert->m_stXYZ.fX / ( (float) poVert->m_stXYZ.fZ / ( 1 - poVert->m_stXYZ.fZ ) + 1 ) );
         poXY->m_stXY.nY = (INT) ( (float) poVert->m_stXYZ.fY / ( (float) poVert->m_stXYZ.fZ / ( 1 - poVert->m_stXYZ.fZ ) + 1 ) );
      }

      // ajust for viewport (middle frame buffer, invert Y, scale)
      poXY->m_stXY.nX = poXY->m_stXY.nX + m_oDraw.m_nMidX;
      poXY->m_stXY.nY = m_oDraw.m_nResY - ( poXY->m_stXY.nY + m_oDraw.m_nMidY );

      // next vertex
      m_oCamVert.mSelectNext ();
      m_oPrjVert.mSelectNext ();
   }
}



/* Define polygon data: projected vertex coordenates, vertex light
levels; make frame buffer clip.
  For clipping is counted the number of vertexes that are outside
  the clip box area. Is a first clip method in pipeline, not really
  powerfull.
poPol - define what vertexes are used in this poly
oLights - the light levels for all mesh vertexes
stClipBox - the bounding box where the vertexes are clipped
dwBehind - how many vertexes are behind the clip plane
Update: m_dwMaxVertArray, m_pastPolVertexCoords, m_pafPolVertexLights
return TRUE - poly is totaly or partialy visible / FALSE - invisible
return dwBehind - number of vertexes behind the near Z clip plane
**********************************************************/
BOOL CRender::mDefinePolygonData ( CNodePolygon *poPol, CList &oLights,
                                      tagRectangle& stClipBox,
                                      DWORD& dwBehind )
{
   // counters for the number of vertexes outside the clip area
   DWORD
      dwLeft = 0,
      dwRight = 0,
      dwTop = 0,
      dwBottom = 0;
   dwBehind = 0;
   
   // verify if have enough space in poly vertex array, if no realocate
   if ( m_dwMaxVertArray < poPol->m_oPolygon.m_dwNumVertexes ) {

      // realocate memory
      // desalocate
      delete m_pastPolVertexCoords;
      // alocate
      if ( ( m_pastPolVertexCoords = new tagXY [ poPol->m_oPolygon.m_dwNumVertexes ] ) == NULL )
         ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      // realocate memory
      // desalocate
      delete m_pafPolVertexLights;
      // alocate
      if ( ( m_pafPolVertexLights = new float [ poPol->m_oPolygon.m_dwNumVertexes ] ) == NULL )
         ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
      #endif

      // set number of elements
      m_dwMaxVertArray = poPol->m_oPolygon.m_dwNumVertexes;
   }

   // scan polygon vertex indexes and copy respective coordenates to array
   DWORD
      *pdwNdxVert;        // point vertex indexes array in polygon
   pdwNdxVert = poPol->m_oPolygon.m_padwVertexes;
   tagXY
      *pstPolVertexes;    // point vertex coordenate array
   pstPolVertexes = m_pastPolVertexCoords;    // point a vertex coord
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   float
      *pfVertLight;       // point vertex light level
   pfVertLight = m_pafPolVertexLights;
   #endif
   for ( DWORD dwcI = 0; dwcI < poPol->m_oPolygon.m_dwNumVertexes; dwcI ++ ) {

      // define vertex projected coordenate
      CNodeXY
         *poNodeVert = (CNodeXY*) m_oPrjVert.mRetNode ( *pdwNdxVert );

      // copy coords
      pstPolVertexes->nX = poNodeVert->m_stXY.nX;
      pstPolVertexes->nY = poNodeVert->m_stXY.nY;
      
      // define light level
      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      *pfVertLight = ( (CNodeLightInfo*) oLights.mRetNode ( *pdwNdxVert ) )->m_stValue.fAmbient +
                     ( (CNodeLightInfo*) oLights.mRetNode ( *pdwNdxVert ) )->m_stValue.fSource;
      // if light level is more than 1 or less than 0, reduce to 1 or 0
      if ( *pfVertLight > 1 )
         *pfVertLight = 1;
      if ( *pfVertLight < 0 )
         *pfVertLight = 0;
      #endif

      // verify if vertex is outside visible area

      // define the 3D camera vertex coordenate
      CNodeXYZ
         *poNodeVertCam = (CNodeXYZ*) m_oCamVert.mRetNode ( *pdwNdxVert );

      // verify if is behind the clip plan
      if ( poNodeVertCam->m_stXYZ.fZ < Z_CLIP )
         dwBehind ++;

      // verify if is outside frame buffer: left, right
      if ( pstPolVertexes->nX < stClipBox.nStartX )
         dwLeft ++;
      else
         if ( pstPolVertexes->nX > stClipBox.nEndX )
            dwRight ++;

      // verify top and bottom
      if ( pstPolVertexes->nY < stClipBox.nStartY )
         dwTop ++;
      else
         if ( pstPolVertexes->nY > stClipBox.nEndY )
            dwBottom ++;

      // next arrays positions (test next)
      pdwNdxVert ++;
      pstPolVertexes ++;
      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      pfVertLight ++;
      #endif
   }

   // verify if all vertexes are outside the frame buffer, if no -> draw
   if ( dwLeft   != poPol->m_oPolygon.m_dwNumVertexes &&
        dwRight  != poPol->m_oPolygon.m_dwNumVertexes &&
        dwTop    != poPol->m_oPolygon.m_dwNumVertexes &&
        dwBottom != poPol->m_oPolygon.m_dwNumVertexes &&
        dwBehind != poPol->m_oPolygon.m_dwNumVertexes )
      return TRUE;

   // no poly draw
   return FALSE;
}
          

          
/* Perform back-face culling: verify if visible poly face is in front
of the view plane. For this is compared the poly normal vector Z value.
return TRUE - poly visible / FALSE - poly invisible
**********************************************************/
BOOL CRender::mPassBackFaceCulling ()
{
   // REVIEW - now does not work very well: when you is close->overflow
   /* fazer ateh dot-product e olhar sinal. usar vetores no SRC
   vetor de ti ate vertice mais proximo em Z?
   Parece que o problema nao eh soh overflow......
   // make normal vector visible cut
   tagfXYZ
      stNormalCam;
   // polygon normal in SRU
   stNormalCam.fX = poPol->m_oPolygon.m_stNormal.fX;
   stNormalCam.fY = poPol->m_oPolygon.m_stNormal.fY;
   stNormalCam.fZ = poPol->m_oPolygon.m_stNormal.fZ;
   // polygon normal in SRC
   poCamSys->mApplyBase ( stNormalCam );
   */

   // define 3 vertexes inside polygon plan, use project x y
   tagXY
      *pstVert0 = &m_pastPolVertexCoords [ 0 ],
      *pstVert1 = &m_pastPolVertexCoords [ 1 ],
      *pstVert2 = &m_pastPolVertexCoords [ 2 ];

   // define 2 vectors (x,y) components inside the poly plan
   tagfXYZ
      stXY1,
      stXY2;
   stXY1.fX = (float) ( pstVert0->nX - pstVert1->nX ) / 1000;
   stXY1.fY = (float) ( pstVert0->nY - pstVert1->nY ) / 1000;
   stXY2.fX = (float) ( pstVert0->nX - pstVert2->nX ) / 1000;
   stXY2.fY = (float) ( pstVert0->nY - pstVert2->nY ) / 1000;

   // define only the normal Z value
   float
      fNormalZ = stXY1.fX * stXY2.fY - stXY1.fY * stXY2.fX;

   // verify if is negative
   if ( fNormalZ < 0 )
      return TRUE;

   return FALSE;
}



/* Get texture bitmap and texel coordenates for a polygon.
If the texture ID == -1 -> poly with no texture
return ppoTx - point the bitmap pointer (bitmap pointer get NULL if no texture)
       ppastTexVertexes - point the texels coord pointer
                          (pointed pointer is NULL if no texture)
       Texels are texture coodenates.
**********************************************************/
void CRender::mGetPolyTextureData ( CNodePolygon *poPol,
                                    CTexture **ppoTx,
                                    tagXY **ppastTexVertexes )
{
   // define polygon texture
   // point texture or NULL if no texture defined
   CNodeTexture
      *poNodeTx = (CNodeTexture*) m_poScen->m_oTextures.mRetValue ( poPol->m_oPolygon.m_dwTextureID );

   // verify if texture not found, or poly with no texture (ID == -1)
   if ( poNodeTx == NULL )
      // verify if texture ID == -1 -> no texture for this poly, else error
      if ( (INT) poPol->m_oPolygon.m_dwTextureID != -1 )
         ABORT ( MSG_ERROR_NO_TEXTURE, __FILE__, __LINE__ )

   // define texture coordenates
   // if texture defined
   if ( poNodeTx != NULL ) {

      // point texture object
      *ppoTx = &poNodeTx->m_oTx;

      // select a predefined array of texture vertexes
      switch ( poPol->m_oPolygon.m_dwNumVertexes ) {

         case 3:
            *ppastTexVertexes = m_astTexVertexes3;
            break;

         case 4:
            *ppastTexVertexes = m_astTexVertexes4;
            break;

         case 5:
            *ppastTexVertexes = m_astTexVertexes5;
            break;

         case 6:
            *ppastTexVertexes = m_astTexVertexes6;
            break;

         case 7:
            *ppastTexVertexes = m_astTexVertexes7;
            break;

         case 8:
            *ppastTexVertexes = m_astTexVertexes8;
            break;

         default:
            ABORT ( MSG_ERROR_MANY_VERTEXES, __FILE__, __LINE__ )
            break;
      }
   }
   else {
      // no texture defined
      *ppoTx = NULL;
      *ppastTexVertexes = NULL;
   }
}
   


/* Create a new poly clipped with the near Z plane.
poPol - the source polygon to be clipped
pastTexVertexes - the texture vertexes
poCamSys - to retransform the new poly
pastClipEdgeBuffer - used to set the poly with it's vertexes, if portal
stClipBox - return with the frame buffer limits if poly is a portal
Update:
m_pastPolZClipCoords,
m_pastPolZClipTex,
m_dwMaxZClipArray,
m_pafPolZClipLight with the new poly data.
return the number of vertexes in new poly and stClipBox (see up)
**********************************************************/
DWORD CRender::mMakeZNearPlaneClip ( CNodePolygon* poPol,
                                     tagXY *pastTexVertexes,
                                     CCameraSystem *poCamSys,
                                     tagInterval* pastClipEdgeBuffer,
                                     tagRectangle& stClipBox )
{ 
   // define number of vertexes in polygon
   DWORD
      dwOldPolNumVertexes = poPol->m_oPolygon.m_dwNumVertexes;

   // search one point in the visible side of Z plan
   DWORD
      *pdwNdxVertClip;    // point vertex indexes array in poly
   pdwNdxVertClip = poPol->m_oPolygon.m_padwVertexes;
   DWORD
      dwNdxFirst = 0;     // the index of the vertex found
   // scan all polygon vertexes
   for ( DWORD dwcV = 0; dwcV < dwOldPolNumVertexes; dwcV ++ )
      // verify if vertex is in front of the Z plan
      if ( ( (CNodeXYZ*) m_oCamVert.mRetNode ( *pdwNdxVertClip ) )->m_stXYZ.fZ >= Z_CLIP )
         break;
      else {
         pdwNdxVertClip ++;
         dwNdxFirst ++;
      }

   // verify if need grow the vertexes array in clipped poly
   if ( m_dwMaxZClipArray < ( m_dwMaxVertArray + 1 ) ) {
      // +1 for new clipped vertex!

      // realocate memory for polygon vertexes
      // desalocate
      if ( m_pastPolZClipCoords != NULL )
         delete m_pastPolZClipCoords;
      // alocate
      if ( ( m_pastPolZClipCoords = new tagXY [ dwOldPolNumVertexes + 1 ] ) == NULL )
         ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

      // if texture defined
      if ( pastTexVertexes != NULL ) {
         // realocate memory for polygon texture vertexes
         // desalocate
         if ( m_pastPolZClipTex != NULL )
            delete m_pastPolZClipTex;
         // alocate
         if ( ( m_pastPolZClipTex = new tagXY [ dwOldPolNumVertexes + 1 ] ) == NULL )
            ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
      }

      // realocate memory for polygon vertexes light levels
      // desalocate
      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      if ( m_pafPolZClipLight != NULL )
         delete m_pafPolZClipLight;
      // alocate
      if ( ( m_pafPolZClipLight = new float [ dwOldPolNumVertexes + 1 ] ) == NULL )
         ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
      #endif

      // define new number of vertexes
      m_dwMaxZClipArray = dwOldPolNumVertexes + 1;
   }

   // if the polygon is a portal, reset the clip edge buffer and set
   // poly to frame buffer size. Don't continue the rest of clip.
   if ( poPol->m_oPolygon.m_stPortal.yCode != 0 ) {

      // set new poly data
      // coords
      m_pastPolZClipCoords [ 0 ].nX = 0;
      m_pastPolZClipCoords [ 0 ].nY = 0;
      m_pastPolZClipCoords [ 1 ].nX = 0;
      m_pastPolZClipCoords [ 1 ].nY = m_oDraw.m_nMaxY;
      m_pastPolZClipCoords [ 2 ].nX = m_oDraw.m_nMaxX;
      m_pastPolZClipCoords [ 2 ].nY = m_oDraw.m_nMaxY;
      m_pastPolZClipCoords [ 3 ].nX = m_oDraw.m_nMaxX;
      m_pastPolZClipCoords [ 3 ].nY = 0;
      // texels
      memcpy ( (void*) m_pastPolZClipTex, (void*) m_astTexVertexes4, sizeof ( tagXY ) * 4 );
      // light levels
      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      memcpy ( (void*) m_pafPolZClipLight, (void*) m_pafPolVertexLights, sizeof ( float ) * 4 );
      #endif
      
      // set clip edge buffer
      m_oDraw.mResetClipEdgeBuffer ( pastClipEdgeBuffer );
      // the initial clip bounding rectangle
      stClipBox.nStartX = 0;
      stClipBox.nEndX = m_oDraw.m_nMaxX;
      stClipBox.nStartY = 0;
      stClipBox.nEndY = m_oDraw.m_nMaxY;
     
      return 4;
   }

   // copy vertexes while they are in front of the Z clip plan

   tagXY
      // source coord
      *pstSrcCoord = &m_pastPolVertexCoords [ dwNdxFirst ],
      // destin coord
      *pstDestCoord = m_pastPolZClipCoords,
      // texture coords
      *pstSrcTex = NULL,
      *pstDestTex = NULL,
      // the end of vertex coord array
      *pstEndPolVertexCoords = &m_pastPolVertexCoords [ dwOldPolNumVertexes - 1 ];
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   float
      // light element
      *pfSrcLight = &m_pafPolVertexLights [ dwNdxFirst ],
      *pfDestLight = m_pafPolZClipLight;
   #endif
   // if texture defined
   if ( pastTexVertexes != NULL ) {
      // src texture coord
      pstSrcTex = &pastTexVertexes [ dwNdxFirst ];
      // dest texture coord
      pstDestTex = m_pastPolZClipTex;
   }
   DWORD
      // the index of source vertex (in polygon array)
      dwNdxSrc = dwNdxFirst,
      // the index of last copyed vertex (in polygon array)
      dwNdxLastSrc,
      // number of vertexes in new poly
      dwNewPolNumVert = 0;
   DWORD
      // point poly vertex index array
      *pdwVertNdxSrc = &poPol->m_oPolygon.m_padwVertexes [ dwNdxFirst ];

   // scan, for each vertex
   do {

      // copy
      pstDestCoord->nX = pstSrcCoord->nX;
      pstDestCoord->nY = pstSrcCoord->nY;
      // if texture defined
      if ( pastTexVertexes != NULL ) {
         pstDestTex->nX = pstSrcTex->nX;
         pstDestTex->nY = pstSrcTex->nY;
      }
      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      *pfDestLight = *pfSrcLight;
      #endif
      // define last data copyed
      dwNdxLastSrc = dwNdxSrc;

      // define next poly vertex and texture vertex to copy
      // verify if is in the end of poly vertex array
      if ( pstSrcCoord == pstEndPolVertexCoords ) {

         // goto start all source pointers, and advance destin pointers

         pstSrcCoord = m_pastPolVertexCoords;
         pstDestCoord ++;
         dwNdxSrc = 0;
         pdwVertNdxSrc = poPol->m_oPolygon.m_padwVertexes;
         dwNewPolNumVert ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstSrcTex = pastTexVertexes;
            pstDestTex ++;
         }
         // light levels
         #ifdef COMP_RENDER_INTERPOLATED_LIGHT
         pfSrcLight = m_pafPolVertexLights;
         pfDestLight ++;
         #endif
      }
      else {

         // advance source and destin pointers

         pstSrcCoord ++;
         pstDestCoord ++;
         dwNdxSrc ++;
         pdwVertNdxSrc ++;
         dwNewPolNumVert ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstDestTex ++;
            pstSrcTex ++;
         }
         // light levels
         #ifdef COMP_RENDER_INTERPOLATED_LIGHT
         pfSrcLight ++;
         pfDestLight ++;
         #endif
      }

   // while the vertex is in front of the clip plane
   } while ( ( (CNodeXYZ*) m_oCamVert.mRetNode ( *pdwVertNdxSrc ) )->m_stXYZ.fZ >= Z_CLIP );

   // define the intercection point between the line and the plan, and
   // all data (texture coords, light levels) at the intersection point
   
   // the 2 point that compose the line
   CNodeXYZ
      *poVert1 = (CNodeXYZ*) m_oCamVert.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ dwNdxSrc ] ),
      *poVert2 = (CNodeXYZ*) m_oCamVert.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ dwNdxLastSrc ] );
   tagXY
      // texture points
      *pstTex1,
      *pstTex2;
   float
      // define delta Z rate between the points
      fZRate = ( (float) Z_CLIP - poVert1->m_stXYZ.fZ ) / ( poVert2->m_stXYZ.fZ - poVert1->m_stXYZ.fZ );
   // define new polygon x and y in camera reference system
   pstDestCoord->nX = (INT) ( poVert1->m_stXYZ.fX + ( ( poVert2->m_stXYZ.fX - poVert1->m_stXYZ.fX ) * fZRate ) );
   pstDestCoord->nY = (INT) ( poVert1->m_stXYZ.fY + ( ( poVert2->m_stXYZ.fY - poVert1->m_stXYZ.fY ) * fZRate ) );
   // project vertex
   pstDestCoord->nX = (INT) ( (float) pstDestCoord->nX / ( Z_CLIP / poCamSys->m_fDistCP + 1 ) );
   pstDestCoord->nY = (INT) ( (float) pstDestCoord->nY / ( Z_CLIP / poCamSys->m_fDistCP + 1 ) );
   // just to view port
   pstDestCoord->nX = pstDestCoord->nX + m_oDraw.m_nMidX;
   pstDestCoord->nY = m_oDraw.m_nResY - ( pstDestCoord->nY + m_oDraw.m_nMidY );
   // if texture defined
   if ( pastTexVertexes != NULL ) {
      // first texture point (behind clip plan)
      pstTex1 = &pastTexVertexes [ dwNdxSrc ],
      // second texture point (in front of clip plan)
      pstTex2 = &pastTexVertexes [ dwNdxLastSrc ];
      // define new texture x and y (use float for better precision)
      pstDestTex->nX = pstTex1->nX + (INT) ( (float) ( pstTex2->nX - pstTex1->nX ) * fZRate );
      pstDestTex->nY = pstTex1->nY + (INT) ( (float) ( pstTex2->nY - pstTex1->nY ) * fZRate );
      pstDestTex ++;
   }
   // light levels
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   float
      *pfSrcLight1 = &m_pafPolVertexLights [ dwNdxSrc ],   // behind
      *pfSrcLight2 = &m_pafPolVertexLights [ dwNdxLastSrc ];  // in front
   // new light value at plane point
   *pfDestLight = *pfSrcLight1 + ( *pfSrcLight2 - *pfSrcLight1 ) * fZRate;
   // next
   pfDestLight ++;
   #endif
   pstDestCoord ++;
   dwNewPolNumVert ++;

   // skip vertexes that are behind the plan

   // for each vertex
   do {
      // define last data copyed
      dwNdxLastSrc = dwNdxSrc;
      // define next poly vertex and texture vertex to copy
      // verify if is in the end of poly vertex array
      if ( pstSrcCoord == pstEndPolVertexCoords ) {
         // goto start
         pstSrcCoord = m_pastPolVertexCoords;
         dwNdxSrc = 0;
         pdwVertNdxSrc = poPol->m_oPolygon.m_padwVertexes;
         // if texture defined
         if ( pastTexVertexes != NULL )
            pstSrcTex = pastTexVertexes;
         // light levels
         #ifdef COMP_RENDER_INTERPOLATED_LIGHT
         pfSrcLight = m_pafPolVertexLights;
         #endif
      }
      else {
         // set to next position
         pstSrcCoord ++;
         dwNdxSrc ++;
         pdwVertNdxSrc ++;
         // if texture defined
         if ( pastTexVertexes != NULL )
            pstSrcTex ++;
         // light levels
         #ifdef COMP_RENDER_INTERPOLATED_LIGHT
         pfSrcLight ++;
         #endif
      }
   } while ( ( (CNodeXYZ*) m_oCamVert.mRetNode ( *pdwVertNdxSrc ) )->m_stXYZ.fZ < Z_CLIP );

   // define the intercection point between the line and the plan, and
   // all data

   // the 2 point that compose the line
   poVert2 = (CNodeXYZ*) m_oCamVert.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ dwNdxSrc ] ),
   poVert1 = (CNodeXYZ*) m_oCamVert.mRetNode ( poPol->m_oPolygon.m_padwVertexes [ dwNdxLastSrc ] );
   // define delta Z rate between the points
   fZRate = ( (float) Z_CLIP - poVert1->m_stXYZ.fZ ) / ( poVert2->m_stXYZ.fZ - poVert1->m_stXYZ.fZ );
   // define new polygon x and y in camera reference system
   pstDestCoord->nX = (INT) ( poVert1->m_stXYZ.fX + ( ( poVert2->m_stXYZ.fX - poVert1->m_stXYZ.fX ) * fZRate ) );
   pstDestCoord->nY = (INT) ( poVert1->m_stXYZ.fY + ( ( poVert2->m_stXYZ.fY - poVert1->m_stXYZ.fY ) * fZRate ) );
   // project vertex
   pstDestCoord->nX = (INT) ( (float) pstDestCoord->nX / ( Z_CLIP / poCamSys->m_fDistCP + 1 ) );
   pstDestCoord->nY = (INT) ( (float) pstDestCoord->nY / ( Z_CLIP / poCamSys->m_fDistCP + 1 ) );
   // just to view port
   pstDestCoord->nX = pstDestCoord->nX + m_oDraw.m_nMidX;
   pstDestCoord->nY = m_oDraw.m_nResY - ( pstDestCoord->nY + m_oDraw.m_nMidY );
   // if texture defined
   if ( pastTexVertexes != NULL ) {
      // first texture point (behind clip plan)
      pstTex2 = &pastTexVertexes [ dwNdxSrc ],
      // second texture point (in front of clip plan)
      pstTex1 = &pastTexVertexes [ dwNdxLastSrc ];
      // define new texture x and y (use float for better precision)
      pstDestTex->nX = pstTex1->nX + (INT) ( (float) ( pstTex2->nX - pstTex1->nX ) * fZRate );
      pstDestTex->nY = pstTex1->nY + (INT) ( (float) ( pstTex2->nY - pstTex1->nY ) * fZRate );
      pstDestTex ++;
   }
   // light levels
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   pfSrcLight2 = &m_pafPolVertexLights [ dwNdxSrc ],   // behind
   pfSrcLight1 = &m_pafPolVertexLights [ dwNdxLastSrc ];  // in front
   // new light value at plane point
   *pfDestLight = *pfSrcLight1 + ( *pfSrcLight2 - *pfSrcLight1 ) * fZRate;
   // next
   pfDestLight ++;
   #endif
   pstDestCoord ++;
   dwNewPolNumVert ++;

   // copy rest of vertexes while be found the start vertex

   // for each vertex
   while ( dwNdxSrc != dwNdxFirst ) {

      // copy
      pstDestCoord->nX = pstSrcCoord->nX;
      pstDestCoord->nY = pstSrcCoord->nY;
      // if texture defined
      if ( pastTexVertexes != NULL ) {
         pstDestTex->nX = pstSrcTex->nX;
         pstDestTex->nY = pstSrcTex->nY;
      }
      #ifdef COMP_RENDER_INTERPOLATED_LIGHT
      *pfDestLight = *pfSrcLight;
      #endif

      // define next poly vertex and texture vertex to copy
      // verify if is in the end of poly vertex array
      if ( pstSrcCoord == pstEndPolVertexCoords ) {

         // all source pointer goto start
         pstSrcCoord = m_pastPolVertexCoords;
         pstDestCoord ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstSrcTex = pastTexVertexes;
            pstDestTex ++;
         }
         dwNdxSrc = 0;
         pdwVertNdxSrc = poPol->m_oPolygon.m_padwVertexes;
         dwNewPolNumVert ++;
         // light levels
         #ifdef COMP_RENDER_INTERPOLATED_LIGHT
         pfSrcLight = m_pafPolVertexLights;
         pfDestLight ++;
         #endif
      }
      else {
         // set to next position
         pstSrcCoord ++;
         pstDestCoord ++;
         dwNdxSrc ++;
         pdwVertNdxSrc ++;
         dwNewPolNumVert ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstSrcTex ++;
            pstDestTex ++;
         }
         // light levels
         #ifdef COMP_RENDER_INTERPOLATED_LIGHT
         pfSrcLight ++;
         pfDestLight ++;
         #endif
      }
   }

   return dwNewPolNumVert;
}
