/* Store all data needed in a scenery. Have the basic definitions (can be
instanciated), object instances, textures, the palette (if needed),
cameras for visualization.
     
A DEFINE is a generic data store, can have one object mesh, for example.
An OBJECT is an instance of one or more DEFINES.
Both DEFINE and OBJECT are stored in CObject objects.

TIPS:
- the data can be loaded from a TXE file, see class CFileTXE for more info.

WARNING:
- after change a define or a sector data you MUST CALL CScenery::mBeginAddDefine
  or CScenery::mBeginAddSector, respectively.
- before change a define or a sector data you MUST CALL CScenery::mEndAddDefine
  or CScenery::mEndAddSector, respectively.
- Use CScenery::mSetCurrentCamera to change the current camera
*/



#include <string.h>

#include "defs.h"

#include "mem.h"

#include "array.h"

#include "string.h"

#include "file.h"

#include "filerem.h"

#include "filepcx.h"

#include "node.h"

#include "nodeflo.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "map.h"

#include "texture.h"

#include "lighttab.h"

#include "nodelgt.h"

#include "math.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"

#include "refsys.h"

#include "camsys.h"

#include "object.h"

#include "nodeint.h"

#include "nodetex.h"

#include "nodeobj.h"

#include "nodexyz.h"

#include "eventtab.h"

#include "comp/nodecomp.h"

#include "comp/cocamsys.h"

#include "comp/comesh.h"

#include "comp/cosecset.h"

#include "filetxe.h"

#include "scenery.h"



/* constructor - init empty scenery
**********************************************************/
CScenery::CScenery ()
{
   // flag what is beeing defined
   m_bDefining = TRUE;

   m_bEnd = FALSE;
   m_dwSectorID = m_dwDefineID = 0;
   m_nClearBackground = -1;
   m_poCamera = NULL;
   m_fDefaultScaleFactor = DEFAULT_SCALE_FACTOR;
   m_fDefaultLightLevel = DEFAULT_LIGHT_LEVEL;
   m_dwDefaultLightTable = DEFAULT_LIGHT_TABLE;
}



/* constructor - load TXE file to this scenery,
if fail abort programm
**********************************************************/
CScenery::CScenery ( CString oszFile )
{
   // flag what is beeing define
   m_bDefining = TRUE;

   m_bEnd = FALSE;
   m_dwSectorID = m_dwDefineID = 0;
   m_nClearBackground = -1;
   m_poCamera = NULL;

   m_fDefaultScaleFactor = DEFAULT_SCALE_FACTOR;
   m_fDefaultLightLevel = DEFAULT_LIGHT_LEVEL;
   m_dwDefaultLightTable = DEFAULT_LIGHT_TABLE;

   // try load
   if ( ! mCreate ( oszFile ) )
      ABORT ( MSG_ERROR_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* destructor - desalocate buffer, if alocated
**********************************************************/
CScenery::~CScenery ()
{
   mDestroy ();
}



/* Load scenery data. (like CScenery::constructor)
return TRUE- ok / FALSE - error
**********************************************************/
BOOL CScenery::mCreate ( CString oszFile )
{
   mDestroy ();

   // try load
   CFileTXE
      oFile;
   if ( ! oFile.mLoad ( oszFile, *this ) )
      return FALSE;

   return TRUE;
}



/* Empty scenery data.
**********************************************************/
void CScenery::mDestroy ()
{
   // flag what is beeing define
   m_bDefining = TRUE;

   m_bEnd = FALSE;
   m_dwSectorID = m_dwDefineID = 0;
   m_nClearBackground = -1;
   m_poCamera = NULL;
   m_fDefaultScaleFactor = DEFAULT_SCALE_FACTOR;
   m_fDefaultLightLevel = DEFAULT_LIGHT_LEVEL;
   m_dwDefaultLightTable = DEFAULT_LIGHT_TABLE;

   m_oRoot.m_oSector.mDestroy ();
   m_oPal.mDestroy ();
   m_oTextures.mRemoveAll ();
   m_oDefines.mRemoveAll ();
   m_oLightTables.mRemoveAll ();
   m_oEventTab.mDestroy ();
}



/* Add palette for graphics mode (if they have)
return TRUE - ok / FALSE - error (palette already defined)
**********************************************************/
BOOL CScenery::mAddPalette ( CMem oPal )
{
   // verify if palette have been already defined
   if ( m_oPal.m_dwSize != 0 ) {
      printf ( MSG_ERROR_PALETTE_ALREADY_DEFINED );
      return FALSE;
   }

   // copy palette from source oPal to m_oPal
   m_oPal = oPal;

   return TRUE;
}



/* Add light table for shaded polys.
If oLgT is empty, create light table from source palette
(CScenery::m_oPal) using dwNumLevels and save it to file oszFile.
dwID - light table ID
fR, fG, fB - light source color (if need), the values are between 0 and 1;
return TRUE - ok / FALSE - error (impossible define light table:
if is already defined or creation error - no source palette)
**********************************************************/
BOOL CScenery::mAddLightTable ( CLightTable& oLgT, DWORD dwID, CString oszFile,
                                DWORD dwNumLevels, float fR, float fG, float fB )
{
   // if key already defined, return
   if ( m_oLightTables [ dwID ] != NULL ) {
      printf ( MSG_ERROR_LIGHTTABLE_ID_ALREADY_DEFINED, dwID );
      return FALSE;
   }

   // verify if loaded light table have error, if yes try create it
   if ( oLgT.m_payLightTable == NULL ) {

      // verify if source palette have been already defined
      if ( m_oPal.m_dwSize == 0 ) {
         printf ( MSG_ERROR_NOPALETTE_FOR_LIGHTTABLE );
         return FALSE;
      }
     
      // create it
      oLgT.mCreate ( m_oPal, dwNumLevels, fR, fG, fB ); 

      // save light table
      oLgT.mSave ( oszFile );
   }

   // try create light table node
   CNodeLightTable
      *poLgT = new CNodeLightTable ( oLgT );
   if ( poLgT == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // add in scenery
   if ( ! m_oLightTables.mAddKey ( dwID, (CNode*) poLgT ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   return TRUE;
}
   


/* Add texture to map relation. Map relations are implemented
by CMap class and define a relationship between a key and one value,
here the key is the texture ID and the value is the texture.
See CMap implementation file for more info.
return TRUE - ok / FALSE - error (texture ID already defined)
**********************************************************/
BOOL CScenery::mAddTexture ( DWORD dwID, CTexture& oTx )
{
   // if key already defined, return
   if ( m_oTextures [ dwID ] != NULL ) {
      printf ( MSG_ERROR_TEXTURE_ID_ALREADY_DEFINED, dwID );
      return FALSE;
   }

   // try create texture node
   CNodeTexture
      *poTx = new CNodeTexture ( oTx );
   if ( poTx == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // add in scenery
   if ( ! m_oTextures.mAddKey ( dwID, (CNode*) poTx ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   return TRUE;
}



/* Add an object instance in scenery.
Before call this function add all possible defines in scenery.
wInstID - this object instance ID - the name (must be uniq in ALL scenery)
dwDestinInstID - inside of what instanciated object insert this instance
dwDestinSectorID - inside of what sector insert this instance
oDefs - a list of CNodeFloat with the defines ID to compose this instance
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mAddObject ( DWORD dwInstID,
                            DWORD dwDestinInstID, DWORD dwDestinSectorID,
                            CList& oDefs )
{
   // verify if the given instance is already defined
   if ( m_oRoot.m_oSector.mSearchObject ( dwInstID ) != NULL ) {
      printf ( MSG_ERROR_OBJECT_ID_ALREADY_DEFINED, dwInstID );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // search sector where insert instance
   CNodeObject
      *poObj;
   CNodeSector
      *poSec;

   // verify if insert instance in root sector (first scenery sector)
   if ( (INT) dwDestinInstID == -1 )

      // use the root sector
      poSec = (CNodeSector*) &m_oRoot;

   else {

      // search for the object to insert instance
      poObj = m_oRoot.m_oSector.mSearchObject ( dwDestinInstID );
      // verify if not found -> error
      if ( poObj == NULL ) {
         printf ( MSG_ERROR_DESTIN_OBJECT_NOT_FOUND, dwDestinInstID );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }

      // search for the sectors component
      CComponentSectorsSet
         *poSectors = (CComponentSectorsSet*) poObj->m_oObject.mRetComponent ( SLOT_INTERIOR );
      // verify if not found -> error
      if ( poSectors == NULL ) {
         printf ( MSG_ERROR_NO_TYPE );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }

      // search for sector
      poSec = (CNodeSector*) poSectors->m_oSectors.mRetValue ( dwDestinSectorID );
      // verify if not found -> error
      if ( poSec == NULL ) {
         printf ( MSG_ERROR_DESTIN_SECTOR_NOT_FOUND, dwDestinSectorID );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }
   }

   // create the object instance
   CNodeObject
      *poNewObj = new CNodeObject;
   // verify error
   if ( poNewObj == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }
   poNewObj->m_oObject.mCreate ( dwInstID, this, poSec );

   // add the defines
   oDefs.mSelectStart ();
   while ( oDefs.mRetExistSelection () ) {

      // define ID
      DWORD
         dwDefID = (DWORD) ( (CNodeFloat*) oDefs.mRetSelected () )->m_fValue;

      // the define
      CNodeObject
         *poDef = (CNodeObject*) m_oDefines.mRetValue ( dwDefID );
      // verify if not found -> error
      if ( poDef == NULL ) {
         printf ( MSG_ERROR_DEFINE_NOT_FOUND, dwDefID );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }

      // add to object
      poNewObj->m_oObject.mAppendObj ( poDef->m_oObject, FALSE );

      // next define
      oDefs.mSelectNext ();
   }

   // add created object instance to respective sector
   poSec->m_oSector.mBeginAddSector ();
   poSec->m_oSector.m_oObjects.mAddKey ( dwInstID, (CNode*) poNewObj );
   poSec->m_oSector.mEndAddSector ();

   return TRUE;
}



/* Start definition of a new define in scenery
dwDefID is the define ID
return TRUE - ok / FALSE - error (ID already defined)
Obs.: before call CScenery::mEndAddDefine.
**********************************************************/
BOOL CScenery::mBeginAddDefine ( DWORD dwDefID )
{
   // if key already defined, ID exist, return
   if ( m_oDefines [ dwDefID ] != NULL ) {
      printf ( MSG_ERROR_DEFINE_ID_ALREADY_DEFINED, dwDefID );
      return FALSE;
   }

   // try create define node to store define data
   CNodeObject
      *poDef = new CNodeObject;
   if ( poDef == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // set the define ID
   poDef->m_oObject.m_dwMyID = dwDefID;

   // try define start data insertion (call for body data addition)
   // get the mesh body component
   CComponentMesh
      *poBody = (CComponentMesh*) poDef->m_oObject.mRetComponent ( SLOT_BODY );
   // verify if body found
   if ( poBody != NULL )
      // signal begin data addition
      return poBody->mActivate ( 1, 0, 0 );

   // add in scenery
   if ( ! m_oDefines.mAddKey ( dwDefID, (CNode*) poDef ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // flag define
   m_bDefining = TRUE;
   m_dwDefineID = dwDefID;       // define ID

   return TRUE;
}



/* Flag end of define. Call after include all define
data and before use this data.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CScenery::mEndAddDefine ()
{
   // update define data
   // point define
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // try define end data insertion (call for body data addition)
   // get the mesh body component
   CComponentMesh
      *poBody = (CComponentMesh*) poDef->m_oObject.mRetComponent ( SLOT_BODY );

   // verify if body not found
   if ( poBody != NULL )
      // signal end data addition
      return poBody->mActivate ( 2, 0, 0 );

   return TRUE;
}



/* Start definition of a new sector in current define.
dwSectorID is the sector ID
return TRUE - ok / FALSE - error (ID already defined,...)
Obs.: before call CScenery::mEndAddSector.
**********************************************************/
BOOL CScenery::mBeginAddSector ( DWORD dwSectorID )
{
   // define where sector will be inserted
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];
   // if point to NULL: define ID not found
   if ( poDef == NULL ) {
      printf ( MSG_ERROR_ID_NOT_FOUND );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }
   // add in object and prepare sector to receive data
   // get the sectors component
   CComponentSectorsSet
      *poSectors = (CComponentSectorsSet*) poDef->m_oObject.mRetComponent ( SLOT_INTERIOR );
   // verify if not found -> error
   if ( poSectors == NULL ) {
      printf ( MSG_ERROR_NO_TYPE );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // add sector
   CNodeSector
      *poNew = (CNodeSector*) poSectors->mActivate ( 2, dwSectorID, 0 );
   // if error
   if ( poNew == NULL )
      return FALSE;

   // try start sector data definition
   if ( ! poNew->m_oSector.mBeginAddSector () )
      return FALSE;

   // flag sector definition
   m_bDefining = FALSE;
   m_dwSectorID = dwSectorID;

   return TRUE;
}



/* Flag end of more sector definition, now is define definition.
Call after define all data and before use this data.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CScenery::mEndAddSector ()
{
   // flag define
   m_bDefining = TRUE;

   // update sector data

   // define where sector is
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // get the sectors component
   CComponentSectorsSet
      *poSectors = (CComponentSectorsSet*) poDef->m_oObject.mRetComponent ( SLOT_INTERIOR );
   // verify if not found -> error
   if ( poSectors == NULL ) {
      printf ( MSG_ERROR_NO_TYPE );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // get the specific sector
   CNodeSector
      *poSec = (CNodeSector*) poSectors->mActivate ( 5, m_dwSectorID, 0 );
   // if error
   if ( poSec == NULL )
      return FALSE;

   // signal end data modification
   return poSec->m_oSector.mEndAddSector ();
}



/* Add polygons in current scenery define or sector.
oIndexes is the list with vertex indexes that compose the polygon,
the indexes are acording vertex list in mesh.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CScenery::mBeginAddDefine or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mAddPolygon ( CList& oIndexes )
{
   // verify if include in current define
   if ( m_bDefining ) {
      // define mesh to set attribute
      // object
      CNodeObject
         *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];
      CComponentMesh
         *poMesh = (CComponentMesh*) poDef->m_oObject.mRetComponent ( SLOT_BODY );
      // set data
      return poMesh->m_oMesh.mAddPolygon ( oIndexes, m_dwDefaultLightTable, m_fDefaultLightLevel );
   }

   // update sector
   // define where sector is
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // get the sectors component
   CComponentSectorsSet
      *poSectors = (CComponentSectorsSet*) poDef->m_oObject.mRetComponent ( SLOT_INTERIOR );
   // verify if not found -> error
   if ( poSectors == NULL ) {
      printf ( MSG_ERROR_NO_TYPE );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // get the specific sector
   CNodeSector
      *poSec = (CNodeSector*) poSectors->mActivate ( 5, m_dwSectorID, 0 );
   // if error
   if ( poSec == NULL )
      return FALSE;

   // add polygon
   return poSec->m_oSector.mAddPolygon ( oIndexes, m_dwDefaultLightTable, m_fDefaultLightLevel );
}



/* Add vertexes coordenates in a define or sector
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CScenery::mBeginAddDefine or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mAddVertex ( float fX, float fY, float fZ )
{
   // verify if include in define
   if ( m_bDefining ) {

      // define mesh to set attribute
      CNodeObject
         *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];
      CComponentMesh
         *poMesh = (CComponentMesh*) poDef->m_oObject.mRetComponent ( SLOT_BODY );
      // set attributes
      return poMesh->m_oMesh.mAddVertex ( fX, fY, fZ, m_fDefaultScaleFactor, m_fDefaultLightLevel );
   }

   // update sector
   // define where sector is
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // get the sectors component
   CComponentSectorsSet
      *poSectors = (CComponentSectorsSet*) poDef->m_oObject.mRetComponent ( SLOT_INTERIOR );
   // verify if not found -> error
   if ( poSectors == NULL ) {
      printf ( MSG_ERROR_NO_TYPE );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // get the specific sector
   CNodeSector
      *poSec = (CNodeSector*) poSectors->mActivate ( 5, m_dwSectorID, 0 );
   // if error
   if ( poSec == NULL )
      return FALSE;
 
   // add vertex
   return poSec->m_oSector.mAddVertex ( fX, fY, fZ, m_fDefaultScaleFactor, m_fDefaultLightLevel );
}



/* Set some initial engine parameters.
dwCameraID - what object instance is the camera (is the viewer)
nClear - background color or -1 if no frame buffer clear is need
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mSetStart ( DWORD dwCameraID, INT nClear )
{
   // define frame buffer clear color, used as background color (-1 for none)
   m_nClearBackground = nClear;

   // what object instance have the camera control
   return mSetCurrentCamera ( dwCameraID );
}   



/* Set the default light table. This light table would be
used on next included polygons.
dwLightID - the light table ID, if -1 is used the engine default
**********************************************************/
void CScenery::mSetLightTable ( DWORD dwLightID )
{
   if ( dwLightID == (DWORD) -1 )
      m_dwDefaultLightTable = DEFAULT_LIGHT_TABLE;
   else
      m_dwDefaultLightTable = dwLightID;
}   



/* Set the default light level. This light level would be
used on next included vertexes and polygons.
fLight - the normalized (between 0 and 1) light level,
         if -1 is used the engine default
**********************************************************/
void CScenery::mSetLightLevel ( float fLight )
{
   if ( fLight == -1 )
      m_fDefaultLightLevel = DEFAULT_LIGHT_LEVEL;
   else
      m_fDefaultLightLevel = fLight;
}   



/* Set the object dwCameraID as the current camera and
Invert the vector K component Z
(does not add/remove components!).
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mSetCurrentCamera ( DWORD dwCameraID )
{
   // invert some camera components of the last camera, if exist
   if ( m_poCamera != NULL ) {
      CCameraSystem
         *poCamSys = &( (CComponentCameraSystem*) m_poCamera->m_oObject.mRetComponent ( SLOT_REFSYS ) )->m_oCamSys;
      poCamSys->mInvertCamera ();
   }
   
   // update camera
   // search for object
   CNodeObject
      *poNewCamera = (CNodeObject*) m_oRoot.m_oSector.mSearchObject ( dwCameraID );
   // verify if object not found
   if ( poNewCamera == NULL ) {
      printf ( MSG_ERROR_CAMERA_OBJECTID_NOT_FOUND, dwCameraID );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }
   CCameraSystem
      *poNewCamSys = &( (CComponentCameraSystem*) poNewCamera->m_oObject.mRetComponent ( SLOT_REFSYS ) )->m_oCamSys;
   poNewCamSys->mInvertCamera ();

   m_poCamera = poNewCamera;

   return TRUE;
}



/* Set polygons attributes to a "definition mesh" or sector mesh.
nTypeID - data type
pafData - the data
nNdxPol1 - first polygon to apply data;
nNdxPol2 - last polygon to apply data;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their data changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change data call CScenery::mBeginAddDefine or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mSetDataValues ( INT nTypeID, float* pafData, DWORD dwNdxPol1, DWORD dwNdxPol2 )
{
   // verify if include in a define
   if ( m_bDefining ) {

      // what mesh to set data
      CNodeObject
         *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];
      CComponentMesh
         *poMesh = (CComponentMesh*) poDef->m_oObject.mRetComponent ( SLOT_BODY );
      // set data
      return poMesh->m_oMesh.mSetDataValues ( nTypeID, pafData, dwNdxPol1, dwNdxPol2 );
   }

   // what sector mesh to set data

   // define where sector is
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // get the sectors component
   CComponentSectorsSet
      *poSectors = (CComponentSectorsSet*) poDef->m_oObject.mRetComponent ( SLOT_INTERIOR );
   // verify if not found -> error
   if ( poSectors == NULL ) {
      printf ( MSG_ERROR_NO_TYPE );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // get the specific sector
   CNodeSector
      *poSec = (CNodeSector*) poSectors->mActivate ( 5, m_dwSectorID, 0 );
   // if error
   if ( poSec == NULL )
      return FALSE;

   // set data
   return poSec->m_oSector.mSetDataValues ( nTypeID, pafData, dwNdxPol1, dwNdxPol2 );
}



/* Add one message to the Event system
poObj - the destin object (to send message), if NULL is used dwObjID
        to define the destin
dwObjID - destin object ID, optional if gived a valid poObj (not NULL)
dwSlotID - the slot to be activated (this is the message code)
pData1, pData2, pData3 - message parameters to object
return the message identifier or -1 if error
**********************************************************/
INT CScenery::mAddMessage ( CNodeObject* poObj, DWORD dwObjID, DWORD dwSlotID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   return m_oEventTab.mAddMessage ( *this, poObj, dwObjID, dwSlotID, pData1, pData2, pData3 );
}



/* Remove the message for the event table.
dwPos - the message identifier
**********************************************************/
void CScenery::mRemoveMessage ( DWORD dwPos )
{
   m_oEventTab.mRemoveMessage ( dwPos );
}



/* Create components into a define.
dwSlotID - slot to add the component
dwTypeID - what component add
pData1 - component creation parameter
pData2 - component creation parameter
pData3 - component creation parameter, point is a CList with
         elements of CNodeFloat.
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mCreateComponent ( DWORD dwSlotID, DWORD dwTypeID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   // what define to add component
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // try add component
   return poDef->m_oObject.mCreateComponent ( dwSlotID, dwTypeID, pData1, pData2, pData3 );
}



/* Activate the component into a define.
dwSlotID - component slot
pData1 - component parameter
pData2 - component parameter
pData3 - component parameter, point is a CList with
         elements of CNodeFloat.
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mActivateComponent ( DWORD dwSlotID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   // what define
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];

   // try run component
   return poDef->m_oObject.mActivateComponent ( dwSlotID, pData1, pData2, pData3 );
}



/* Remove components from a define.
dwSlotID - component slot
pData1 - component parameter
pData2 - component parameter
pData3 - component parameter, point is a CList with
         elements of CNodeFloat.
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mDestroyComponent ( DWORD dwSlotID, PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   // what define
   CNodeObject
      *poDef = (CNodeObject*) m_oDefines [ m_dwDefineID ];
   
   // try del component
   return poDef->m_oObject.mDestroyComponent ( dwSlotID, pData1, pData2, pData3 );
}
