/* Frame buffer view functions (graphic mode independent).
You have functions to start a specific video mode, programm palette (if the
video mode have palette) and show frame buffers on screen.

Use CDraw to draw an image in a CMem object (this CMem object is the frame
buffer), use CScreen to have access to video (write CMem objects as images
or program palette). CGr have MCGA functions.

FEATURES:

  +------------------+-----------------+--------------------------+
  |                    CURRENT VIDEO MODES                        |
  +------------------+-----------------+--------------------------+
  |    Video Mode    |   Identifier    |   Resolution and colors  |
  +------------------+-----------------+--------------------------+
  |    MCGA          |   MODE_MCGA     |   320x200 256 colors     |
  |    Text mode     |   MODE_TEXT     |   80x25 standart text    |
  +------------------+-----------------+--------------------------+

TIPS:
- you don't need to explicit call text mode when ending the aplication,
  only call CScreen::mDestroy or wait CScreen destructor execution.
- the reference system is the same as the video:   (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX

BUGS:
- mShow (2) not test screen dimensions (if you want to draw a texture bigger
  than the current video mode resolution you would got a GFP!).
- mShow (2) only draw complete bitmaps if they have width multiple of 4
*/



#include <string.h>   // necessary to access sys/movedata.h

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "texture.h"

#ifdef TRG_DJGPP
   #include <go32.h>
   #include "djgpp\gr.h"
#endif

#ifdef TRG_LINUXX11
   #include "linx11/xwinapp.h"
#endif

#include "screen.h"



/* constructor - init data
**********************************************************/
CScreen::CScreen ()
{
   m_dwScreenMode = 0;

   #ifdef TRG_LINUXX11
      m_poApp = NULL;
   #endif
   
   #ifdef TRG_DJGPP
      m_poGr = NULL;
   #endif
}



/* constructor - create a Screen
**********************************************************/
CScreen::CScreen ( DWORD dwMode )
{
   // to mDestroy work
   m_dwScreenMode = 0;
   
   #ifdef TRG_DJGPP
      m_poGr = NULL;
   #endif

   #ifdef TRG_LINUXX11
      m_poApp = NULL;
   #endif

   mCreate ( dwMode );
}



/* destructor
**********************************************************/
CScreen::~CScreen ()
{
   mDestroy ();
}



/* Change current video mode
**********************************************************/
void CScreen::mCreate ( DWORD dwMode )
{
   mDestroy ();

   // create the specific graphic object needed
   switch ( dwMode ) {

      case MODE_MCGA:

         m_dwScreenMode = MODE_MCGA;

         // if the target is linux with SvgaLib
         #ifdef TRG_LINUX
            vga_init ();
            vga_setmode ( G320x200x256 );
         #endif

         // if the target is DJGPP
         #ifdef TRG_DJGPP
            // alocate object with the specific mode
            if ( ( m_poGr = new CGr ( m_dwScreenMode ) ) == NULL )
               ABORT ( MSG_ERROR_CREATING_GR, __FILE__, __LINE__ )
         #endif
               
         // if the target is Linux X11
         #ifdef TRG_LINUXX11
            // alocate object with the specific mode
            if ( ( m_poApp = new CXWindowApp ) == NULL )
               ABORT ( MSG_ERROR_CREATING_GR, __FILE__, __LINE__ )
         #endif
         
      break;
   }
}



/* Change video mode to text mode and delete gr object
**********************************************************/
void CScreen::mDestroy ()
{
   // verify if graphics object exist
   if ( m_dwScreenMode != MODE_TEXT ) {
      
      #ifdef TRG_LINUX
         vga_setmode ( TEXT );
      #endif
      
      #ifdef TRG_DJGPP
         // set text mode
         m_poGr->mSetVideoMode ( MODE_TEXT );
         // delete object
         delete m_poGr;
         m_poGr = NULL;
      #endif
      
      #ifdef TRG_LINUXX11
         // set text mode
         m_poApp->mClose ();
         // delete object
         delete m_poApp;
         m_poApp = NULL;
      #endif
      
      m_dwScreenMode = MODE_TEXT;
   }
}



/* Show frame buffer on screen. Mode independent. Wait retrace.
**********************************************************/
void CScreen::mShow ( CMem& oMem )
{
   // select the current graphics object
   switch ( m_dwScreenMode ) {

      case MODE_MCGA:
         
         #ifdef TRG_LINUX
            
            #ifdef COMP_WAIT_RETRACE
            vga_waitretrace ();
            #endif

            // draw scanline per scanline
            for ( int ncL = 0; ncL < 200; ncL ++ )
               vga_drawscansegment ( oMem.m_payMem + ncL * 320, 0, ncL, 320 );
         #endif
      
         #ifdef TRG_DJGPP
            
            #ifdef COMP_WAIT_RETRACE
            m_poGr->mWaitRetrace ();
            #endif

            m_poGr->mShow ( oMem );
         #endif
      
         #ifdef TRG_LINUXX11
            m_poApp->mShow ( oMem );
         #endif

      break;
   }
}



/* Set palette if the current graphic mode have palette
**********************************************************/
void CScreen::mSetPalette ( CMem& oMem )
{
   // set palette
   switch ( m_dwScreenMode ) {

      case MODE_MCGA:
         
         #ifdef TRG_LINUX
            // seek all palette writing RGB
            for ( DWORD dwcCol = 0; dwcCol < 768; dwcCol += 3 )
               vga_setpalette ( dwcCol / 3, oMem.m_payMem [ dwcCol ],
                                            oMem.m_payMem [ dwcCol + 1 ],
                                            oMem.m_payMem [ dwcCol + 2 ] );
         #endif
         
         #ifdef TRG_DJGPP
            m_poGr->mWritePal ( oMem );
         #endif
         
         #ifdef TRG_LINUXX11
            m_poApp->mWritePal ( oMem );
         #endif
      
      break;
   }
}
