/* Create and manage a socket stream for Internet domain.

Implement the client side of the client/server association.

The functions can be called if sources compiled for DJGPP target
but there is no socket execution (only for compatibility).
Under Linux this class has full functionality.

TIPS:
- in case of error:
  for a more complete error description see the value of 
  the system variable errno.
*/



#include "defs.h"

#ifndef TRG_DJGPP
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#endif

#include <string.h>

#include "mem.h"

#include "string.h"

#include "socket.h"



/* constructor - init socket without connection
**********************************************************/
CSocket::CSocket ()
{
   m_nSocket = 0;
}



/* constructor - init socket with other socket
**********************************************************/
CSocket::CSocket ( INT nSock )
{
   m_nSocket = nSock;
}



/* constructor - create socket with address+port, if error abort
**********************************************************/
CSocket::CSocket ( CString oszIPAddress, UINT uPort )
{
   if ( ! mCreate ( oszIPAddress, uPort ) )
      ABORT ( MSG_ERROR_CREATING_SOCKET, __FILE__, __LINE__ )
}



/* destructor - desalocate socket
**********************************************************/
CSocket::~CSocket ()
{
   mDestroy ();
}



/* Create a socket of Internet domain and using streams
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CSocket::mCreate ()
{
   mDestroy ();

   #ifndef TRG_DJGPP
   // call socket creation and verify if return less than 0, if yes abort
   if ( ( m_nSocket = socket ( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
      return FALSE;
   #endif

   return TRUE;
}



/* Create a socket of Internet domain and using streams.
Define try connection to the given address.
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CSocket::mCreate ( CString oszIPAddress, UINT uPort )
{
   mDestroy ();

   #ifndef TRG_DJGPP
   // call socket creation and verify if return less than 0, if yes return
   if ( ( m_nSocket = socket ( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
      return FALSE;

   // try connect a socket and verify if error return FALSE
   if ( ! mConnect ( oszIPAddress, uPort ) )
      return FALSE;
   #endif

   return TRUE;
}



/* Close socket, if is open
**********************************************************/
void CSocket::mDestroy ()
{
   #ifndef TRG_DJGPP
   // close socket if is open
   if ( m_nSocket > 0 ) {
      shutdown ( m_nSocket, 0 );
      m_nSocket = 0;        // signals socket close
   }
   #endif
}



/* Assign a socket destriptor to this object.
return it's value.
**********************************************************/
INT CSocket::mSetObj ( INT nSock )
{
   m_nSocket = nSock;
   return m_nSocket;
}



/* Return the socket destriptor of this object
**********************************************************/
INT CSocket::mRetSocket ()
{
   return m_nSocket;
}



/* Make a connection, client call.
oszIPAddress - the Internet address (127.0.0.1 for loopback)
uPort - the port to make connection
return TRUE ok / FALSE error
**********************************************************/
BOOL CSocket::mConnect ( CString oszIPAddress, UINT uPort )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP

   // define destin address (server address)
   // look for the specific address in Internet
   hostent
     *pstHost = gethostbyname ( (char*) oszIPAddress );
   // define socket address
   struct sockaddr_in
      stName;
   stName.sin_family = AF_INET;
   bcopy ( pstHost->h_addr, &stName.sin_addr, pstHost->h_length );  
   stName.sin_port = htons ( uPort );

   // make connection
   if ( connect ( m_nSocket, (struct sockaddr *) &stName, sizeof ( stName ) ) == -1 )
      return FALSE;

   #endif

   return TRUE;
}



/* 1 - Send data to socket.
pvBuffer - is a buffer with data to be sended.
dwSize - the buffer size.
return TRUE - ok / FALSE - error sending data
**********************************************************/
BOOL CSocket::mSend ( void *pvBuffer, DWORD dwSize )
{
   #ifndef TRG_DJGPP
   return ! send ( m_nSocket, pvBuffer, dwSize, 0 );
   #endif

   // if TRG_DJGPP
   return 0;
}



/* 2 - Send data to socket.
The buffer is a CMem object. See
BOOL CSocket::mSend ( void *pvBuffer, DWORD dwSize )
for more info.
**********************************************************/
BOOL CSocket::mSend ( CMem& oMem )
{
   return mSend ( (void*) oMem, oMem.m_dwSize );
}



/* 1 - Receive data from socket.
pvBuffer - point data to be received
dwSize - the buffer size
&nCode (is a return value): 0 - end of connection;
                  less than 0 - error;
              greatter than 0 - the number of bytes received;
dwWaitSec - the amount of seconds to wait data to be received (default 0);
dwWaitUSec - the amount of micro seconds to wait data to be
             received (default 0);
This function can stay blocked dwWaitSec + dwWaitUSec while there is no
data to receive. If data commes during this time the function unblock,
get the data and return TRUE. Else, with during this time there is no
data to receive the function unblock and return FALSE, nCode can have
unknow value in this case.
If dwWaitSec + dwWaitUSec are 0 there is no wait time.
return TRUE - data received / FALSE no data to receive (no data sended)
**********************************************************/
BOOL CSocket::mReceive ( void *pvBuffer, DWORD dwSize, INT& nCode,
                         DWORD dwWaitSec, DWORD dwWaitUSec )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP

   // time to wait
   struct timeval
      stWait;
   stWait.tv_sec = dwWaitSec;
   stWait.tv_usec = dwWaitUSec;

   // define the set of descriptors to test
   fd_set
      stTest;
   FD_ZERO ( &stTest );
   FD_SET ( m_nSocket, &stTest );

   // Block waiting read data is ok, if the function select
   // return 0 or less -> no data to receive
   // there is data to read else no.
   if ( select ( m_nSocket + 1, &stTest, (fd_set*) 0,
                 (fd_set*) 0, &stWait ) <= 0 )
      return FALSE;

   // receive data
   nCode = recv ( m_nSocket, pvBuffer, dwSize, 0 );
   #endif

   return TRUE;
}



/* 2 - Receive data from socket. The buffer is a CMem object.
See INT CSocket::mReceive ( void *pvBuffer, DWORD dwSize ),
for more info.
**********************************************************/
BOOL CSocket::mReceive ( CMem& oMem, INT& nCode,
                         DWORD dwWaitSec, DWORD dwWaitUSec )
{
   return mReceive ( (void*) oMem, oMem.m_dwSize, nCode, dwWaitSec, dwWaitUSec );
}



/* like CSocket::mSetObj, see it for info
**********************************************************/
INT CSocket::operator = ( INT nSock )
{
   return mSetObj ( nSock );
}



/* Get this socket name: IP addredd and port.
Use the primitive getpeername.
chIP is an array of 4 chas where would be stored the IP
address as in format: pachIP [ 0 ].pachIP [ 1 ].pachIP [ 2 ].pachIP [ 3 ]
nPort - the socket port
return - TRUE ok / FALSE error
**********************************************************/
BOOL CSocket::mRetName ( char* pachIP, INT& nPort )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP

   // define socket address
   struct sockaddr_in
      stName;
   INT
      nSize = sizeof ( stName );
   // try get name
   if ( getpeername ( m_nSocket, (struct sockaddr *) &stName, &nSize) < 0 )
      return FALSE;

   // set values
   pachIP [ 0 ] = ( (char *) &stName.sin_addr.s_addr ) [ 0 ];
   pachIP [ 1 ] = ( (char *) &stName.sin_addr.s_addr ) [ 1 ];
   pachIP [ 2 ] = ( (char *) &stName.sin_addr.s_addr ) [ 2 ];
   pachIP [ 3 ] = ( (char *) &stName.sin_addr.s_addr ) [ 3 ];
   nPort  = ntohs ( stName.sin_port );

   #endif

   return TRUE;
}
