/* Create and manage a socket stream for Internet domain.

Implement the server side of client/server.

The functions can be called if sources compiled for DJGPP target
but there is no socket execution (only for compatibility).
Under Linux this class has all functionality.

TIPS:
- in case of error:
  for a more complete error description see the value of 
  the system variable errno.
- is more recomendable use a CSocket instance to send/receive
  messages.
- this class is derived from CSocket, see CSocket for more info.
- because this class derive from CSocket, it can access the 
  send/receive functions implemented from here. Use them to
  exchange messages with the client.
*/



#include "defs.h"

#ifndef TRG_DJGPP
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#endif

#include <string.h>

#include "mem.h"

#include "string.h"

#include "socket.h"

#include "svsocket.h"



/* constructor - init socket without connection
**********************************************************/
CServerSocket::CServerSocket ()
{
   m_nSocketServer = 0;
}



/* constructor - create socket with address+port, if error abort
Bind a address to socket, alocate a queue for clients.
oszIPAddress - the IP address of this socket
uPort - the port to be associated to this socket
dwMaxConnections - max number of clients in queue, waiting connection
**********************************************************/
CServerSocket::CServerSocket ( CString oszIPAddress, UINT uPort, DWORD dwMaxConnections )
{
   if ( ! mCreate ( oszIPAddress, uPort, dwMaxConnections ) )
      ABORT ( MSG_ERROR_CREATING_SOCKET, __FILE__, __LINE__ )
}



/* destructor - desalocate socket
**********************************************************/
CServerSocket::~CServerSocket ()
{
   mDestroy ();
}



/* Create a socket of Internet domain and using streams
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CServerSocket::mCreate ()
{
   mDestroy ();

   #ifndef TRG_DJGPP
   // call socket creation and verify if return less than 0, if yes abort
   if ( ( m_nSocketServer = socket ( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
      return FALSE;
   #endif

   return TRUE;
}



/* Create a socket of Internet domain and using streams
Make the binding between the socket and a name to enable clients
connections recall, alocate a queue for clients.
oszIPAddress - the IP address of this socket
uPort - the port to be associated to this socket
dwMaxConnections - max number of clients in queue, waiting connection
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CServerSocket::mCreate ( CString oszIPAddress, UINT uPort, DWORD dwMaxConnections )
{
   mDestroy ();

   #ifdef TRG_DJGPP
   return TRUE;
   #endif

   #ifndef TRG_DJGPP
   // call socket creation and verify if return less than 0, if yes abort
   if ( ( m_nSocketServer = socket ( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
      return FALSE;
   #endif

   // bind a name to socket and verify if error return FALSE
   if ( ! mBind ( oszIPAddress, uPort ) )
      return FALSE;

   // try create the queue, if error return FALSE
   if ( ! mListen ( dwMaxConnections ) )
      return FALSE;

   return TRUE;
}



/* Close socket, if is open
**********************************************************/
void CServerSocket::mDestroy ()
{
   CSocket::mDestroy ();

   #ifndef TRG_DJGPP
   // close socket if is open
   if ( m_nSocketServer > 0 ) {
      shutdown ( m_nSocketServer, 0 );
      m_nSocketServer = 0;  // signals socket close
   }
   #endif
}



/* Assign a socket destriptor to this object.
return it's value.
**********************************************************/
INT CServerSocket::mSetObj ( INT nSock )
{
   m_nSocketServer = nSock;
   return m_nSocketServer;
}



/* Return the socket destriptor of this object
**********************************************************/
INT CServerSocket::mRetSocket ()
{
   return m_nSocketServer;
}



/* Bind, associate, a name to a socket.
oszIPAddress - the Internet address (127.0.0.1 for loopback)
uPort - the port to make connection
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CServerSocket::mBind ( CString oszIPAddress, UINT uPort )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP
   
   // define destin address (server address)
   // look for the specific address in Internet
   hostent
      *pstHost = gethostbyname ( (char*) oszIPAddress );

   // define socket address
   struct sockaddr_in
      stNameServer;
   stNameServer.sin_family = AF_INET;
   bcopy ( pstHost->h_addr, &stNameServer.sin_addr, pstHost->h_length );  
   stNameServer.sin_port = htons ( uPort );

   // make association
   if ( bind ( m_nSocketServer, (struct sockaddr*) &stNameServer, sizeof ( stNameServer ) ) )
      return FALSE;
   
   #endif

   return TRUE;
}



/* Create the connection queue. After this clients can
try connections.
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CServerSocket::mListen ( DWORD dwMaxConnections )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP

   // allocate queue, enable connections
   if ( listen ( m_nSocketServer, dwMaxConnections ) )
      return FALSE;

   #endif

   return TRUE;
}



/* Enable connections, waiting a client connection.
dwWaitSec - the amount of seconds to wait a connection (default 0);
dwWaitUSec - the amount of micro seconds to wait a connection (default 0);
This function can stay blocked dwWaitSec + dwWaitUSec while there is no
connection recall from a client. If client try connection during this time
the function unblock, accept the connection and return the client socket
descriptor. Else, with during this time there is no connection the function
unblock and return 0.
If dwWaitSec + dwWaitUSec are 0 there is no wait time.
return the client socket descriptor if success or 0 if no connection
**********************************************************/
INT CServerSocket::mAccept ( DWORD dwWaitSec, DWORD dwWaitUSec )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP

   // time to wait
   struct timeval
      stWait;
   stWait.tv_sec = dwWaitSec;
   stWait.tv_usec = dwWaitUSec;

   // define the set of descriptors to test
   fd_set
      stTest;
   FD_ZERO ( &stTest );
   FD_SET ( m_nSocketServer, &stTest );

   // Block waiting connection, if the function select
   // return 0 or less -> no connection.
   if ( select ( m_nSocketServer + 1, &stTest, (fd_set*) 0,
                 (fd_set*) 0, &stWait ) <= 0 )
      return 0;

   // the client address
   struct sockaddr_in
      stNameClient;
   INT
      nSize = sizeof ( stNameClient );
   m_nSocket = accept ( m_nSocketServer, (struct sockaddr *) &stNameClient, (int*) &nSize );

   #endif

   return m_nSocket;
}



/* like CServerSocket::mSetObj, see it for info
**********************************************************/
INT CServerSocket::operator = ( INT nSock )
{
   return mSetObj ( nSock );
}



/* Get this socket name: IP addredd and port.
Use the primitive: getsockname.
pachIP is an array of 4 chas where would be stored the IP
address as in format: pachIP [ 0 ].pachIP [ 1 ].pachIP [ 2 ].pachIP [ 3 ]
nPort - the socket port
return - TRUE ok / FALSE error
**********************************************************/
BOOL CServerSocket::mRetName ( char* pachIP, INT& nPort )
{
   #ifdef TRG_DJGPP
   return FALSE;
   #endif

   #ifndef TRG_DJGPP

   // define socket address
   struct sockaddr_in
      stName;
   INT
      nSize = sizeof ( stName );
   // try get name
   if ( getsockname ( m_nSocketServer, (struct sockaddr *) &stName, &nSize) < 0 )
      return FALSE;

   // set values
   pachIP [ 0 ] = ( (char *) &stName.sin_addr.s_addr ) [ 0 ];
   pachIP [ 1 ] = ( (char *) &stName.sin_addr.s_addr ) [ 1 ];
   pachIP [ 2 ] = ( (char *) &stName.sin_addr.s_addr ) [ 2 ];
   pachIP [ 3 ] = ( (char *) &stName.sin_addr.s_addr ) [ 3 ];
   nPort  = ntohs ( stName.sin_port );

   #endif

   return TRUE;
}
