/* Implements bitmap data storage.

This bitmap can be used in texture map polygons or as a frame buffer.
See CDraw for more info about texture mapping and frame buffers.

Textures used for polygons have a fixed size defined in TEXTURE_SIZE,
file defs.h

TIPS:
- the bitmap can be load from a PCX file using the class CFilePCX
- you can load RAW bitmaps using
  BOOL CFile::mRetMem ( DWORD dwMax, CMem& oMem ). Before you must set
  bitmap x and y sizes manualy.
- this class derive from CMem, see CMem for more info.
*/



#include "defs.h"

#include "mem.h"

#include "string.h"

#include "file.h"

#include "filepcx.h"

#include "texture.h"



/* constructor - no image
**********************************************************/
CTexture::CTexture ()
{
   m_nWidth = m_nHeight = 0;
}



/* constructor - copy another texture
**********************************************************/
CTexture::CTexture ( CTexture& oTx )
{
   m_nWidth = m_nHeight = 0;
   mSetObj ( oTx );
}



/* constructor - copy and scale a source texture.
nWidth and nHeight are this texture size, oTxSrc is the
source texture (can't be the same object).
**********************************************************/
CTexture::CTexture ( INT nWidth, INT nHeight, CTexture& oTxSrc )
{
   m_nWidth = m_nHeight = 0;
   mCreate ( nWidth, nHeight, oTxSrc );
}



/* destrutor - set image dimensions to 0
**********************************************************/
CTexture::~CTexture ()
{
   mDestroy ();
}



/* Set image dimensions to 0 and desalocate memory
**********************************************************/
void CTexture::mDestroy ()
{
   m_nWidth = m_nHeight = 0;
   CMem::mDestroy ();
}



/* Create a new object copying and scaling the bitmap of
other texture. nWidth and nHeight are this texture size,
oTxSrc is the source texture (can't be the same object).
**********************************************************/
void CTexture::mCreate ( INT nWidth, INT nHeight, CTexture& oTxSrc )
{
   CMem::mDestroy ();

   // define texture
   mSetObj ( oTxSrc );
   mScale ( nWidth, nHeight );
}



/* Init texture with a copy of other object
(like CTexture::constructor and CTexture::operator =)
**********************************************************/
void CTexture::mSetObj ( CTexture& oTx )
{
   mDestroy ();

   CMem::mSetObj ( oTx );

   m_nWidth = oTx.m_nWidth;
   m_nHeight = oTx.m_nHeight;
}



/* like CTexture::constructor and CTexture::mSetObj
**********************************************************/
void CTexture::operator = ( CTexture& oTx )
{
   mSetObj ( oTx );
}



/* Change bitmap size doing scale (zoom in or out)
**********************************************************/
void CTexture::mScale ( INT nWidth, INT nHeight )
{
   // source texture
   CTexture
      oTxSrc ( *this );

   // alocate space to new size
   CMem::mCreate ( nWidth * nHeight );

   // dimensions
   m_nWidth = nWidth;
   m_nHeight = nHeight;

   // scan all destin image surface
   for ( INT ncY = 0; ncY < m_nHeight; ncY ++ )

      for ( INT ncX = 0; ncX < m_nWidth; ncX ++ )

         // copy and scale
         m_payMem [ ncY * m_nWidth + ncX ] = oTxSrc.m_payMem [
          ( ( ncX * oTxSrc.m_nWidth  ) / m_nWidth  ) +
          ( ( ncY * oTxSrc.m_nHeight ) / m_nHeight ) * oTxSrc.m_nWidth ];
}
